@@ -1,12 +1,10 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-
 /*
- * $Id: cpp.c,v 1.88 2003/09/30 16:40:00 grubba Exp $
- */
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: cpp.c,v 1.111 2004/03/31 14:28:22 grubba Exp $
+*/
+
 #include "global.h"
 #include "stralloc.h"
 #include "module_support.h"
@@ -29,6 +27,11 @@
 
 #include <ctype.h>
 
+#define sp Pike_sp
+
+#undef ATTRIBUTE
+#define ATTRIBUTE(X)
+
 #define CPP_NO_OUTPUT 1
 #define CPP_EXPECT_ELSE 2
 #define CPP_EXPECT_ENDIF 4
@@ -57,7 +60,9 @@ struct pike_predef_s
   char *value;
 };
 
-static struct pike_predef_s *pike_predefs=0;
+static int use_initial_predefs;
+static struct pike_predef_s *first_predef = NULL, *last_predef = NULL;
+
 struct define_part
 {
   int argument;
@@ -118,11 +123,8 @@ void cpp_error(struct cpp *this,char *err)
     ref_push_string(this->current_file);
     push_int(this->current_line);
     push_text(err);
-    if (this->handler && this->handler->prog) {
-      safe_apply(this->handler, "compile_error", 3);
-    } else {
-      SAFE_APPLY_MASTER("compile_error", 3);
-    }
+    low_safe_apply_handler("compile_error", this->handler,
+			   this->compat_handler, 3);
     pop_stack();
   }else{
     (void)fprintf(stderr, "%s:%ld: %s\n",
@@ -133,6 +135,88 @@ void cpp_error(struct cpp *this,char *err)
   }
 }
 
+void cpp_error_vsprintf(struct cpp *this, const char *fmt, va_list args)
+{
+  char buf[8192];
+
+#ifdef HAVE_VSNPRINTF
+  vsnprintf(buf, 8190, fmt, args);
+#else /* !HAVE_VSNPRINTF */
+  VSPRINTF(buf, fmt, args);
+#endif /* HAVE_VSNPRINTF */
+  if((size_t)strlen(buf) >= (size_t)sizeof(buf))
+    Pike_fatal("Buffer overflow in cpp_error.\n");
+
+  cpp_error(this, buf);
+}
+
+void cpp_error_sprintf(struct cpp *this, const char *fmt, ...)
+     ATTRIBUTE((format(printf,2,3)))
+{
+  va_list args;
+  va_start(args, fmt);
+  cpp_error_vsprintf(this, fmt, args);
+  va_end(args);
+}
+
+void cpp_handle_exception(struct cpp *this, const char *cpp_error_fmt, ...)
+     ATTRIBUTE((format(printf,2,3)))
+{
+  struct svalue thrown;
+  move_svalue (&thrown, &throw_value);
+  throw_value.type = T_INT;
+
+  if (cpp_error_fmt) {
+    va_list args;
+    va_start (args, cpp_error_fmt);
+    cpp_error_vsprintf (this, cpp_error_fmt, args);
+    va_end (args);
+  }
+
+  push_svalue(&thrown);
+  low_safe_apply_handler("compile_exception", error_handler, compat_handler, 1);
+
+  if (SAFE_IS_ZERO(sp-1)) {
+    /* FIXME: Doesn't handle wide string error messages. */
+    struct pike_string *s = format_exception_for_error_msg (&thrown);
+    if (s) {
+      if (!s->size_shift) cpp_error (this, s->str);
+      free_string (s);
+    }
+  }
+
+  pop_stack();
+  free_svalue(&thrown);
+}
+
+void cpp_describe_exception(struct cpp *this, struct svalue *thrown)
+{
+  /* FIXME: Doesn't handle wide string error messages. */
+  struct pike_string *s = 0;
+
+  if ((thrown->type == T_ARRAY) && thrown->u.array->size &&
+      (thrown->u.array->item[0].type == T_STRING)) {
+    /* Old-style backtrace */
+    s = thrown->u.array->item[0].u.string;
+  } else if (thrown->type == T_OBJECT) {
+    struct generic_error_struct *ge;
+    if ((ge = (struct generic_error_struct *)
+	 get_storage(thrown->u.object, generic_error_program))) {
+      s = ge->desc;
+    }
+  }
+
+  if (s && !s->size_shift) {
+    ref_push_string(s);
+    f_string_trim_all_whites(1);
+    push_constant_text("\n");
+    push_constant_text(" ");
+    f_replace(3);
+    cpp_error(this, sp[-1].u.string->str);
+    pop_stack();
+  }
+}
+
 void cpp_change_compat(struct cpp *this, int major, int minor)
 {
   if(this->compat_major == major &&
@@ -157,12 +241,29 @@ void cpp_change_compat(struct cpp *this, int major, int minor)
   if(sp[-1].type == T_OBJECT)
   {
     this->compat_handler=sp[-1].u.object;
+    dmalloc_touch_svalue(Pike_sp-1);
     sp--;
   }
   this->compat_major=major;
   this->compat_minor=minor;
 }
 
+static struct mapping *initial_predefs_mapping()
+{
+  struct pike_predef_s *def;
+  struct mapping *map = allocate_mapping (0);
+#ifdef PIKE_DEBUG
+  if (!use_initial_predefs) Pike_fatal ("Initial predefs has been taken over.\n");
+#endif
+  for (def = first_predef; def; def = def->next) {
+    push_string (make_shared_string (def->name));
+    push_string (make_shared_string (def->value));
+    mapping_insert (map, sp - 2, sp - 1);
+    pop_n_elems (2);
+  }
+  return map;
+}
+
 /* devours one reference to 'name'! */
 static struct define *alloc_empty_define(struct pike_string *name,
 					 ptrdiff_t parts)
@@ -177,6 +278,7 @@ static struct define *alloc_empty_define(struct pike_string *name,
   def->num_parts=parts;
   def->first=0;
   def->link.s=name;
+  debug_malloc_touch(name);
   return def;
 }
 
@@ -214,6 +316,17 @@ static void do_magic_define(struct cpp *this,
   this->defines=hash_insert(this->defines, & def->link);
 }
 
+static void add_define(struct cpp *this,
+		       struct pike_string *name,
+		       struct pike_string *what)
+{
+  struct define* def;
+  add_ref (name);
+  def=alloc_empty_define(name,0);
+  add_ref (def->first = what);
+  this->defines=hash_insert(this->defines, & def->link);
+}
+
 static void simple_add_define(struct cpp *this,
 			    char *name,
 			    char *what)
@@ -369,9 +482,9 @@ static void simple_add_define(struct cpp *this,
     }								\
     if (data[pos] == '\\') {					\
       if (data[pos+1] == '\n') {				\
-	pos+=2;                                                 \
+	pos+=2;							\
       } else if ((data[pos+1] == '\r') &&			\
-		 (data[pos+2] == '\n')) {                       \
+		 (data[pos+2] == '\n')) {			\
 	pos+=3;							\
       } else {							\
 	break;							\
@@ -937,7 +1050,7 @@ static ptrdiff_t low_cpp(struct cpp *this, void *data, ptrdiff_t len,
     return lower_cpp2(this, (p_wchar2 *)data, len,
 		      flags, auto_convert, charset);
   default:
-    fatal("low_cpp(): Bad shift: %d\n", shift);
+    Pike_fatal("low_cpp(): Bad shift: %d\n", shift);
   }
   /* NOT_REACHED */
   return 0;
@@ -1042,7 +1155,7 @@ static void check_defined(struct cpp *this,
     s=binary_findstring2((p_wchar2 *)args[0].arg.ptr, args[0].len);
     break;
   default:
-    fatal("cpp(): Symbol has unsupported shift: %d\n", args[0].arg.shift);
+    Pike_fatal("cpp(): Symbol has unsupported shift: %d\n", args[0].arg.shift);
     break;
   }
   if(s && find_define(s))
@@ -1053,67 +1166,7 @@ static void check_defined(struct cpp *this,
   }
 }
 
-static void dumpdef(struct cpp *this,
-		    struct define *def,
-		    struct define_argument *args,
-		    struct string_builder *tmp)
-{
-  struct pike_string *s = NULL;
-  struct define *d;
-
-  switch(args[0].arg.shift) {
-  case 0:
-    s=binary_findstring((char *)args[0].arg.ptr, args[0].len);
-    break;
-  case 1:
-    s=binary_findstring1((p_wchar1 *)args[0].arg.ptr, args[0].len);
-    break;
-  case 2:
-    s=binary_findstring2((p_wchar2 *)args[0].arg.ptr, args[0].len);
-    break;
-  default:
-    fatal("cpp(): Bad shift in macroname: %d\n", args[0].arg.shift);
-    break;
-  }
-  if(s && (d=find_define(s)))
-  {
-    int e;
-    char buffer[42];
-    PUSH_STRING_SHIFT(d->link.s->str, d->link.s->len,
-		      d->link.s->size_shift, tmp);
-    if(d->magic)
-    {
-      string_builder_binary_strcat(tmp, " defined magically ", 19);
-    }else{
-      string_builder_binary_strcat(tmp, " defined as ", 12);
-      
-      if(d->first)
-	PUSH_STRING_SHIFT(d->first->str, d->first->len,
-			  d->first->size_shift, tmp);
-      for(e=0;e<d->num_parts;e++)
-      {
-	if(!(d->parts[e].argument & DEF_ARG_NOPRESPACE))
-	  string_builder_putchar(tmp, ' ');
-	
-	if(d->parts[e].argument & DEF_ARG_STRINGIFY)
-	  string_builder_putchar(tmp, '#');
-	
-	sprintf(buffer,"%ld",(long)(d->parts[e].argument & DEF_ARG_MASK));
-	string_builder_binary_strcat(tmp, buffer, strlen(buffer));
-	
-	if(!(d->parts[e].argument & DEF_ARG_NOPOSTSPACE))
-	  string_builder_putchar(tmp, ' ');
-	
-	PUSH_STRING_SHIFT(d->parts[e].postfix->str, d->parts[e].postfix->len,
-			  d->parts[e].postfix->size_shift, tmp);
-      } 
-    }
-  }else{
-    string_builder_binary_strcat(tmp, " 0 ", 3);
-  }
-}
-
-static int do_safe_index_call(struct pike_string *s);
+static int do_safe_index_call(struct cpp *this, struct pike_string *s);
 
 static void check_constant(struct cpp *this,
 			  struct define *def,
@@ -1162,28 +1215,71 @@ static void check_constant(struct cpp *this,
 	push_int(0);
       }
 
-      safe_apply_handler("resolv", this->handler, this->compat_handler, 3);
+      if (safe_apply_handler("resolv", this->handler,
+			     this->compat_handler, 3, 0))
+	res = !(SAFE_IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED);
+      else {
+	if (throw_value.type == T_STRING && !throw_value.u.string->size_shift) {
+	  cpp_error(this, throw_value.u.string->str);
+	  free_svalue(&throw_value);
+	  throw_value.type = T_INT;
+	  res = 0;
+	}
+	else {
+	  struct svalue thrown = throw_value;
+	  throw_value.type = T_INT;
 
-      res=(throw_value.type!=T_STRING) &&
-	(!(IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED));
+	  if (!data.shift) {
+	    char *str = malloc(dlen + 1);
+	    MEMCPY(str, data.ptr, dlen);
+	    str[dlen] = 0;
+	    cpp_error_sprintf(this, "Error resolving '%s'.", str);
+	    free(str);
+	  }
+	  else
+	    cpp_error(this, "Error resolving identifier.");
+
+	  push_svalue(&thrown);
+	  low_safe_apply_handler("compile_exception", this->handler,
+				 this->compat_handler, 1);
+	  if (SAFE_IS_ZERO(sp-1)) cpp_describe_exception(this, &thrown);
+	  pop_stack();
+	  free_svalue(&thrown);
+	  res = 0;
+	}
+      }
     }else{
       res=0;
     }
   }else{
-    /* Handle contant(.foo) */
+    /* Handle constant(.foo) */
     push_text(".");
     ref_push_string(this->current_file);
 
-    /* FIXME: use this->compat_handler */
-    if (this->handler && this->handler->prog) {
+    if (this->handler) {
       ref_push_object(this->handler);
-      SAFE_APPLY_MASTER("handle_import", 3);
     } else {
-      SAFE_APPLY_MASTER("handle_import", 2);
+      push_int(0);
+    }
+
+    if (safe_apply_handler("handle_import", this->handler,
+			   this->compat_handler, 3,
+			   BIT_MAPPING|BIT_OBJECT|BIT_PROGRAM))
+      res = !(SAFE_IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED);
+    else {
+      struct svalue thrown = throw_value;
+      throw_value.type = T_INT;
+
+      cpp_error(this, "Error importing '.'.");
+
+      push_svalue(&thrown);
+      low_safe_apply_handler("compile_exception", this->handler,
+			     this->compat_handler, 1);
+      if (SAFE_IS_ZERO(sp-1)) cpp_describe_exception(this, &thrown);
+      pop_stack();
+      free_svalue(&thrown);
+      res = 0;
     }
-    
-    res=(throw_value.type!=T_STRING) &&
-      (!(IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED));
   }
 
   while(1)
@@ -1269,8 +1365,8 @@ static void check_constant(struct cpp *this,
         struct pike_string *s = begin_wide_shared_string(dlen, data.shift);
 	MEMCPY(s->str, data.ptr, dlen<<data.shift);
 	s = end_shared_string(s);
-	res=do_safe_index_call(s);
-        free_string(s);
+	res=do_safe_index_call(this, s);
+	free_string(s);
       }
     }else{
       cpp_error(this, "Garbage characters in constant()\n");
@@ -1283,35 +1379,168 @@ static void check_constant(struct cpp *this,
 }
 
 
-static int do_safe_index_call(struct pike_string *s)
+static int do_safe_index_call(struct cpp *this, struct pike_string *s)
 {
   int res;
   JMP_BUF recovery;
   if(!s) return 0;
-  
-  if (SETJMP(recovery)) {
+
+  if (SETJMP_SP(recovery, 1)) {
+    if (!s->size_shift)
+      cpp_handle_exception (this, "Error indexing module with \"%s\".", s->str);
+    else
+      cpp_handle_exception (this, "Error indexing module in '.' operator.");
     res = 0;
-    free_svalue(&throw_value);
-    throw_value.type = T_INT;
+    push_undefined();
   } else {
     ref_push_string(s);
     f_index(2);
     
-    res=!(IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED);
+    res=!(UNSAFE_IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED);
   }
   UNSETJMP(recovery);
   return res;
 }
 
+/*! @decl constant __VERSION__
+ *!
+ *! This define contains the current Pike version as a float. If
+ *! another Pike version is emulated, this define is updated
+ *! accordingly.
+ *!
+ *! @seealso
+ *!   @[__REAL_VERSION__]
+ */
 
-/* string cpp(string data, string|void current_file,
- *            int|string|void charset, object|void handler,
- *            void|int compat_major, void|int compat_minor);
+/*! @decl constant __REAL_VERSION__
+ *!
+ *! This define always contains the version of the current Pike,
+ *! represented as a float.
+ *!
+ *! @seealso
+ *!   @[__VERSION__]
+ */
+
+/*! @decl constant __MAJOR__
+ *!
+ *! This define contains the major part of the current Pike version,
+ *! represented as an integer. If another Pike version is emulated,
+ *! this define is updated accordingly.
+ *!
+ *! @seealso
+ *!   @[__REAL_MAJOR__]
+ */
+
+/*! @decl constant __REAL_MAJOR__
+ *!
+ *! This define always contains the major part of the version of the
+ *! current Pike, represented as an integer.
+ *!
+ *! @seealso
+ *!   @[__MAJOR__]
+ */
+
+/*! @decl constant __MINOR__
+ *! This define contains the minor part of the current Pike version,
+ *! represented as an integer. If another Pike version is emulated,
+ *! this define is updated accordingly.
+ *!
+ *! @seealso
+ *!   @[__REAL_MINOR__]
+ */
+
+/*! @decl constant __REAL_MINOR__
+ *!
+ *! This define always contains the minor part of the version of the
+ *! current Pike, represented as an integer.
+ *!
+ *! @seealso
+ *!   @[__MINOR__]
+ */
+
+/*! @decl constant __BUILD__
+ *! This constant contains the build number of the current Pike version,
+ *! represented as an integer. If another Pike version is emulated,
+ *! this constant remains unaltered.
+ *!
+ *! @seealso
+ *!   @[__REAL_MINOR__]
+ */
+
+/*! @decl constant __REAL_BUILD__
+ *!
+ *! This define always contains the minor part of the version of the
+ *! current Pike, represented as an integer.
+ *!
+ *! @seealso
+ *!   @[__BUILD__]
+ */
+
+/*! @decl constant __LINE__
+ *!
+ *! This define contains the current line number, represented as an
+ *! integer, in the source file.
+ */
+
+/*! @decl constant __FILE__
+ *!
+ *! This define contains the file path and name of the source file.
+ */
+
+/*! @decl constant __DATE__
+ *!
+ *! This define contains the current date at the time of compilation,
+ *! e.g. "Jul 28 2001".
+ */
+
+/*! @decl constant __TIME__
+ *!
+ *! This define contains the current time at the time of compilation,
+ *! e.g. "12:20:51".
+ */
+
+/*! @decl constant __PIKE__
+ *!
+ *! This define is always true.
+ */
+
+/*! @decl constant __AUTO_BIGNUM__
+ *!
+ *! This define is defined when automatic bignum conversion is enabled.
+ *! When enabled all integers will automatically be converted to
+ *! bignums when they get bigger than what can be represented by
+ *! an integer, hampering performance slightly instead of crashing
+ *! the program.
+ */
+
+/*! @decl constant __NT__
+ *!
+ *! This define is defined when the Pike is running on a Microsoft Windows OS,
+ *! not just Microsoft Windows NT, as the name implies.
+ */
+
+/*! @decl constant __amigaos__
+ *!
+ *! This define is defined when the Pike is running on Amiga OS.
+ */
+
+/*! @decl string cpp(string data, string|void current_file, @
+ *!                  int|string|void charset, object|void handler, @
+ *!                  void|int compat_major, void|int compat_minor)
+ *!
+ *! Run a string through the preprocessor.
+ *!
+ *! Preprocesses the string @[data] with Pike's builtin ANSI-C look-alike
+ *! preprocessor. If the @[current_file] argument has not been specified,
+ *! it will default to @expr{"-"@}. @[charset] defaults to @expr{"ISO-10646"@}.
+ *!
+ *! @seealso
+ *!   @[compile()]
  */
 void f_cpp(INT32 args)
 {
   struct pike_string *data;
-  struct pike_predef_s *tmpf;
+  struct mapping *predefs = NULL;
   struct svalue *save_sp = sp - args;
   struct cpp this;
   int auto_convert = 0;
@@ -1322,10 +1551,10 @@ void f_cpp(INT32 args)
 #endif /* PIKE_DEBUG */
 
   if(args<1)
-    Pike_error("Too few arguments to cpp()\n");
+    SIMPLE_TOO_FEW_ARGS_ERROR("cpp", 1);
 
   if(sp[-args].type != T_STRING)
-    Pike_error("Bad argument 1 to cpp()\n");
+    SIMPLE_BAD_ARG_ERROR("cpp", 1, "string");
 
   data = sp[-args].u.string;
 
@@ -1339,13 +1568,13 @@ void f_cpp(INT32 args)
     if(args > 5)
     {
       if(sp[4-args].type != T_INT)
-	Pike_error("Bad argument 5 to cpp()\n");
+	SIMPLE_BAD_ARG_ERROR("cpp", 5, "int");
       if(sp[5-args].type != T_INT)
-	Pike_error("Bad argument 6 to cpp()\n");
+	SIMPLE_BAD_ARG_ERROR("cpp", 6, "int");
     }
     if(sp[1-args].type != T_STRING) {
       free_string(data);
-      Pike_error("Bad argument 2 to cpp()\n");
+      SIMPLE_BAD_ARG_ERROR("cpp", 2, "string");
     }
     copy_shared_string(this.current_file, sp[1-args].u.string);
 
@@ -1366,15 +1595,17 @@ void f_cpp(INT32 args)
 	auto_convert = sp[2-args].u.integer;
       } else {
 	free_string(data);
-	Pike_error("Bad argument 3 to cpp()\n");
+	SIMPLE_BAD_ARG_ERROR("cpp", 3, "string|int");
       }
       if (args > 3) {
 	if (sp[3-args].type == T_OBJECT) {
-	  this.handler = sp[3-args].u.object;
+	  if ((this.handler = sp[3-args].u.object)) {
+	    add_ref(this.handler);
+	  }
 	} else if (sp[3-args].type != T_INT) {
 	  free_string(data);
 	  free_string(this.current_file);
-	  Pike_error("Bad argument 4 to cpp()\n");
+	  SIMPLE_BAD_ARG_ERROR("cpp", 4, "object");
 	}
       }
     }
@@ -1382,6 +1613,67 @@ void f_cpp(INT32 args)
     this.current_file=make_shared_string("-");
   }
 
+  this.compat_major=PIKE_MAJOR_VERSION;
+  this.compat_minor=PIKE_MINOR_VERSION;
+  if(args > 5)
+  {
+    cpp_change_compat(&this, sp[4-args].u.integer, sp[5-args].u.integer);
+  }
+
+  if (use_initial_predefs)
+    /* Typically compiling the master here. */
+    predefs = initial_predefs_mapping();
+  else {
+    low_unsafe_apply_handler ("get_predefines", this.handler, this.compat_handler, 0);
+    if (!UNSAFE_IS_ZERO (sp - 1)) {
+      struct keypair *k;
+      int e, sprintf_args = 0;
+      if (sp[-1].type != T_MAPPING) {
+	push_constant_text ("Invalid return value from get_predefines\n");
+	push_constant_text ("Invalid return value from get_predefines, got %O\n");
+	push_svalue (sp - 3);
+	sprintf_args = 2;
+      }
+      else {
+	predefs = copy_mapping (sp[-1].u.mapping);
+	NEW_MAPPING_LOOP (predefs->data) {
+	  if (k->ind.type != T_STRING || !k->ind.u.string->len) {
+	    push_constant_text ("Expected nonempty string as predefine name\n");
+	    push_constant_text ("Expected nonempty string as predefine name, got %O\n");
+	    push_svalue (&k->ind);
+	    sprintf_args = 2;
+	    free_mapping (predefs);
+	    predefs = NULL;
+	    goto predef_map_error;
+	  }
+	  if (k->val.type != T_STRING &&
+	      (k->val.type != T_INT || k->val.u.integer)) {
+	    push_constant_text ("Expected zero or string value for predefine\n");
+	    push_constant_text ("Expected zero or string value for predefine %O\n");
+	    push_svalue (&k->ind);
+	    sprintf_args = 2;
+	    free_mapping (predefs);
+	    predefs = NULL;
+	    goto predef_map_error;
+	  }
+	}
+      }
+      if (!predefs) {
+      predef_map_error:
+	free_string (data);
+	free_string (this.current_file);
+	if (this.handler) free_object (this.handler);
+	if (this.compat_handler) free_object (this.compat_handler);
+	f_sprintf (sprintf_args);
+	if (!sp[-1].u.string->size_shift)
+	  Pike_error ("%s", sp[-1].u.string->str);
+	else
+	  Pike_error ("%s", sp[-2].u.string->str);
+      }
+    }
+    pop_stack();
+  }
+
   if (auto_convert && (!data->size_shift) && (data->len > 1)) {
     /* Try to determine if we need to recode the string */
     struct pike_string *new_data = recode_string(data);
@@ -1399,8 +1691,6 @@ void f_cpp(INT32 args)
   this.current_line=1;
   this.compile_errors=0;
   this.defines=0;
-  this.compat_major=PIKE_MAJOR_VERSION;
-  this.compat_minor=PIKE_MINOR_VERSION;
 
   do_magic_define(&this,"__LINE__",insert_current_line);
   do_magic_define(&this,"__FILE__",insert_current_file_as_string);
@@ -1410,13 +1700,6 @@ void f_cpp(INT32 args)
   do_magic_define(&this,"__MAJOR__",insert_current_major);
   do_magic_define(&this,"__MINOR__",insert_current_minor);
 
-  {
-    struct define* def=alloc_empty_define(make_shared_string("__dumpdef"),0);
-    def->magic=dumpdef;
-    def->args=1;
-    this.defines=hash_insert(this.defines, & def->link);
-  }
-
   {
     simple_add_define(&this, "__PIKE__", " 1 ");
 
@@ -1442,8 +1725,17 @@ void f_cpp(INT32 args)
 #endif
   }
 
-  for (tmpf=pike_predefs; tmpf; tmpf=tmpf->next)
-    simple_add_define(&this, tmpf->name, tmpf->value);
+  if (predefs) {
+    struct keypair *k;
+    int e;
+    NEW_MAPPING_LOOP (predefs->data) {
+      if (k->val.type == T_STRING)
+	add_define (&this, k->ind.u.string, k->val.u.string);
+      else
+	add_define (&this, k->ind.u.string, empty_pike_string);
+    }
+    free_mapping (predefs);
+  }
 
   string_builder_binary_strcat(&this.buf, "# 1 ", 4);
   PUSH_STRING_SHIFT(this.current_file->str, this.current_file->len,
@@ -1454,11 +1746,6 @@ void f_cpp(INT32 args)
   SET_ONERROR(tmp, fatal_on_error, "Preprocessor exited with longjump!\n");
 #endif /* PIKE_DEBUG */
 
-  if(args > 5)
-  {
-    cpp_change_compat(&this, sp[4-args].u.integer, sp[5-args].u.integer);
-  }
-
 
   low_cpp(&this, data->str, data->len, data->size_shift,
 	  0, auto_convert, charset);
@@ -1476,6 +1763,12 @@ void f_cpp(INT32 args)
     this.compat_handler=0;
   }
 
+  if (this.handler)
+  {
+    free_object(this.handler);
+    this.handler = 0;
+  }
+
   free_string(this.current_file);
 
   free_string(data);
@@ -1483,15 +1776,38 @@ void f_cpp(INT32 args)
   if(this.compile_errors)
   {
     free_string_builder(&this.buf);
-    Pike_error("Cpp() failed\n");
+    throw_error_object(low_clone(cpp_error_program), 0, 0, 0,
+		       "Cpp() failed\n");
   }else{
     pop_n_elems(sp - save_sp);
     push_string(finish_string_builder(&this.buf));
   }
 }
 
+void f__take_over_initial_predefines (INT32 args)
+{
+  pop_n_elems (args);
+  if (use_initial_predefs) {
+    struct pike_predef_s *tmp;
+    push_mapping (initial_predefs_mapping());
+    use_initial_predefs = 0;
+
+    while((tmp=first_predef))
+    {
+      free(tmp->name);
+      free(tmp->value);
+      first_predef=tmp->next;
+      free((char *)tmp);
+    }
+    last_predef = 0;
+  }
+  else Pike_error ("Initial predefines already taken over.\n");
+}
+
 void init_cpp()
 {
+  struct svalue s;
+
   defined_macro=alloc_empty_define(make_shared_string("defined"),0);
   defined_macro->magic=check_defined;
   defined_macro->args=1;
@@ -1500,14 +1816,27 @@ void init_cpp()
   constant_macro->magic=check_constant;
   constant_macro->args=1;
 
-  
-/* function(string, string|void, int|string|void, object|void:string) */
+  use_initial_predefs = 1;
+
+/* function(string, string|void, int|string|void, object|void, int|void, int|void:string) */
   ADD_EFUN("cpp", f_cpp, tFunc(tStr tOr(tStr,tVoid)
 			       tOr(tInt,tOr(tStr,tVoid))
 			       tOr(tObj,tVoid)
 			       tOr(tInt,tVoid)
 			       tOr(tInt,tVoid)
-			       , tStr), OPT_EXTERNAL_DEPEND);
+			       , tStr),
+	   /* OPT_SIDE_EFFECT since we might instantiate modules etc. */
+	   OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+
+  /* Somewhat tricky to add a _constant_ function in _static_modules.Builtin. */
+  s.type = T_FUNCTION;
+  s.subtype = FUNCTION_BUILTIN;
+  s.u.efun = make_callable (f__take_over_initial_predefines,
+			    "_take_over_initial_predefines",
+			    "function(void:mapping(string:string))",
+			    OPT_SIDE_EFFECT, NULL, NULL);
+  simple_add_constant ("_take_over_initial_predefines", &s, 0);
+  free_svalue (&s);
 }
 
 
@@ -1530,19 +1859,23 @@ void add_predefine(char *s)
     tmp->value=(char *)xalloc(4);
     MEMCPY(tmp->value," 1 ",4);
   }
-  tmp->next=pike_predefs;
-  pike_predefs=tmp;
+  tmp->next = NULL;
+  if (first_predef) {
+    last_predef->next = tmp;
+    last_predef = tmp;
+  }
+  else first_predef = last_predef = tmp;
 }
 
 void exit_cpp(void)
 {
 #ifdef DO_PIKE_CLEANUP
   struct pike_predef_s *tmp;
-  while((tmp=pike_predefs))
+  while((tmp=first_predef))
   {
     free(tmp->name);
     free(tmp->value);
-    pike_predefs=tmp->next;
+    first_predef=tmp->next;
     free((char *)tmp);
   }
   free_string(defined_macro->link.s);
