@@ -1,11 +1,12 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-/**/
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: main.c,v 1.163 2004/05/14 08:49:40 grubba Exp $
+*/
+
 #include "global.h"
-RCSID("$Id: main.c,v 1.116 2003/05/07 21:01:03 mast Exp $");
+RCSID("$Id: main.c,v 1.163 2004/05/14 08:49:40 grubba Exp $");
 #include "fdlib.h"
 #include "backend.h"
 #include "module.h"
@@ -29,9 +30,17 @@ RCSID("$Id: main.c,v 1.116 2003/05/07 21:01:03 mast Exp $");
 #include "cpp.h"
 #include "main.h"
 #include "operators.h"
+#include "rbtree.h"
 #include "security.h"
 #include "constants.h"
 #include "version.h"
+#include "program.h"
+#include "rusage.h"
+#include "module_support.h"
+
+#ifdef AUTO_BIGNUM
+#include "bignum.h"
+#endif
 
 #if defined(__linux__) && defined(HAVE_DLOPEN) && defined(HAVE_DLFCN_H)
 #include <dlfcn.h>
@@ -56,6 +65,14 @@ RCSID("$Id: main.c,v 1.116 2003/05/07 21:01:03 mast Exp $");
 int try_use_mmx;
 #endif
 
+/* Define this to trace the execution of main(). */
+/* #define TRACE_MAIN */
+
+#ifdef TRACE_MAIN
+#define TRACE(X)	fprintf X
+#else /* !TRACE_MAIN */
+#define TRACE(X)
+#endif /* TRACE_MAIN */
 
 char *master_file;
 char **ARGV;
@@ -69,9 +86,9 @@ PMOD_EXPORT int default_t_flag=0;
 PMOD_EXPORT int a_flag=0;
 PMOD_EXPORT int l_flag=0;
 PMOD_EXPORT int p_flag=0;
-#ifdef YYDEBUG
+#if defined(YYDEBUG) || defined(PIKE_DEBUG)
 extern int yydebug;
-#endif /* YYDEBUG */
+#endif /* YYDEBUG || PIKE_DEBUG */
 static long instructions_left;
 
 #define MASTER_COOKIE "(#*&)@(*&$Master Cookie:"
@@ -146,9 +163,12 @@ static void get_master_key(HKEY cat)
   HKEY k;
   char buffer[4096];
   DWORD len=sizeof(buffer)-1,type=REG_SZ;
-  long ret;
+
   if(RegOpenKeyEx(cat,
-		  (LPCTSTR)"SOFTWARE\\Idonex\\Pike\\7.2",
+		  (LPCTSTR)("SOFTWARE\\Pike\\"
+			    DEFINETOSTR(PIKE_MAJOR_VERSION)
+			    "."
+			    DEFINETOSTR(PIKE_MINOR_VERSION)),
 		  0,KEY_READ,&k)==ERROR_SUCCESS)
   {
     if(RegQueryValueEx(k,
@@ -168,34 +188,86 @@ static void get_master_key(HKEY cat)
 int dbm_main(int argc, char **argv)
 {
   JMP_BUF back;
-  int e, num, do_backend;
+  int e, num;
   char *p;
   struct array *a;
 #ifdef DECLARE_ENVIRON
   extern char **environ;
 #endif
 
+  TRACE((stderr, "dbm_main()\n"));
+
+  /* Attempt to make sure stderr is unbuffered. */
+#ifdef HAVE_SETVBUF
+  setvbuf(stderr, NULL, _IONBF, 0);
+#else /* !HAVE_SETVBUF */
+#ifdef HAVE_SETBUF
+  setbuf(stderr, NULL);
+#endif /* HAVE_SETBUF */
+#endif /* HAVE_SETVBUF */
+
+  TRACE((stderr, "Init CPU lib...\n"));
+  
+  init_pike_cpulib();
+
 #ifdef TRY_USE_MMX
+  TRACE((stderr, "Init MMX...\n"));
+  
   try_use_mmx=mmx_ok();
 #endif
 #ifdef OWN_GETHRTIME
 /* initialize our own gethrtime conversion /Mirar */
+  TRACE((stderr, "Init gethrtime...\n"));
+  
   own_gethrtime_init();
 #endif
 
   ARGV=argv;
 
+  TRACE((stderr, "Main init...\n"));
+  
   fd_init();
+  {
+    extern void init_mapping_blocks(void);
+    extern void init_callable_blocks(void);
+    extern void init_gc_frame_blocks(void);
+    extern void init_pike_frame_blocks(void);
+    extern void init_node_s_blocks(void);
+    extern void init_object_blocks(void);
+    extern void init_callback_blocks(void);
+
+    init_mapping_blocks();
+    init_callable_blocks();
+    init_gc_frame_blocks();
+    init_pike_frame_blocks();
+    init_node_s_blocks();
+    init_object_blocks();
+#ifndef DEBUG_MALLOC
+    /* This has already been done by initialize_dmalloc(). */
+    init_callback_blocks();
+#endif /* !DEBUG_MALLOC */
+#ifdef PIKE_NEW_MULTISETS
+    init_multiset();
+#endif
+    init_builtin_constants();
+  }
 
 #ifdef SHARED_NODES
+  TRACE((stderr, "Init shared nodes...\n"));
+  
   node_hash.table = malloc(sizeof(node *)*32831);
   if (!node_hash.table) {
-    fatal("Out of memory!\n");
+    Pike_fatal("Out of memory!\n");
   }
   MEMSET(node_hash.table, 0, sizeof(node *)*32831);
   node_hash.size = 32831;
 #endif /* SHARED_NODES */
 
+#ifdef HAVE_TZSET
+  tzset();
+#endif /* HAVE_TZSET */
+
+#ifndef DONT_USE_SYSTEM_LOCALE
 #ifdef HAVE_SETLOCALE
 #ifdef LC_NUMERIC
   setlocale(LC_NUMERIC, "C");
@@ -213,8 +285,10 @@ int dbm_main(int argc, char **argv)
   setlocale(LC_MESSAGES, "");
 #endif
 #endif  
+#endif /* !DONT_USE_SYSTEM_LOCALE */
 
-  init_backend();
+  TRACE((stderr, "Init master...\n"));
+  
   master_file = 0;
 
 #ifdef HAVE_GETENV
@@ -239,8 +313,12 @@ int dbm_main(int argc, char **argv)
     master_file=master_location;
   }
 
+  TRACE((stderr, "Default master at \"%s\"...\n", master_file));
+  
   for(e=1; e<argc; e++)
   {
+    TRACE((stderr, "Parse argument %d:\"%s\"...\n", e, argv[e]));
+  
     if(argv[e][0]=='-')
     {
       for(p=argv[e]+1; *p;)
@@ -344,9 +422,9 @@ int dbm_main(int argc, char **argv)
 
 	    case 'c':
 	      p++;
-#ifdef YYDEBUG
+#if defined(YYDEBUG) || defined(PIKE_DEBUG)
 	      yydebug++;
-#endif /* YYDEBUG */
+#endif /* YYDEBUG || PIKE_DEBUG */
 	      break;
 
 	    case 's':
@@ -358,7 +436,17 @@ int dbm_main(int argc, char **argv)
 	      debug_options|=NO_TAILRECURSION;
 	      p++;
 	      goto more_d_flags;
- 
+
+	    case 'g':
+	      debug_options|=GC_RESET_DMALLOC;
+	      p++;
+	      goto more_d_flags;
+
+	    case 'p':
+	      debug_options|=NO_PEEP_OPTIMIZING;
+	      p++;
+	      goto more_d_flags;
+
 	    case 'T':
 	      debug_options |= ERRORCHECK_MUTEXES;
 	      p++;
@@ -438,7 +526,7 @@ int dbm_main(int argc, char **argv)
     }
   }
 
-#if !defined(HAVE_PTHREAD_MUTEX_ERRORCHECK) && !defined(HAVE_PTHREAD_MUTEX_ERRORCHECK_NP)
+#ifndef PIKE_MUTEX_ERRORCHECK
   if (debug_options & ERRORCHECK_MUTEXES)
     fputs ("Warning: -dT (error checking mutexes) not supported on this system.\n",
 	   stderr);
@@ -449,6 +537,8 @@ int dbm_main(int argc, char **argv)
 #define RLIMIT_NOFILE RLIMIT_OFILE
 #endif
 
+  TRACE((stderr, "Init C stack...\n"));
+  
   Pike_interpreter.stack_top = (char *)&argv;
 
   /* Adjust for anything already pushed on the stack.
@@ -510,7 +600,7 @@ int dbm_main(int argc, char **argv)
       Pike_interpreter.stack_top += lim.rlim_cur;
 #endif /* STACK_DIRECTION < 0 */
 
-#if defined(__linux__) && defined(HAVE_DLOPEN) && defined(HAVE_DLFCN_H)
+#if defined(__linux__) && defined(HAVE_DLOPEN) && defined(HAVE_DLFCN_H) && !defined(PPC)
       {
 	char ** bos_location;
 	void *handle;
@@ -518,13 +608,14 @@ int dbm_main(int argc, char **argv)
 	if((handle=dlopen(0, RTLD_LAZY)))
 	{
 	  bos_location=dlsym(handle,"__pthread_initial_thread_bos");
-	  dlclose(handle);
 
 	  if(bos_location && *bos_location &&
 	     (*bos_location - Pike_interpreter.stack_top) *STACK_DIRECTION < 0)
 	  {
 	    Pike_interpreter.stack_top=*bos_location;
 	  }
+
+	  dlclose(handle);
 	}
       }
 #else
@@ -542,9 +633,14 @@ int dbm_main(int argc, char **argv)
 	}
       }
 #endif /* HAVE_PTHREAD_INITIAL_THREAD_BOS */
-#endif /* __linux__ && HAVE_DLOPEN && HAVE_DLFCN_H */
+#endif /* __linux__ && HAVE_DLOPEN && HAVE_DLFCN_H && !PPC*/
+
+#if STACK_DIRECTION < 0
+      Pike_interpreter.stack_top += 8192 * sizeof(char *);
+#else /* STACK_DIRECTION >= 0 */
+      Pike_interpreter.stack_top -= 8192 * sizeof(char *);
+#endif /* STACK_DIRECTION < 0 */
 
-      Pike_interpreter.stack_top -= STACK_DIRECTION * 8192 * sizeof(char *);
 
 #ifdef STACK_DEBUG
       fprintf(stderr, "1: C-stack: 0x%08p - 0x%08p, direction:%d\n",
@@ -580,35 +676,66 @@ int dbm_main(int argc, char **argv)
 #endif
 #endif
   
+  TRACE((stderr, "Init time...\n"));
+  
   GETTIMEOFDAY(&current_time);
+  
+  TRACE((stderr, "Init interpreter...\n"));
 
   low_th_init();
   
   init_shared_string_table();
   init_interpreter();
   init_types();
-  init_cpp();
   init_lex();
   init_program();
   init_object();
 
+  TRACE((stderr, "Init modules...\n"));
+
   init_modules();
-  master();
-  call_callback(& post_master_callbacks, 0);
-  free_callback_list(& post_master_callbacks);
-  
+
+#ifdef TEST_MULTISET
+  /* A C-level testsuite for the low level stuff in multisets. */
+  test_multiset();
+#endif
+
   if(SETJMP(back))
   {
     if(throw_severity == THROW_EXIT)
     {
       num=throw_value.u.integer;
     }else{
-      call_handle_error();
+      if (throw_value.type == T_OBJECT &&
+	  throw_value.u.object->prog == master_load_error_program &&
+	  !get_master()) {
+	/* Report this specific error in a nice way. Since there's no
+	 * master it'd be reported with a raw error dump otherwise. */
+	struct generic_error_struct *err;
+	*(Pike_sp++) = throw_value;
+	dmalloc_touch_svalue(Pike_sp-1);
+	throw_value.type=T_INT;
+	err = (struct generic_error_struct *)
+	  get_storage (Pike_sp[-1].u.object, generic_error_program);
+	push_string (err->desc);
+	f_werror (1);
+	pop_stack();
+      }
+      else
+	call_handle_error();
       num=10;
     }
   }else{
     back.severity=THROW_EXIT;
 
+    TRACE((stderr, "Init master...\n"));
+
+    master();
+    call_callback(& post_master_callbacks, 0);
+    free_callback_list(& post_master_callbacks);
+
+    TRACE((stderr, "Call master->_main()...\n"));
+
     a=allocate_array_no_init(argc,0);
     for(num=0;num<argc;num++)
     {
@@ -628,12 +755,12 @@ int dbm_main(int argc, char **argv)
   
     apply(master(),"_main",2);
     pop_stack();
-    
-    backend();
     num=0;
   }
   UNSETJMP(back);
 
+  TRACE((stderr, "Exit %d...\n", num));
+  
   pike_do_exit(num);
   return num; /* avoid warning */
 }
@@ -649,12 +776,11 @@ DECLSPEC(noreturn) void pike_do_exit(int num) ATTRIBUTE((noreturn))
   exit_modules();
 
 #ifdef DEBUG_MALLOC
-  {
-    extern void cleanup_memhdrs(void);
-    cleanup_memhdrs();
-  }
+  cleanup_memhdrs();
+  cleanup_debug_malloc();
 #endif
 
+
 #ifdef PROFILING
   {
     int q;
@@ -664,17 +790,39 @@ DECLSPEC(noreturn) void pike_do_exit(int num) ATTRIBUTE((noreturn))
   }
 #endif
 
+#ifdef PIKE_DEBUG
+  /* For profiling */
+  exit_lex();
+#endif
+
+#ifdef INTERNAL_PROFILING
+  fprintf (stderr, "Evaluator callback calls: %lu\n", evaluator_callback_calls);
+#ifdef PIKE_THREADS
+  fprintf (stderr, "Thread yields: %lu\n", thread_yields);
+#endif
+  fprintf (stderr, "Main thread summary:\n");
+  debug_print_rusage (stderr);
+#endif
+
   exit(num);
 }
 
 
 void low_init_main(void)
 {
+  void init_iterators(void);
+
+  init_cpp();
+  init_backend();
+  init_iterators();
   init_pike_searching();
   init_error();
   init_pike_security();
   th_init();
   init_operators();
+
+  init_builtin();
+
   init_builtin_efuns();
   init_signals();
   init_dynamic_load();
@@ -694,35 +842,28 @@ void init_main(void)
 void low_exit_main(void)
 {
 #ifdef DO_PIKE_CLEANUP
-  void cleanup_added_efuns(void);
-  void cleanup_pike_types(void);
-  void cleanup_program(void);
-  void cleanup_compiler(void);
-  void cleanup_backend(void);
-  void free_all_mapping_blocks(void);
-  void free_all_object_blocks(void);
-  void exit_builtin(void);
+  void exit_iterators(void);
 
 #ifdef AUTO_BIGNUM
-  void exit_auto_bignum(void);
   exit_auto_bignum();
 #endif
+
   exit_pike_searching();
   th_cleanup();
   exit_object();
   exit_dynamic_load();
   exit_signals();
   exit_builtin();
-  exit_lex();
   exit_cpp();
   cleanup_interpret();
   cleanup_added_efuns();
+  cleanup_module_support();
   exit_operators();
-  cleanup_pike_types();
+  exit_iterators();
   cleanup_program();
   cleanup_compiler();
   cleanup_error();
-  cleanup_backend();
+  exit_backend();
 
 #ifdef SHARED_NODES
   free(node_hash.table);
@@ -731,10 +872,18 @@ void low_exit_main(void)
   exit_pike_security();
   free_svalue(& throw_value);
   throw_value.type=T_INT;
-  do_gc();
+  {
+    while(1) {
+      int tmp=num_objects;
+      do_gc();
+      if(num_objects >= tmp) break;
+    }
+  }
 
   cleanup_gc();
 
+  cleanup_pike_types();
+
 #if defined(PIKE_DEBUG) && defined(DEBUG_MALLOC)
   if(verbose_debug_exit)
   {
@@ -776,8 +925,10 @@ void low_exit_main(void)
       fprintf(stderr,"Multisets left: %d (%d bytes)\n",num,size);
 
 
+    destruct_objects_to_destruct_cb();
     if(recount)
     {
+
       fprintf(stderr,"Garbage collecting..\n");
       do_gc();
       
@@ -797,7 +948,7 @@ void low_exit_main(void)
     {
       struct program *p;
       for(p=first_program;p;p=p->next)
-	describe_something(p, T_PROGRAM, 0,2,0);
+	describe_something(p, T_PROGRAM, 0,2,0, NULL);
     }
 
 
@@ -808,17 +959,20 @@ void low_exit_main(void)
     {
       struct object *o;
       for(o=first_object;o;o=o->next)
-	describe_something(o, T_OBJECT, 0,2,0);
+	describe_something(o, T_OBJECT, 0,2,0, NULL);
     }
 
-    cleanup_shared_string_table();
+    count_memory_in_pike_types(&num, &size);
+    if (num)
+      fprintf(stderr, "Types left: %d (%d bytes)\n", num, size);
+    describe_all_types();
   }
 #else
 
   zap_all_arrays();
   zap_all_mappings();
 
-  cleanup_shared_string_table();
+  destruct_objects_to_destruct_cb();
 #endif
 
   really_clean_up_interpret();
@@ -827,12 +981,18 @@ void low_exit_main(void)
   free_all_callable_blocks();
   exit_destroy_called_mark_hash();
 
+  cleanup_pike_type_table();
+  cleanup_shared_string_table();
+
   free_dynamic_load();
   first_mapping=0;
   free_all_mapping_blocks();
   first_object=0;
   free_all_object_blocks();
+  first_program=0;
+  free_all_program_blocks();
+#ifdef PIKE_NEW_MULTISETS
+  exit_multiset();
+#endif
 #endif
 }
-
-
