@@ -1,3 +1,10 @@
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: module_support.c,v 1.55 2004/05/18 07:59:13 grubba Exp $
+*/
+
 #include "global.h"
 #include "module_support.h"
 #include "interpret.h"
@@ -9,7 +16,9 @@
 #include "object.h"
 #include "opcodes.h"
 
-RCSID("$Id: module_support.c,v 1.42 2004/03/04 17:03:05 grubba Exp $");
+#define sp Pike_sp
+
+RCSID("$Id: module_support.c,v 1.55 2004/05/18 07:59:13 grubba Exp $");
 
 /* Checks that args_to_check arguments are OK.
  * Returns 1 if everything worked ok, zero otherwise.
@@ -38,7 +47,7 @@ static int va_check_args(struct svalue *s,
 
     if (!((1UL << s[res->argno].type) & res->expected))
     {
-      res->got=s[res->argno].type;
+      res->got = DO_NOT_WARN((unsigned char)s[res->argno].type);
       res->error_type = ERR_BAD_ARG;
       return 0;
     }
@@ -77,7 +86,6 @@ PMOD_EXPORT void check_all_args(const char *fnname, int args, ... )
 {
   va_list arglist;
   struct expect_result tmp;
-  int argno;
 
   va_start(arglist, args);
   va_check_args(sp - args, args, &tmp, arglist);
@@ -94,7 +102,7 @@ PMOD_EXPORT void check_all_args(const char *fnname, int args, ... )
   case ERR_BAD_ARG:
   {
     char buf[1000];
-    int e,d;
+    int e;
     buf[0]=0;
     for(e=0;e<16;e++)
     {
@@ -125,6 +133,7 @@ PMOD_EXPORT void check_all_args(const char *fnname, int args, ... )
  * usage: get_args(sp-args, args, "%i",&an_int)
  * format specifiers:
  *   %i: INT_TYPE
+ *   %l: LONGEST
  *   %s: char *				Only 8bit strings
  *   %S: struct pike_string *		Only 8bit strings
  *   %W: struct pike_string *		Allow wide strings
@@ -142,7 +151,7 @@ PMOD_EXPORT void check_all_args(const char *fnname, int args, ... )
 
 int va_get_args(struct svalue *s,
 		INT32 num_args,
-		char *fmt,
+		const char *fmt,
 		va_list ap)
 {
   int ret=0;
@@ -150,7 +159,7 @@ int va_get_args(struct svalue *s,
   while(*fmt)
   {
     if(*fmt != '%')
-      fatal("Error in format for get_args.\n");
+      Pike_fatal("Error in format for get_args.\n");
 
     if(ret == num_args) return ret;
 
@@ -177,7 +186,7 @@ int va_get_args(struct svalue *s,
 	  DO_NOT_WARN((int)s->u.float_number);
       else 
       {
-        ref_push_string(int_type_string);
+        ref_push_type_value(int_type_string);
         push_svalue( s );
         f_cast( );
 	if(sp[-1].type == T_INT)
@@ -197,7 +206,7 @@ int va_get_args(struct svalue *s,
         *va_arg(ap, INT_TYPE *) = DO_NOT_WARN((INT_TYPE)s->u.float_number);
       else 
       {
-        ref_push_string(int_type_string);
+        ref_push_type_value(int_type_string);
         push_svalue( s );
         f_cast( );
 	if(sp[-1].type == T_INT)
@@ -210,6 +219,17 @@ int va_get_args(struct svalue *s,
         pop_stack();
       }
       break;
+    case 'l':
+      if (s->type == T_INT) {
+	*va_arg(ap, LONGEST *)=s->u.integer;
+	break;
+#ifdef AUTO_BIGNUM
+      } else if (!is_bignum_object_in_svalue(s) ||
+		 !int64_from_bignum(va_arg(ap, LONGEST *), s->u.object)) {
+	return ret;
+#endif /* AUTO_BIGNUM */
+      }
+      break;
     case 's':
       if(s->type != T_STRING) return ret;
       if(s->u.string->size_shift) return ret;
@@ -231,7 +251,7 @@ int va_get_args(struct svalue *s,
     case 'A':
       if(s->type == T_ARRAY)
 	*va_arg(ap, struct array **)=s->u.array;
-      else if (IS_ZERO(s))
+      else if (UNSAFE_IS_ZERO(s))
 	*va_arg(ap, struct array **)=NULL;
       else
 	return ret;
@@ -247,7 +267,7 @@ int va_get_args(struct svalue *s,
 	 *va_arg(ap, FLOAT_TYPE *)=(float)s->u.integer;
       else 
       {
-        ref_push_string(float_type_string);
+        ref_push_type_value(float_type_string);
         push_svalue( s );
         f_cast( );
         *va_arg(ap, FLOAT_TYPE *)=sp[-1].u.float_number;
@@ -270,7 +290,7 @@ int va_get_args(struct svalue *s,
     case 'O':
       if(s->type == T_OBJECT) 
         *va_arg(ap, struct object **)=s->u.object;
-      else if(IS_ZERO(s))
+      else if(UNSAFE_IS_ZERO(s))
         *va_arg(ap, struct object **)=NULL;
       else
         return ret;
@@ -296,7 +316,7 @@ int va_get_args(struct svalue *s,
       break;
       
     default:
-      fatal("Unknown format character in get_args.\n");
+      Pike_fatal("Unknown format character in get_args.\n");
     }
     ret++;
     s++;
@@ -306,8 +326,8 @@ int va_get_args(struct svalue *s,
 }
 
 PMOD_EXPORT int get_args(struct svalue *s,
-	     INT32 num_args,
-	     char *fmt, ...)
+			 INT32 num_args,
+			 const char *fmt, ...)
 {
   va_list ptr;
   int ret;
@@ -317,7 +337,8 @@ PMOD_EXPORT int get_args(struct svalue *s,
   return ret;
 }
 
-PMOD_EXPORT void get_all_args(char *fname, INT32 args, char *format,  ... )
+PMOD_EXPORT void get_all_args(const char *fname, INT32 args,
+			      const char *format,  ... )
 {
   va_list ptr;
   int ret;
@@ -327,7 +348,7 @@ PMOD_EXPORT void get_all_args(char *fname, INT32 args, char *format,  ... )
   if((ptrdiff_t)ret*2 != (ptrdiff_t)strlen(format)) {
     char *expected_type;
     switch(format[ret*2+1]) {
-    case 'd': case 'i': expected_type = "int"; break;
+    case 'd': case 'i': case 'l': expected_type = "int"; break;
     case 'D': case 'I': expected_type = "int|float"; break;
     case 's': case 'S': expected_type = "string (8bit)"; break;
     case 'W': expected_type = "string"; break;
@@ -370,24 +391,24 @@ PMOD_EXPORT void get_all_args(char *fname, INT32 args, char *format,  ... )
  */
    
 static struct mapping *exported_symbols;
-static struct program *function_encapsulation_program;
 
-PMOD_EXPORT void pike_module_export_symbol(char *name,
+PMOD_EXPORT void pike_module_export_symbol(const char *name,
 					   int len,
 					   void *ptr)
 {
   struct pike_string *str=make_shared_binary_string(name,len);
   struct svalue s;
   if(!exported_symbols) exported_symbols=allocate_mapping(10);
-  s.u.refs=(INT32 *)ptr;
+  s.u.ptr=ptr;
   s.type=T_INT;
   s.subtype=4711;
   mapping_string_insert(exported_symbols, str, &s);
+  free_string(str);
 }
 
-PMOD_EXPORT void *pike_module_import_symbol(char *name,
+PMOD_EXPORT void *pike_module_import_symbol(const char *name,
 					    int len,
-					    char *module,
+					    const char *module,
 					    int module_len)
 {
   struct svalue *s;
@@ -395,10 +416,14 @@ PMOD_EXPORT void *pike_module_import_symbol(char *name,
   if(exported_symbols)
   {
     s=low_mapping_string_lookup(exported_symbols, str);
-    if(s && s->type == T_INT && s->subtype == 4711)
+    if(s)
     {
+#ifdef PIKE_DEBUG
+      if (s->type != T_INT || s->subtype != 4711)
+	Pike_fatal("Unexpected value in exported_symbols.\n");
+#endif
       free_string(str);
-      return s->u.refs;
+      return s->u.ptr;
     }
   }
 
@@ -411,13 +436,27 @@ PMOD_EXPORT void *pike_module_import_symbol(char *name,
   {
     s=low_mapping_string_lookup(exported_symbols, str);
 
-    if(s && s->type == T_INT && s->subtype == 4711)
+    if(s)
     {
+#ifdef PIKE_DEBUG
+      if (s->type != T_INT || s->subtype != 4711)
+	Pike_fatal("Unexpected value in exported_symbols.\n");
+#endif
       free_string(str);
-      return s->u.refs;
+      return s->u.ptr;
     }
   }
 
   free_string(str);
   return 0;
 }
+
+#ifdef DO_PIKE_CLEANUP
+void cleanup_module_support (void)
+{
+  if (exported_symbols) {
+    free_mapping (exported_symbols);
+    exported_symbols = NULL;
+  }
+}
+#endif
