@@ -1,6 +1,11 @@
 /*
- * $Id: jvm.c,v 1.46 2003/04/01 13:13:11 grubba Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: jvm.c,v 1.50 2004/01/02 12:11:04 jonasw Exp $
+*/
+
+/*
  * Pike interface to Java Virtual Machine
  *
  * Marcus Comstedt
@@ -10,7 +15,6 @@
  * Includes
  */
 
-//#define JAVA_DEBUG
 #define NO_PIKE_SHORTHAND
 
 #ifdef HAVE_CONFIG_H
@@ -18,7 +22,7 @@
 #endif /* HAVE_CONFIG_H */
 
 #include "global.h"
-RCSID("$Id: jvm.c,v 1.46 2003/04/01 13:13:11 grubba Exp $");
+RCSID("$Id: jvm.c,v 1.50 2004/01/02 12:11:04 jonasw Exp $");
 #include "program.h"
 #include "interpret.h"
 #include "stralloc.h"
@@ -34,6 +38,8 @@ RCSID("$Id: jvm.c,v 1.46 2003/04/01 13:13:11 grubba Exp $");
 
 #ifdef HAVE_JAVA
 
+#include <stdarg.h>
+
 #ifdef HAVE_JNI_H
 #include <jni.h>
 #endif /* HAVE_JNI_H */
@@ -43,9 +49,10 @@ RCSID("$Id: jvm.c,v 1.46 2003/04/01 13:13:11 grubba Exp $");
 #endif /* HAVE_WINBASE_H */
 
 #ifdef _REENTRANT
-#if defined(HAVE_SPARC_CPU) || defined(HAVE_X86_CPU) || defined(HAVE_PPC_CPU)
+#if defined(HAVE_SPARC_CPU) || defined(HAVE_X86_CPU) || defined(HAVE_PPC_CPU) \
+ || defined(HAVE_ALPHA_CPU)
 #define SUPPORT_NATIVE_METHODS
-#endif /* HAVE_SPARC_CPU || HAVE_X86_CPU */
+#endif /* HAVE_SPARC_CPU || HAVE_X86_CPU || HAVE_PPC_CPU || HAVE_ALPHA_CPU */
 #endif /* _REENTRANT */
 
 #ifdef __NT__
@@ -65,7 +72,7 @@ struct jvm_storage {
   JavaVM *jvm;			/* Denotes a Java VM */
   JNIEnv *env;			/* pointer to native method interface */
   JavaVMInitArgs vm_args;       /* JDK 1.2 VM initialization arguments */
-  JavaVMOption vm_options[10];	/* Should be large enough to hold all opts. */
+  JavaVMOption vm_options[4];	/* Should be large enough to hold all opts. */
   struct pike_string *classpath_string;
   jclass class_object, class_class, class_string, class_throwable;
   jclass class_runtimex, class_system;
@@ -134,8 +141,6 @@ struct att_storage {
 #define THIS_ATT ((struct att_storage *)(Pike_fp->current_storage))
 #endif /* _REENTRANT */
 
-#include "module_magic.h"
-
 /*
 
 TODO(?):
@@ -371,10 +376,10 @@ static void f_jobj_cast(INT32 args)
       (*env)->ReleaseStringChars(env, jstr, wstr);
       (*env)->DeleteLocalRef(env, jstr);
     } else
-      push_int(0);
+      Pike_error("cast() to string failed.\n");
     jvm_vacate_env(jo->jvm, env);
   } else
-    push_int(0);
+    Pike_error("cast() to string failed (no JNIEnv).\n");
 }
 
 static void f_jobj_eq(INT32 args)
@@ -636,7 +641,7 @@ static void make_jargs(jvalue *jargs, INT32 args, char *dorelease, char *sig,
 	jargs->j = sv->u.integer;
 	break;
       case 'F':
-	jargs->f = sv->u.integer;
+	jargs->f = DO_NOT_WARN((float)sv->u.integer);
 	break;
       case 'D':
 	jargs->d = sv->u.integer;
@@ -651,16 +656,16 @@ static void make_jargs(jvalue *jargs, INT32 args, char *dorelease, char *sig,
 	jargs->z = sv->u.float_number!=0;
 	break;
       case 'B':
-	jargs->b = sv->u.float_number;
+	jargs->b = DO_NOT_WARN((char)sv->u.float_number);
 	break;
       case 'C':
-	jargs->c = sv->u.float_number;
+	jargs->c = DO_NOT_WARN((unsigned short)sv->u.float_number);
 	break;
       case 'S':
-	jargs->s = sv->u.float_number;
+	jargs->s = DO_NOT_WARN((short)sv->u.float_number);
 	break;
       case 'I':
-	jargs->i = sv->u.float_number;
+	jargs->i = DO_NOT_WARN((long)sv->u.float_number);
 	break;
       case 'J':
 	jargs->j = sv->u.float_number;
@@ -1475,7 +1480,8 @@ struct cpu_context {
 };
 
 static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
-			   void (*dispatch)(), int args)
+			   void (*dispatch)(), int args,
+			   int flt_args, int dbl_args)
 {
   unsigned INT32 *p = ctx->code;
 
@@ -1524,7 +1530,8 @@ struct cpu_context {
 };
 
 static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
-			   void (*dispatch)(), int args)
+			   void (*dispatch)(), int args,
+			   int flt_args, int dbl_args)
 {
   unsigned char *p = ctx->code;
 
@@ -1562,12 +1569,108 @@ static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
 #else
 #ifdef HAVE_PPC_CPU
 
+#ifdef __linux
+
+/* SVR4 ABI */
+
+#define VARARG_NATIVE_DISPATCH
+
+#define NUM_FP_SAVE 8
+#define REG_SAVE_AREA_SIZE (8*NUM_FP_SAVE+4*8+8)
+#define STACK_FRAME_SIZE (8+REG_SAVE_AREA_SIZE+12+4)
+#define VAOFFS0 (8+REG_SAVE_AREA_SIZE)
+#define VAOFFS(x) ((((char*)&((*(va_list*)NULL))[0].x)-(char*)NULL)+VAOFFS0)
+
+struct cpu_context {
+  unsigned INT32 code[32+NUM_FP_SAVE];
+};
+
+static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
+			   void (*dispatch)(), int args,
+			   int flt_args, int dbl_args)
+{
+  unsigned INT32 *p = ctx->code;
+  int i;
+
+  *p++ = 0x7c0802a6;  /* mflr r0         */
+  *p++ = 0x90010004;  /* stw r0,4(r1)    */
+  *p++ = 0x94210000|(0xffff&-STACK_FRAME_SIZE);
+		      /* stwu r1,-STACK_FRAME_SIZE(r1) */
+  if(!statc)
+    *p++ = 0x9081000c;  /* stw r4,12(r1)   */
+  *p++ = 0x90a10010;  /* stw r5,16(r1)   */
+  *p++ = 0x90c10014;  /* stw r6,20(r1)   */
+  *p++ = 0x90e10018;  /* stw r7,24(r1)   */
+  *p++ = 0x9101001c;  /* stw r8,28(r1)   */
+  *p++ = 0x91210020;  /* stw r9,32(r1)   */
+  *p++ = 0x91410024;  /* stw r10,36(r1)  */
+
+  *p++ = 0x40a60000|(4*NUM_FP_SAVE+4);
+		      /* bne+ cr1,.nofpsave */
+  for(i=0; i<NUM_FP_SAVE; i++)
+    *p++ = 0xd8010000|((i+1)<<21)|(8+4*8+8*i);
+		      /* stfd fN,M(r1)   */
+
+		      /* .nofpsave:      */
+  if(statc) {
+    *p++ = 0x7c852378;  /* mr r5,r4        */
+    *p++ = 0x38000002;  /* li r0,2	   */
+  } else {
+    *p++ = 0x38a00000;  /* li r5,0         */
+    *p++ = 0x38000001;  /* li r0,1	   */
+  }
+  
+  *p++ = 0x7c641b78;  /* mr r4,r3        */
+  *p++ = 0x98010000|VAOFFS(gpr);
+		      /* stb r0,gpr      */
+  *p++ = 0x38000000;  /* li r0,0         */
+  *p++ = 0x98010000|VAOFFS(fpr);
+		      /* stb r0,fpr      */
+  *p++ = 0x38010000|(STACK_FRAME_SIZE+8);
+		      /* addi r0,r1,STACK_FRAME_SIZE+8   */
+  *p++ = 0x90010000|VAOFFS(overflow_arg_area);
+		      /* stw r0,overflow_arg_area        */
+  *p++ = 0x38010008;  /* addi r0,r1,8                    */
+  *p++ = 0x90010000|VAOFFS(reg_save_area);
+		      /* stw r0,reg_save_area            */
+
+  *p++ = 0x38c10000|VAOFFS0;
+		      /* addi r6,r1,va_list              */
+
+  *p++ = 0x3c600000|(((unsigned INT32)(void *)data)>>16);
+                      /* lis r3,hi16(data)          */
+  *p++ = 0x60630000|(((unsigned INT32)(void *)data)&0xffff);
+                      /* ori r3,r3,lo16(data)       */
+ 
+  *p++ = 0x3d800000|(((unsigned INT32)(void *)dispatch)>>16);
+                      /* lis r12,hi16(dispatch)     */
+  *p++ = 0x618c0000|(((unsigned INT32)(void *)dispatch)&0xffff);
+                      /* ori r12,r12,lo16(dispatch) */
+
+  *p++ = 0x7d8803a6;  /* mtlr r12        */
+  *p++ = 0x4e800021;  /* blrl            */
+
+  *p++ = 0x80210000;  /* lwz r1,0(r1)    */
+  *p++ = 0x80010004;  /* lwz r0,4(r1)    */
+  *p++ = 0x7c0803a6;  /* mtlr r0         */
+  *p++ = 0x4e800020;  /* blr             */
+
+  return ctx->code;
+}
+
+#else /* __linux */
+
+/* PowerOpen ABI */
+
 struct cpu_context {
   unsigned INT32 code[23];
 };
 
+#define FLT_ARG_OFFS (args+wargs)
+
 static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
-			   void (*dispatch)(), int args)
+			   void (*dispatch)(), int args,
+			   int flt_args, int dbl_args)
 {
   unsigned INT32 *p = ctx->code;
 
@@ -1576,12 +1679,26 @@ static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
   *p++ = 0x9421ffc8;  /* stwu r1,-56(r1) */
   if(!statc)
     *p++ = 0x90810054;  /* stw r4,84(r1)   */
+#ifdef __APPLE__
+  {
+    int i, fp=1;
+    for(i=0; i<6; i++)
+      if(flt_args&(1<<i))
+	*p++ = 0xd0010000|((fp++)<<21)|(4*i+88);  /* stfs fN,X(r1)   */	
+      else if(i<5 && dbl_args&(1<<i)) {
+	*p++ = 0xd8010000|((fp++)<<21)|(4*i+88);  /* stfd fN,X(r1)   */	
+	i++;
+      } else
+	*p++ = 0x90010000|((i+5)<<21)|(4*i+88);  /* stw rN,X(r1)   */
+  }
+#else
   *p++ = 0x90a10058;  /* stw r5,88(r1)   */
   *p++ = 0x90c1005c;  /* stw r6,92(r1)   */
   *p++ = 0x90e10060;  /* stw r7,96(r1)   */
   *p++ = 0x91010064;  /* stw r8,100(r1)  */
   *p++ = 0x91210068;  /* stw r9,104(r1)  */
   *p++ = 0x9141006c;  /* stw r10,108(r1) */
+#endif
 
   if(statc) {
     *p++ = 0x7c852378;  /* mr r5,r4        */
@@ -1614,8 +1731,103 @@ static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
   return ctx->code;
 }
 
+#endif /* __linux */
+
+#else
+#ifdef HAVE_ALPHA_CPU
+
+/* NB: Assumes that pointers are 64bit! */
+
+#define GET_NATIVE_ARG(ty) (((args)=((void**)(args))+1),*(ty *)(((void**)(args))-1))
+
+struct cpu_context {
+  void *code[10];
+};
+
+static void *low_make_stub(struct cpu_context *ctx, void *data, int statc,
+			   void (*dispatch)(), int args,
+			   int flt_args, int dbl_args)
+{
+  unsigned INT32 *p = (unsigned INT32 *)ctx->code;
+
+  /* lda sp,-48(sp) */
+  *p++ = 0x23deffd0;
+  /* stq ra,0(sp) */
+  *p++ = 0xb75e0000;
+  if(!statc)
+    /* stq a1,8(sp) */
+    *p++ = 0xb63e0008;
+  if(dbl_args & (1<<0))
+    /* stt $f18,16(sp) */
+    *p++ = 0x9e5e0010;
+  else if(flt_args & (1<<0))
+    /* sts $f18,16(sp) */
+    *p++ = 0x9a5e0010;
+  else
+    /* stq a2,16(sp) */
+    *p++ = 0xb65e0010;
+  if(dbl_args & (1<<1))
+    /* stt $f19,24(sp) */
+    *p++ = 0x9e7e0018;
+  else if(flt_args & (1<<1))
+    /* sts $f19,24(sp) */
+    *p++ = 0x9a7e0018;
+  else
+    /* stq a3,24(sp) */
+    *p++ = 0xb67e0018;
+  if(dbl_args & (1<<2))
+    /* stt $f20,32(sp) */
+    *p++ = 0x9e9e0020;
+  else if(flt_args & (1<<2))
+    /* sts $f20,32(sp) */
+    *p++ = 0x9a9e0020;
+  else
+    /* stq a4,32(sp) */
+    *p++ = 0xb69e0020;
+  if(dbl_args & (1<<3))
+    /* stt $f21,40(sp) */
+    *p++ = 0x9ebe0028;
+  else if(flt_args & (1<<3))
+    /* sts $f21,40(sp) */
+    *p++ = 0x9abe0028;
+  else
+    /* stq a5,40(sp) */
+    *p++ = 0xb6be0028;
+  if(statc) { 
+    /* mov a1,a2 */
+    *p++ = 0x46310412;
+    /* lda a3,16(sp) */
+    *p++ = 0x227e0010;
+  } else { 
+    /* clr a2 */
+    *p++ = 0x47ff0412;
+    /* lda a3,8(sp) */
+    *p++ = 0x227e0008;
+  } 
+  /* mov a0,a1 */
+  *p++ = 0x46100411;
+  /* ldq a0,64(t12) */
+  *p++ = 0xa61b0040;
+  /* ldq t12,72(t12) */
+  *p++ = 0xa77b0048;
+  /* jsr ra,(t12) */
+  *p++ = 0x6b5b4000;
+  /* ldq ra,0(sp) */
+  *p++ = 0xa75e0000;
+  /* lda sp,48(sp) */
+  *p++ = 0x23de0030;
+  /* ret zero,(ra) */
+  *p++ = 0x6bfa8001;
+
+  ctx->code[8] = data;
+  ctx->code[9] = dispatch;
+
+  return ctx->code;
+}
+
 #else
 #error How did you get here?  It should never happen.
+#endif /* HAVE_ALPHA_CPU */
 #endif /* HAVE_PPC_CPU */
 #endif /* HAVE_X86_CPU */
 #endif /* HAVE_SPARC_CPU */
@@ -1662,7 +1874,27 @@ static void make_java_exception(struct object *jvm, JNIEnv *env,
   }
 }
 
+#ifdef VARARG_NATIVE_DISPATCH
+
+#define ARGS_TYPE va_list*
+#define GET_NATIVE_ARG(ty) va_arg(*args,ty)
+#define NATIVE_ARG_JFLOAT_TYPE jdouble
+
+#else
+
+#ifndef ARGS_TYPE
+#define ARGS_TYPE void*
+#endif
+
+#ifndef GET_NATIVE_ARG
 #define GET_NATIVE_ARG(ty) (((args)=((ty *)(args))+1),(((ty *)(args))[-1]))
+#endif
+
+#ifndef NATIVE_ARG_JFLOAT_TYPE
+#define NATIVE_ARG_JFLOAT_TYPE jfloat
+#endif
+
+#endif
 
 static void do_native_dispatch(struct native_method_context *ctx,
 			       JNIEnv *env, jclass cls, void *args_,
@@ -1683,7 +1915,7 @@ static void do_native_dispatch(struct native_method_context *ctx,
 
   {
     int nargs = 0;
-    void *args = args_;
+    ARGS_TYPE args = args_;
 
     if(!cls) {
       push_java_anyobj(GET_NATIVE_ARG(jobject), ctx->nat->jvm, env);
@@ -1711,7 +1943,7 @@ static void do_native_dispatch(struct native_method_context *ctx,
 	break;
       
       case 'F':
-	push_float(GET_NATIVE_ARG(jfloat));
+	push_float(GET_NATIVE_ARG(NATIVE_ARG_JFLOAT_TYPE));
 	break;
       
       case 'D':
@@ -1746,7 +1978,7 @@ static void do_native_dispatch(struct native_method_context *ctx,
          released automatically when we return to java */
       make_jargs(rc, -1, NULL, p, ctx->nat->jvm, env);
       if((*p == 'L' || *p == '[') && rc->l != NULL)
-        rc->l = (*env)->NewLocalRef(env, rc->l);
+	rc->l = (*env)->NewLocalRef(env, rc->l);
     }
   }
 
@@ -1768,11 +2000,6 @@ static void native_dispatch(struct native_method_context *ctx,
     } else {
       /* Nope, let's get it... */
       mt_lock_interpreter();
-      while (threads_disabled) {
-	THREADS_FPRINTF(1, (stderr,
-			    "jvm/native_dispatch(): Threads disabled\n"));
-	co_wait_interpreter(&threads_disabled_change);
-      }
       SWAP_IN_THREAD(state);
 
       do_native_dispatch(ctx, env, cls, args, rc);
@@ -1784,11 +2011,6 @@ static void native_dispatch(struct native_method_context *ctx,
   } else {
     /* Not a pike thread.  Create a temporary thread_id... */
     mt_lock_interpreter();
-    while (threads_disabled) {
-      THREADS_FPRINTF(1, (stderr,
-			  "jvm/native_dispatch(): Threads disabled\n"));
-      co_wait_interpreter(&threads_disabled_change);
-    }
     init_interpreter();
     Pike_interpreter.stack_top=((char *)&state)+ (thread_stack_size-16384) * STACK_DIRECTION;
     Pike_interpreter.recoveries = NULL;
@@ -1807,8 +2029,8 @@ static void native_dispatch(struct native_method_context *ctx,
       co_signal(& OBJ2THREAD(Pike_interpreter.thread_id)->status_change);
       thread_table_delete(Pike_interpreter.thread_id);
       free_object(Pike_interpreter.thread_id);
+      Pike_interpreter.thread_id=NULL;
     }
-    Pike_interpreter.thread_id=NULL;
     cleanup_interpret();
     num_threads--;
     mt_unlock_interpreter();
@@ -1894,8 +2116,8 @@ static void JNICALL native_dispatch_v(struct native_method_context *ctx,
   native_dispatch(ctx, env, obj, args, &v);
 }
 
-static void make_stub(struct cpu_context *ctx, void *data, int statc, int rt,
-		      int args)
+static void *make_stub(struct cpu_context *ctx, void *data, int statc, int rt,
+		       int args, int flt_args, int dbl_args)
 {
   void *disp = native_dispatch_v;
 
@@ -1914,16 +2136,20 @@ static void make_stub(struct cpu_context *ctx, void *data, int statc, int rt,
     disp = native_dispatch_v;
   }
 
-  low_make_stub(ctx, data, statc, disp, args);
+  return low_make_stub(ctx, data, statc, disp, args, flt_args, dbl_args);
 }
 
+#ifndef FLT_ARG_OFFS
+#define FLT_ARG_OFFS args
+#endif
+
 static void build_native_entry(JNIEnv *env, jclass cls,
 			       struct native_method_context *con,
 			       JNINativeMethod *jnm,
 			       struct pike_string *name,
 			       struct pike_string *sig)
 {
-  int statc, args=0;
+  int statc, args=0, wargs=0, flt_args=0, dbl_args=0;
   char *p = sig->str;
 
   con->name = name;
@@ -1945,13 +2171,19 @@ static void build_native_entry(JNIEnv *env, jclass cls,
 
   jnm->name = name->str;
   jnm->signature = sig->str;
-  jnm->fnPtr = (void*)&con->cpu;
   while(*p && *p != ')')
     switch(*p++) {
     case '(':
       break;
-    case 'J': case 'D':
-      args += 2;
+    case 'D':
+      dbl_args |= 1<<FLT_ARG_OFFS;
+    case 'J':
+      args ++;
+      wargs ++;
+      break;
+    case 'F':
+      flt_args |= 1<<FLT_ARG_OFFS;
+      args ++;
       break;
     case '[':
       if(!*p)
@@ -1963,7 +2195,8 @@ static void build_native_entry(JNIEnv *env, jclass cls,
       args++;
     }
   if(*p) p++;
-  make_stub(&con->cpu, con, statc, *p, args+2);
+  jnm->fnPtr = make_stub(&con->cpu, con, statc, *p, args+wargs+2,
+			 flt_args, dbl_args);
 }
 
 static void init_natives_struct(struct object *o)
@@ -2844,7 +3077,6 @@ static void f_create(INT32 args)
   char *classpath = NULL;
   jclass cls;
 
-  /* FIXME: This doesn't loop... */
   while(j->jvm) {
     JavaVM *jvm = j->jvm;
     JNIEnv *env;
@@ -2903,22 +3135,6 @@ static void f_create(INT32 args)
 #endif
 #endif
 
-#ifdef JAVA_DEBUG
-  j->vm_args.options[j->vm_args.nOptions].optionString =
-    "-Xdebug";
-  j->vm_args.options[j->vm_args.nOptions].extraInfo = NULL;
-  j->vm_args.nOptions++;
-
-  j->vm_args.options[j->vm_args.nOptions].optionString =
-#ifdef __NT__
-    "-Xrunjdwp:transport=dt_shmem,address=roxenjdb,server=y,suspend=n";
-#else
-    "-Xrunjdwp:transport=dt_socket,address=8000,server=y,suspend=n";
-#endif
-  j->vm_args.options[j->vm_args.nOptions].extraInfo = NULL;
-  j->vm_args.nOptions++;
-#endif
-
   /* load and initialize a Java VM, return a JNI interface 
    * pointer in env */
   if(JNI_CreateJavaVM(&j->jvm, (void**)&j->env, &j->vm_args))
@@ -3297,11 +3513,9 @@ static void f_new_double_array(INT32 args)
     push_int(0);
 }
 
-#else
-#include "module_magic.h"
 #endif /* HAVE_JAVA */
 
-void pike_module_init(void)
+PIKE_MODULE_INIT
 {
 #ifdef HAVE_JAVA
   struct svalue prog;
@@ -3488,7 +3702,7 @@ void pike_module_init(void)
 #endif /* HAVE_JAVA */
 }
 
-void pike_module_exit(void)
+PIKE_MODULE_EXIT
 {
 #ifdef HAVE_JAVA
   if(jarray_program) {
@@ -3544,5 +3758,3 @@ void pike_module_exit(void)
 #endif /* __NT __ */
 #endif /* HAVE_JAVA */
 }
-
-
