@@ -1,6 +1,11 @@
 /*
- * $Id: acconfig.h,v 1.6 2002/03/18 13:23:27 grubba Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: acconfig.h,v 1.11 2003/03/28 21:33:31 marcus Exp $
+*/
+
+/*
  * Config-file for the Pike mySQL-module.
  *
  * Henrik Grubbstrm 1997-01-30
