@@ -1,6 +1,11 @@
 /*
- * $Id: oracle.c,v 1.60 2002/10/03 15:55:52 mast Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: oracle.c,v 1.75 2002/11/28 23:45:49 marcus Exp $
+*/
+
+/*
  * Pike interface to Oracle databases.
  *
  * original design by Marcus Comstedt
@@ -34,10 +39,6 @@
 #include "pike_macros.h"
 #include "version.h"
 
-#if (PIKE_MAJOR_VERSION - 0 > 7) || (PIKE_MAJOR_VERSION - 0 == 7 && PIKE_MINOR_VERSION - 0 >= 1)
-/* must be included last */
-#include "module_magic.h"
-#endif
 
 #ifdef HAVE_ORACLE
 
@@ -53,15 +54,35 @@
 
 #include <math.h>
 
-RCSID("$Id: oracle.c,v 1.60 2002/10/03 15:55:52 mast Exp $");
+RCSID("$Id: oracle.c,v 1.75 2002/11/28 23:45:49 marcus Exp $");
 
 
+/* User-changable defines: */
+
 #define BLOB_FETCH_CHUNK 16384
 
 /* #define ORACLE_DEBUG */
+
+/* Undefining this if you don't want pike
+ * muexes protecting everything
+ */
 #define ORACLE_USE_THREADS
+
+/* Define this to zero if you don't want
+ * oracle internal mutexes
+ */
+#define ORACLE_INIT_FLAGS OCI_THREADED
+
+/* This define puts a mutex around the connect calls */
 #define SERIALIZE_CONNECT
 
+/* Define This if you want one environment for each
+ * database connection. (This crashes for me - Hubbe)
+ */
+/* #define LOCAL_ENV */
+
+/* #define REUSE_DEFINES */
+
 /*
  * This define crippes the Pike module to use static define buffers.
  * It may be required to work around bugs in some versions of the
@@ -75,6 +96,9 @@ RCSID("$Id: oracle.c,v 1.60 2002/10/03 15:55:52 mast Exp $");
 
 /* #define STATIC_BUFFERS 8000 */
 
+/* End user-changable defines */
+
+
 #ifndef ORACLE_USE_THREADS
 
 #undef THREADS_ALLOW
@@ -112,7 +136,7 @@ DEFINE_MUTEX(oracle_serialization_mutex);
 }while(0)
 
 #define UNLOCK(X) do { \
-   fprintf(stderr,"unlocking " #X "      from %s:%d\n",__FUNCTION__,__LINE__);\
+   fprintf(stderr,"unocking " #X "      from %s:%d\n",__FUNCTION__,__LINE__); \
    mt_unlock( & (X) ); \
 }while(0)
 
@@ -183,8 +207,11 @@ DEFINE_MUTEX(oracle_serialization_mutex);
 #endif
 
 
+#ifdef PARENT_INFO
+#define PARENTOF(X) PARENT_INFO(X)->parent
+#else
 #define PARENTOF(X) (X)->parent
-
+#endif
 
 /* This define only exists in Pike 7.1.x, if it isn't defined
  * we have to provide this function ourselves -Hubbe
@@ -225,20 +252,20 @@ void *parent_storage(int depth)
     if (o->refs == 0x55555555) {
       fprintf(stderr, "The object %p has been zapped!\n", o);
       describe(p);
-      fatal("Object zapping detected.\n");
+      Pike_fatal("Object zapping detected.\n");
     }
     if (p->refs == 0x55555555) {
       fprintf(stderr, "The program %p has been zapped!\n", p);
       describe(p);
       fprintf(stderr, "Which taken from the object %p\n", o);
       describe(o);
-      fatal("Looks like the program %p has been zapped!\n", p);
+      Pike_fatal("Looks like the program %p has been zapped!\n", p);
     }
 #endif /* DEBUG_MALLOC */
     
 #ifdef PIKE_DEBUG
     if(i < 0 || i > p->num_identifier_references)
-      fatal("Identifier out of range!\n");
+      Pike_fatal("Identifier out of range!\n");
 #endif
     
     inherit=INHERIT_FROM_INT(p, i);
@@ -251,7 +278,7 @@ void *parent_storage(int depth)
       describe(p);
       fprintf(stderr, "Which was in turn taken from the object %p\n", o);
       describe(o);
-      fatal("Looks like the program %p has been zapped!\n", p);
+      Pike_fatal("Looks like the program %p has been zapped!\n", p);
     }
 #endif /* DEBUG_MALLOC */
     
@@ -263,17 +290,29 @@ void *parent_storage(int depth)
 }
 #endif
 
+
+
+#define ocimalloc NULL
+#define ocirealloc NULL
+#define ocifree NULL
+
+
 #ifdef PIKE_DEBUG
-void *check_storage(void *storage, unsigned long magic, char *prog)
+void *low_check_storage(void *storage, unsigned long magic, char *prog)
 {
   if( storage && magic != *((unsigned long *)storage))
   {
     fprintf(stderr, "Wrong magic number! expected a %s\n",prog);
     fprintf(stderr, "Expected %lx, got %lx\n",magic,*((unsigned long *)storage));
-    fatal("Wrong program, expected %s!\n",prog);
+    Pike_fatal("Wrong program, expected %s!\n",prog);
   }
   return storage;
 }
+#ifdef DEBUG_MALLOC 
+#define check_storage(X,Y,Z) (debug_malloc_touch(THISOBJ),low_check_storage((X),(Y),(Z)))
+#else
+#define check_storage(X,Y,Z) (X)
+#endif
 #else
 #define check_storage(X,Y,Z) (X)
 #endif
@@ -291,14 +330,14 @@ void *check_storage(void *storage, unsigned long magic, char *prog)
 
 static struct program *oracle_program = NULL;
 static struct program *compile_query_program = NULL;
-static struct program *big_query_program = NULL;
+static struct program *big_typed_query_program = NULL;
 static struct program *dbresultinfo_program = NULL;
 static struct program *Date_program = NULL;
 static struct program *NULL_program = NULL;
 
 static int oracle_identifier;
 static int compile_query_identifier;
-static int big_query_identifier;
+static int big_typed_query_identifier;
 static int dbresultinfo_identifier;
 static int Date_identifier;
 
@@ -308,19 +347,25 @@ static struct object *nullint_object;
 static struct object *nulldate_object;
 
 static OCIEnv *oracle_environment=0;
-
-static OCIEnv *get_oracle_environment(void)
+static OCIEnv *low_get_oracle_environment(void)
 {
   sword rc;
   if(!oracle_environment)
   {
     rc=OCIEnvInit(&oracle_environment, OCI_DEFAULT, 0, 0);
     if(rc != OCI_SUCCESS)
-      Pike_error("Failed to initialize oracle environment.\n");
+      Pike_error("Failed to initialize oracle environment, err=%d.\n",rc);
   }
   return oracle_environment;
 }
 
+
+#ifdef DEBUG_MALLOC
+#define get_oracle_environment() dmalloc_touch(OCIEnv*,low_get_oracle_environment())
+#else
+#define get_oracle_environment() low_get_oracle_environment()
+#endif
+
 struct inout
 {
   sb2 indicator;
@@ -360,6 +405,7 @@ struct dbcon
 #ifdef PIKE_DEBUG
   unsigned long magic;
 #endif
+  OCIEnv *env;
   OCIError *error_handle;
   OCISvcCtx *context;
 
@@ -381,6 +427,9 @@ static void init_dbcon_struct(struct object *o)
   THIS_DBCON->context=0;
   THIS_DBCON->timeout_limit = 30; /* default value */
   THIS_DBCON->resultobject_busy = 0;
+#ifdef LOCAL_ENV
+  THIS_DBCON->env=0;
+#endif
   mt_init( & THIS_DBCON->lock );
 }
 
@@ -389,11 +438,26 @@ static void exit_dbcon_struct(struct object *o)
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"%s\n",__FUNCTION__);
 #endif
+  debug_malloc_touch(THIS_DBCON->context);
   OCILogoff(THIS_DBCON->context, THIS_DBCON->error_handle);
+
+  debug_malloc_touch(THIS_DBCON->error_handle);
   OCIHandleFree(THIS_DBCON->error_handle, OCI_HTYPE_ERROR);
+
+#ifdef LOCAL_ENV
+  debug_malloc_touch(THIS_DBCON->env);
+  OCIHandleFree(THIS_DBCON->env, OCI_HTYPE_ENV);
+#endif
+
   mt_destroy( & THIS_DBCON->lock );
 }
 
+#ifdef LOCAL_ENV
+#define ENVOF(X) (X)->env
+#else
+#define ENVOF(X) get_oracle_environment()
+#endif
+
 /****** query ******/
 
 struct dbquery
@@ -429,6 +493,7 @@ void exit_dbquery_struct(struct object *o)
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"%s\n",__FUNCTION__);
 #endif
+  debug_malloc_touch(THIS_QUERY->statement);
   OCIHandleFree(THIS_QUERY->statement, OCI_HTYPE_STMT);
   mt_destroy(& THIS_QUERY->lock);
 }
@@ -597,7 +662,7 @@ static void ora_error_handler(OCIError *err, sword rc, char *func)
 }
 
 
-static OCIError *global_error_handle=0;;
+static OCIError *global_error_handle=0;
 
 OCIError *get_global_error_handle(void)
 {
@@ -666,6 +731,15 @@ static sb4 output_callback(struct inout *inout,
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"%s(%d)\n",__FUNCTION__,inout->ftype);
 #endif
+  debug_malloc_touch(bufpp);
+  debug_malloc_touch(*bufpp);
+  debug_malloc_touch(alenpp);
+  debug_malloc_touch(*alenpp);
+  debug_malloc_touch(indpp);
+  debug_malloc_touch(*indpp);
+  debug_malloc_touch(rcodepp);
+  debug_malloc_touch(*rcodepp);
+
   inout->has_output=1;
   *indpp = (dvoid *) &inout->indicator;
   *rcodepp=&inout->rcode;
@@ -925,12 +999,12 @@ static void f_fetch_fields(INT32 args)
 	  type_name = "blob";
 	else
 	  type_name="clob";
-	if (rc = OCIDescriptorAlloc(
+	if ((rc = OCIDescriptorAlloc(
 				    (dvoid *) get_oracle_environment(),
 				    (dvoid **) &info->data.lob, 
 				    (ub4)OCI_DTYPE_LOB, 
 				    (size_t) 0, 
-				    (dvoid **) 0))
+				    (dvoid **) 0)))
 	  {
 #ifdef ORACLE_DEBUG
 	    fprintf(stderr,"OCIDescriptorAlloc failed!\n");
@@ -989,8 +1063,8 @@ static void f_fetch_fields(INT32 args)
 			dbcon->error_handle,
 			i+1,
 			info->data.lob ? 
-			& info->data.lob :
-			& info->data.u,
+			(void *)(&info->data.lob):
+			(void *)(&info->data.u),
 #ifdef STATIC_BUFFERS
 			data_size<0? STATIC_BUFFERS :data_size,
 #else
@@ -1025,7 +1099,9 @@ static void f_fetch_fields(INT32 args)
       if(!IS_SUCCESS(rc))
 	ora_error_handler(dbcon->error_handle, rc, "OCIDefineDynamic");
 #endif
-
+      debug_malloc_touch(dbcon->error_handle);
+      debug_malloc_touch(dbquery->statement);
+      debug_malloc_touch(info->define_handle);
     }
     f_aggregate(dbquery->cols);
     add_ref( dbquery->field_info=Pike_sp[-1].u.array );
@@ -1124,7 +1200,7 @@ static void push_inout_value(struct inout *inout,
 	errfunc = "OCILobGetLength";
       } else {
 	amtp = loblen;
-	if(bufp = malloc(loblen)) {
+	if((bufp = malloc(loblen))) {
 	  if((ret = OCILobRead(dbcon->context,
 			       dbcon->error_handle,
 			       inout->lob, 
@@ -1152,7 +1228,7 @@ static void push_inout_value(struct inout *inout,
       fprintf(stderr,"LOB length: %d\n",loblen);
 #endif
       if(ret == OCI_SUCCESS)
-	push_string(make_shared_binary_string(loblen ? bufp : "",loblen));
+	push_string(make_shared_binary_string(bufp, loblen));
       else
 	ora_error_handler(dbcon->error_handle, ret, errfunc);
       
@@ -1201,7 +1277,7 @@ static void push_inout_value(struct inout *inout,
 
       if(IS_SUCCESS(ret))
       {
-	push_int(integer);
+	push_int64(integer);
       }else{
 #ifdef AUTO_BIGNUM
 	unsigned char buffer[80];
@@ -1236,7 +1312,8 @@ static void push_inout_value(struct inout *inout,
       break;
       
     case SQLT_FLT:
-      push_float(inout->u.f); /* We might need to push a Matrix here */
+      /* We might need to push a Matrix here */
+      push_float(DO_NOT_WARN((FLOAT_TYPE)inout->u.f));
       break;
       
     default:
@@ -1279,7 +1356,7 @@ static void f_fetch_row(INT32 args)
 
   pop_n_elems(args);
 
-  if(dbquery->cols==-2)
+  if(!dbquery->field_info)
   {
     f_fetch_fields(0);
     pop_stack();
@@ -1344,13 +1421,36 @@ static void f_oracle_create(INT32 args)
   else
     passwd = NULL;
 
+
+#ifdef LOCAL_ENV
+  if(!dbcon->env)
+  {
+#ifdef ORACLE_DEBUG
+    fprintf(stderr,"%s: creating new environment\n",__FUNCTION__);
+#endif
+    LOCK(dbcon->lock);
+    rc=OCIEnvInit(&dbcon->env, OCI_DEFAULT,0,0);
+    UNLOCK(dbcon->lock);
+    if(rc != OCI_SUCCESS) 
+      Pike_error("Failed to initialize oracle environment, err=%d.\n",rc);
+#ifdef ORACLE_DEBUG
+  } else {
+    fprintf(stderr,"%s: environment exists\n",__FUNCTION__);
+#endif
+  }
+#endif
+
+  debug_malloc_touch(ENVOF(dbcon));
+
   THREADS_ALLOW();
 
   LOCK(dbcon->lock);
+#ifdef SERIALIZE_CONNECT
   LOCK(oracle_serialization_mutex);
+#endif
 
   do  {
-    rc=OCIHandleAlloc(get_oracle_environment(),
+    rc=OCIHandleAlloc(ENVOF(dbcon),
 		      (void **)& dbcon->error_handle,
 		      OCI_HTYPE_ERROR,
 		      0,
@@ -1362,17 +1462,18 @@ static void f_oracle_create(INT32 args)
 #endif
 
 #if 0
-    if(OCIHandleAlloc(get_oracle_environment(),&THIS_DBCON->srvhp,
+    if(OCIHandleAlloc(ENVOF(dbcon),
+		      &THIS_DBCON->srvhp,
 		      OCI_HTYPE_SERVER, 0,0)!=OCI_SUCCESS)
       Pike_error("Failed to allocate server handle.\n");
     
-    if(OCIHandleAlloc(get_oracle_environment(),&THIS_DBCON->srchp,
+    if(OCIHandleAlloc(ENVOF(dbcon),
+		      &THIS_DBCON->srchp,
 		      OCI_HTYPE_SVCCTX, 0,0)!=OCI_SUCCESS)
       Pike_error("Failed to allocate service context.\n");
 #endif
 
-
-    rc=OCILogon(get_oracle_environment(),
+    rc=OCILogon(ENVOF(dbcon),
 		dbcon->error_handle,
 		&dbcon->context,
 		uid->str, uid->len, 
@@ -1381,7 +1482,9 @@ static void f_oracle_create(INT32 args)
   
   }while(0);
 
+#ifdef SERIALIZE_CONNECT
   UNLOCK(oracle_serialization_mutex);
+#endif
   UNLOCK(dbcon->lock);
 
   THREADS_DISALLOW();
@@ -1390,6 +1493,9 @@ static void f_oracle_create(INT32 args)
   if(rc != OCI_SUCCESS)
     ora_error_handler(dbcon->error_handle, rc, 0);
 
+  debug_malloc_touch(dbcon->error_handle);
+  debug_malloc_touch(dbcon->context);
+
   pop_n_elems(args);
 
 #ifdef ORACLE_DEBUG
@@ -1436,8 +1542,8 @@ static void f_compile_query_create(INT32 args)
 
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"f_compile_query_create: dbquery: %p\n",dbquery);
-  fprintf(stderr,"            dbcon: %p\n",dbcon);
-  fprintf(stderr,"     error_handle: %p\n",dbcon->error_handle);
+  fprintf(stderr,"         dbcon:   %p\n",dbcon);
+  fprintf(stderr,"  error_handle: %p\n",dbcon->error_handle);
   fprintf(stderr,"resultobject_busy: %d\n", dbcon->resultobject_busy);
 #endif
 
@@ -1448,7 +1554,11 @@ static void f_compile_query_create(INT32 args)
     for(t = 0; t < timeout_limit && dbcon->resultobject_busy; ++t)
     {
       THREADS_ALLOW();
+#ifdef __NT__
+      Sleep(1000);
+#else
       sleep(1);
+#endif
       THREADS_DISALLOW();
     }
     if (dbcon->resultobject_busy)
@@ -1459,7 +1569,7 @@ static void f_compile_query_create(INT32 args)
   THREADS_ALLOW();
   LOCK(dbcon->lock);
   
-  rc=OCIHandleAlloc(get_oracle_environment(),
+  rc=OCIHandleAlloc(ENVOF(dbcon),
 		    (void **)&dbquery->statement,
 		    OCI_HTYPE_STMT,
 		    0,0);
@@ -1499,12 +1609,8 @@ static void f_compile_query_create(INT32 args)
   }
   
   THREADS_DISALLOW();
-#ifdef ORACLE_DEBUG
-  fprintf(stderr, "    Unlock; rc = %d, OCI_SUCCESS = %d\n", rc, OCI_SUCCESS);
-#endif
   UNLOCK(dbcon->lock);
 
-
   if(rc != OCI_SUCCESS)
     ora_error_handler(dbcon->error_handle, rc, 0);
 
@@ -1602,7 +1708,7 @@ static void free_bind_block(struct bind_block *bind)
  * FIXME: This function should probably lock the statement
  * handle until it is freed...
  */
-static void f_big_query_create(INT32 args)
+static void f_big_typed_query_create(INT32 args)
 {
   sword rc;
   struct mapping *bnds=0;
@@ -1623,7 +1729,7 @@ static void f_big_query_create(INT32 args)
   fprintf(stderr,"%s\n",__FUNCTION__);
 #endif
 
-  check_all_args("Oracle.oracle.compile_query->big_query", args,
+  check_all_args("Oracle.oracle.compile_query->big_typed_query", args,
 		 BIT_VOID | BIT_MAPPING | BIT_INT, 
 		 BIT_VOID | BIT_INT,
 		 BIT_VOID | BIT_OBJECT,
@@ -1634,7 +1740,7 @@ static void f_big_query_create(INT32 args)
   {
       CHECK_INTERPRETER_LOCK();
       if(d_flag>1 && thread_for_id(th_self()) != Pike_thread_id)
-        fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
+        Pike_fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
   }
 #endif
 
@@ -1668,8 +1774,8 @@ static void f_big_query_create(INT32 args)
   if(new_parent &&
      PARENTOF(PARENTOF(THISOBJ)) != new_parent)
   {
-    if(new_parent->prog != PARENTOF(PARENTOF(THISOBJ))->parent->prog)
-      Pike_error("Bad argument 3 to big_query.\n");
+    if(new_parent->prog != PARENTOF(PARENTOF(PARENTOF(THISOBJ)))->prog)
+      Pike_error("Bad argument 3 to big_typed_query.\n");
 
     /* We might need to check that there are no locks held here
      * but I don't beleive that could happen, so just go with it...
@@ -1683,7 +1789,7 @@ static void f_big_query_create(INT32 args)
   {
       CHECK_INTERPRETER_LOCK();
       if(d_flag>1 && thread_for_id(th_self()) != Pike_thread_id)
-        fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
+        Pike_fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
   }
 #endif
 
@@ -1696,12 +1802,11 @@ static void f_big_query_create(INT32 args)
     MAPPING_LOOP(bnds)
       {
 	struct svalue *value=&k->val;
-	OCIBind *bindp;
 	sword rc = 0;
 	void *addr;
 	sword len, fty;
 	int mode=OCI_DATA_AT_EXEC;
-	long rlen=0x7fffffff;
+	long rlen=4000;
 	bind.bindnum++;
 	
 	assign_svalue_no_free(& bind.bind[bind.bindnum].ind, & k->ind);
@@ -1729,7 +1834,7 @@ static void f_big_query_create(INT32 args)
 	      goto retry;
 	    }
 	    Pike_error("Bad value type in argument 2 to "
-		  "Oracle.oracle->big_query()\n");
+		       "Oracle.oracle->big_typed_query()\n");
 	    break;
 
 	  case T_STRING:
@@ -1763,18 +1868,35 @@ static void f_big_query_create(INT32 args)
 	    break;
 	    
 	  case T_MULTISET:
-	    if(value->u.multiset->ind->size == 1 &&
-	       ITEM(value->u.multiset->ind)[0].type == T_STRING)
-	    {
-	      addr = (ub1 *)ITEM(value->u.multiset->ind)[0].u.string->str;
-	      len = ITEM(value->u.multiset->ind)[0].u.string->len;
-	      fty = SQLT_LBI;
-	      break;
+	    if(multiset_sizeof(value->u.multiset) == 1) {
+	      struct pike_string *s;
+#ifdef PIKE_NEW_MULTISETS
+	      {
+		struct svalue tmp;
+		if (use_multiset_index (value->u.multiset,
+					multiset_first (value->u.multiset),
+					tmp)->type == T_STRING)
+		  s = tmp.u.string;
+		else
+		  s = NULL;
+	      }
+#else
+	      if (ITEM(value->u.multiset->ind)[0].type == T_STRING)
+		s = ITEM(value->u.multiset->ind)[0].u.string;
+	      else
+		s = NULL;
+#endif
+	      if (s) {
+		addr = (ub1 *)s->str;
+		len = s->len;
+		fty = SQLT_LBI;
+		break;
+	      }
 	    }
 	    
 	  default:
 	    Pike_error("Bad value type in argument 2 to "
-		  "Oracle.oracle->big_query()\n");
+		       "Oracle.oracle->big_typed_query()\n");
 	}
 	
 	bind.bind[bind.bindnum].addr=addr;
@@ -1822,7 +1944,7 @@ static void f_big_query_create(INT32 args)
 	else
 	{
 	  Pike_error("Bad index type in argument 2 to "
-		"Oracle.oracle->big_query()\n");
+		     "Oracle.oracle->big_typed_query()\n");
 	}
 	if(rc)
 	{
@@ -1846,10 +1968,19 @@ static void f_big_query_create(INT32 args)
 	}
       }
   }
+  debug_malloc_touch(dbcon->context);
+#ifndef REUSE_DEFINES
+  if(dbquery->field_info)
+  {
+    free_array(dbquery->field_info);
+    dbquery->field_info=0;
+  }
+#endif
   THREADS_ALLOW();
   LOCK(dbcon->lock);
   dbresult->dbcon_lock=1;
 
+
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"OCIExec query_type=%d\n",dbquery->query_type);
 #endif
@@ -1871,7 +2002,7 @@ static void f_big_query_create(INT32 args)
   {
       CHECK_INTERPRETER_LOCK();
       if(d_flag>1 && thread_for_id(th_self()) != Pike_thread_id)
-        fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
+        Pike_fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
   }
 #endif
 
@@ -1916,7 +2047,7 @@ static void f_big_query_create(INT32 args)
   {
       CHECK_INTERPRETER_LOCK();
       if(d_flag>1 && thread_for_id(th_self()) != Pike_thread_id)
-        fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
+        Pike_fatal("thread_for_id() (or Pike_thread_id) failed in interpreter.h! %p != %p\n",thread_for_id(th_self()),Pike_thread_id);
   }
 #endif
 
@@ -1959,6 +2090,14 @@ static void dbdate_sprintf(INT32 args)
   char buffer[100];
   sword rc;
   sb4 bsize=100;
+  int mode = 0;
+  if(args>0 && Pike_sp[-args].type == PIKE_T_INT)
+    mode = Pike_sp[-args].u.integer;
+  if(mode != 'O' && mode != 's') {
+    pop_n_elems(args);
+    push_undefined();
+    return;
+  }
   rc=OCIDateToText(get_global_error_handle(),
 		   &THIS_DBDATE->date,
 		   0,
@@ -1981,8 +2120,6 @@ static void dbdate_cast(INT32 args)
   get_all_args("Oracle.Date->cast",args,"%s",&s);
   if(!strcmp(s,"int"))
   {
-    struct tm *tm;
-    time_t t;
     ub1 hour, min, sec, month,day;
     sb2 year;
 
@@ -2016,6 +2153,14 @@ static void dbnull_create(INT32 args)
 
 static void dbnull_sprintf(INT32 args)
 {
+  int mode = 0;
+  if(args>0 && Pike_sp[-args].type == PIKE_T_INT)
+    mode = Pike_sp[-args].u.integer;
+  pop_n_elems(args);
+  if(mode != 'O') {
+    push_undefined();
+    return;
+  }
   switch(THIS_DBNULL->type.type)
   {
     case T_INT: push_text("Oracle.NULLint"); break;
@@ -2033,7 +2178,7 @@ static void dbnull_not(INT32 args)
   push_int(1);
 }
 
-void pike_module_init(void)
+PIKE_MODULE_INIT
 {
   long offset=0;
 #ifdef ORACLE_HOME
@@ -2049,23 +2194,19 @@ void pike_module_init(void)
   fprintf(stderr,"%s\n",__FUNCTION__);
 #endif
 
+  
   if(OCIInitialize(
-    OCI_OBJECT | 
-#ifdef ORACLE_USE_THREADS
-    OCI_DEFAULT,
-#else
-    OCI_THREADED,
-#endif
-    0, 0, 0, 0) != OCI_SUCCESS)
+    OCI_OBJECT | ORACLE_INIT_FLAGS,
+    0, ocimalloc, ocirealloc, ocifree) != OCI_SUCCESS)
   {
 #ifdef ORACLE_DEBUG
-    fprintf(stderr,"OCIInitialize failed\n");
+    fprintf(stderr,"OCIInitizlie failed\n");
 #endif
     return;
   }
 
   if(oracle_program)
-    fatal("Oracle module initiated twice!\n");
+    Pike_fatal("Oracle module initiated twice!\n");
 
   MY_START_CLASS(dbcon); {
 
@@ -2084,8 +2225,9 @@ void pike_module_init(void)
 /*      add_function("query_type",f_query_type,"function(:int)",0); */
 
       MY_START_CLASS(dbresult); {
-	ADD_FUNCTION("create", f_big_query_create,
-		     tFunc(tOr(tVoid,tMap(tStr,tMix)) tComma tOr(tVoid,tInt) tComma tOr(tVoid,tObj),tVoid), ID_PUBLIC);
+	ADD_FUNCTION("create", f_big_typed_query_create,
+		     tFunc(tOr(tVoid,tMap(tStr,tMix)) tComma tOr(tVoid,tInt)
+			   tComma tOr(tVoid,tObj),tVoid), ID_PUBLIC);
 	
 	/* function(:int) */
 	ADD_FUNCTION("num_fields", f_num_fields,tFunc(tNone,tInt), ID_PUBLIC);
@@ -2096,11 +2238,11 @@ void pike_module_init(void)
 	/* function(:int|array(string|int)) */
 	ADD_FUNCTION("fetch_row", f_fetch_row,tFunc(tNone,tOr(tInt,tArr(tOr(tStr,tInt)))), ID_PUBLIC);
 	
-	MY_END_CLASS(big_query);
 #ifdef PROGRAM_USES_PARENT
-	big_query_program->flags|=PROGRAM_USES_PARENT;
+	Pike_compiler->new_program->flags|=PROGRAM_USES_PARENT;
 #endif
-	big_query_program->flags|=PROGRAM_DESTRUCT_IMMEDIATE;
+	MY_END_CLASS(big_typed_query);
+	big_typed_query_program->flags|=PROGRAM_DESTRUCT_IMMEDIATE;
       }
       
       
@@ -2115,14 +2257,17 @@ void pike_module_init(void)
 	ADD_FUNCTION("`[]=",protect_dbresultinfo,tFunc(tStr tComma tMix,tVoid),0);
 #ifdef ORACLE_DEBUG
 	set_gc_check_callback(gc_dbresultinfo_struct);
+#endif
+#ifdef PROGRAM_USES_PARENT
+	Pike_compiler->new_program->flags|=PROGRAM_USES_PARENT;
 #endif
 	MY_END_CLASS(dbresultinfo);
       }
 
-      MY_END_CLASS(compile_query);
 #ifdef PROGRAM_USES_PARENT
-      compile_query_program->flags|=PROGRAM_USES_PARENT;
+	Pike_compiler->new_program->flags|=PROGRAM_USES_PARENT;
 #endif
+      MY_END_CLASS(compile_query);
     }
 
     ADD_FUNCTION("timeout_limit", f_dbcon_timeout_limit,
@@ -2165,7 +2310,7 @@ void pike_module_init(void)
 
 static void call_atexits(void);
 
-void pike_module_exit(void)
+PIKE_MODULE_EXIT
 {
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"%s\n",__FUNCTION__);
@@ -2177,11 +2322,17 @@ void pike_module_exit(void)
     global_error_handle=0;
   }
 
+  if(oracle_environment)
+  {
+    OCIHandleFree(oracle_environment, OCI_HTYPE_ENV);
+    oracle_environment=0;
+  }
+
 #define FREE_PROG(X) if(X) { free_program(X) ; X=NULL; }
 #define FREE_OBJ(X) if(X) { free_object(X) ; X=NULL; }
   FREE_PROG(oracle_program);
   FREE_PROG(compile_query_program);
-  FREE_PROG(big_query_program);
+  FREE_PROG(big_typed_query_program);
   FREE_PROG(dbresultinfo_program);
   FREE_PROG(Date_program);
   FREE_PROG(NULL_program);
@@ -2227,8 +2378,7 @@ static void call_atexits(void)
 
 #else /* HAVE_ORACLE */
 
-void pike_module_init(void)  {}
-void pike_module_exit(void)  {}
+PIKE_MODULE_INIT  {}
+PIKE_MODULE_EXIT  {}
 
 #endif
-
