@@ -1,10 +1,13 @@
 /*
- * $Id: pgresult.c,v 1.16 2002/02/07 13:33:03 grubba Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: pgresult.c,v 1.20 2002/10/21 17:06:22 marcus Exp $
+*/
+
+/*
  * Postgres95 support for pike/0.5 and up
  *
- * (C) 1997 Francesco Chemolli <kinkie@kame.usr.dsi.unimi.it>
- *
  * This code is provided AS IS, and may be copied and distributed freely,
  * under the terms of the GNU General Public License, version 2.
  *
@@ -65,7 +68,7 @@
 #include "builtin_functions.h"
 #include "module_support.h"
 
-RCSID("$Id: pgresult.c,v 1.16 2002/02/07 13:33:03 grubba Exp $");
+RCSID("$Id: pgresult.c,v 1.20 2002/10/21 17:06:22 marcus Exp $");
 
 #ifdef _REENTRANT
 PIKE_MUTEX_T pike_postgres_result_mutex STATIC_MUTEX_INIT;
@@ -78,8 +81,6 @@ PIKE_MUTEX_T pike_postgres_result_mutex STATIC_MUTEX_INIT;
 
 #include "pg_types.h"
 
-/* must be included last */
-#include "module_magic.h"
 
 
 #define THIS ((struct postgres_result_object_data *) Pike_fp->current_storage)
