@@ -1,3 +1,10 @@
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: sane.c,v 1.15 2002/10/21 17:06:22 marcus Exp $
+*/
+
 #include "config.h"
 
 #if defined(HAVE_SANE_SANE_H) || defined(HAVE_SANE_H)
@@ -29,20 +36,16 @@
 
 #include "../Image/image.h"
 
-/* must be included last */
-#include "module_magic.h"
 
-RCSID("$Id: sane.c,v 1.8 2000/12/01 08:10:23 hubbe Exp $");
+#define sp Pike_sp
 
-/*
-**! module SANE
-**!
-**!	This module enables access to the SANE (Scanner Access Now Easy)
-**!     library from pike
-**!
-**! note
-**!	$Id: sane.c,v 1.8 2000/12/01 08:10:23 hubbe Exp $
-*/
+RCSID("$Id: sane.c,v 1.15 2002/10/21 17:06:22 marcus Exp $");
+
+/*! @module SANE
+ *!
+ *!  This module enables access to the SANE (Scanner Access Now Easy)
+ *!  library from pike
+ */
 
 static int sane_is_inited;
 
@@ -68,32 +71,28 @@ static void push_device( SANE_Device *d )
 }
 
 
-/*
-**! method array(mapping) list-scanners()
-**!
-**!  Returns an array with all available scanners.
-**!
-**!    Example:
-**!     <pre>
-**!    Pike v0.7 release 120 running Hilfe v2.0 (Incremental Pike Frontend)
-**!   > SANE.list_scanners();
-**!     Result: ({
-**!            ([
-**!              "model":"Astra 1220S     ",
-**!              "name":"umax:/dev/scg1f",
-**!              "type":"flatbed scanner",
-**!              "vendor":"UMAX    "
-**!            ]),
-**!            ([
-**!              "model":"Astra 1220S     ",
-**!              "name":"net:lain.idonex.se:umax:/dev/scg1f",
-**!              "type":"flatbed scanner",
-**!              "vendor":"UMAX    "
-**!            ])
-**!        })
-**!
-**!     </pre>
-*/
+/*! @decl array(mapping) list_scanners()
+ *!
+ *!  Returns an array with all available scanners.
+ *!
+ *! @example
+ *!   Pike v0.7 release 120 running Hilfe v2.0 (Incremental Pike Frontend)
+ *!   > SANE.list_scanners();
+ *!     Result: ({
+ *!            ([
+ *!              "model":"Astra 1220S     ",
+ *!              "name":"umax:/dev/scg1f",
+ *!              "type":"flatbed scanner",
+ *!              "vendor":"UMAX    "
+ *!            ]),
+ *!            ([
+ *!              "model":"Astra 1220S     ",
+ *!              "name":"net:lain.idonex.se:umax:/dev/scg1f",
+ *!              "type":"flatbed scanner",
+ *!              "vendor":"UMAX    "
+ *!            ])
+ *!        })
+ */
 static void f_list_scanners( INT32 args )
 {
   SANE_Device **devices;
@@ -206,10 +205,11 @@ static void push_option_descriptor( const SANE_Option_Descriptor *o )
   f_aggregate_mapping( sp - osp );
 }
 
-/*
-**! class Scanner
-**!    Scanner s = Scanner( scanner name )
-*/
+/*! @class Scanner
+ */
+
+/*! @decl void create(string name)
+ */
 static void f_scanner_create( INT32 args )
 {
   char *name;
@@ -220,9 +220,73 @@ static void f_scanner_create( INT32 args )
     Pike_error("Failed to open scanner \"%s\"\n", name );
 }
 
-/*
-**! function array(mapping) list_options( )
-*/
+/*! @decl array(mapping) list_options()
+ *!
+ *! This method returns an array where every element is a
+ *! mapping, layed out as follows.
+ *!
+ *! @mapping
+ *!   @member string "name"
+ *!
+ *!   @member string "title"
+ *!
+ *!   @member string "desc"
+ *!
+ *!   @member string "type"
+ *!     @string
+ *!       @value "boolean"
+ *!       @value "int"
+ *!       @value "float"
+ *!       @value "string"
+ *!       @value "button"
+ *!       @value "group"
+ *!     @endstring
+ *!   @member string "unit"
+ *!     @string
+ *!       @value "none"
+ *!       @value "pixel"
+ *!       @value "bit"
+ *!       @value "mm"
+ *!       @value "dpi"
+ *!       @value "percent"
+ *!       @value "microsend"
+ *!     @endstring
+ *!   @member int "size"
+ *!
+ *!   @member multiset "cap"
+ *!     @multiset
+ *!       @index "soft_select"
+ *!       @index "hard_select"
+ *!       @index "emulate"
+ *!       @index "automatic"
+ *!       @index "inactive"
+ *!       @index "advanced"
+ *!     @endmultiset
+ *!   @member int(0..0)|mapping "constraint"
+ *!     Constraints can be of three different types; range, word list or string
+ *!     list. Constraint contains 0 if there are no constraints.
+ *!
+ *!     @mapping range
+ *!       @member string "type"
+ *!         Contains the value "range".
+ *!       @member int "max"
+ *!       @member int "min"
+ *!       @member int "quant"
+ *!     @endmapping
+ *!
+ *!     @mapping "word list"
+ *!       @member string "type"
+ *!         Contains the value "list".
+ *!       @member array(float|int) "list"
+ *!     @endmapping
+ *!
+ *!     @mapping "string list"
+ *!       @member string "type"
+ *!         Contains the value "list".
+ *!       @member array(string) "list"
+ *!     @endmapping
+ *! @endmapping
+ */
 static void f_scanner_list_options( INT32 args )
 {
   int i, n;
@@ -248,11 +312,10 @@ static int find_option( char *name, const SANE_Option_Descriptor **p )
 }
 
 
-/*
-**! function void set_option( string name, mixed new_value )
-**! function void set_option( string name )
-**!    If no value is specified, the option is set to it's default value
-*/
+/*! @decl void set_option( string name, mixed new_value )
+ *! @decl void set_option( string name )
+ *!    If no value is specified, the option is set to it's default value
+ */
 static void f_scanner_set_option( INT32 args )
 {
   char *name;
@@ -297,9 +360,8 @@ static void f_scanner_set_option( INT32 args )
 }
 
 
-/*
-**! function mixed get_option( string name )
-*/
+/*! @decl mixed get_option( string name )
+ */
 static void f_scanner_get_option( INT32 args )
 {
   char *name;
@@ -338,9 +400,18 @@ static void f_scanner_get_option( INT32 args )
   }
 }
 
-/*
-**! function mapping(string:int) get_parameters(  )
-*/
+/*! @decl mapping(string:int) get_parameters()
+ *!
+ *! @returns
+ *!   @mapping
+ *!  	@member int "format"
+ *!  	@member int "last_frame"
+ *!  	@member int "lines"
+ *!  	@member int "depth"
+ *!  	@member int "pixels_per_line"
+ *!  	@member int "bytes_per_line"
+ *!   @endmapping
+ */
 static void f_scanner_get_parameters( INT32 args )
 {
   SANE_Parameters p;
@@ -420,9 +491,8 @@ static void assert_image_program()
     Pike_error("No Image.Image?!\n");
 }
 
-/*
-**! function Image.Image simple_scan(  )
-*/
+/*! @decl Image.Image simple_scan()
+ */
 static void f_scanner_simple_scan( INT32 args )
 {
   SANE_Parameters p;
@@ -475,9 +545,8 @@ static void f_scanner_simple_scan( INT32 args )
   push_object( o );
 }
 
-/*
-**! function void row_scan(function(Image.Image,int,Scanner:void) callback)
-*/
+/*! @decl void row_scan(function(Image.Image,int,Scanner:void) callback)
+ */
 static void f_scanner_row_scan( INT32 args )
 {
   SANE_Parameters p;
@@ -620,9 +689,8 @@ static void nonblocking_row_scan_callback( int fd, void *_c )
   }
 }
 
-/*
-**! function void nonblocking_row_scan(function(Image.Image,int,Scanner,int:void) callback)
-*/
+/*! @decl void nonblocking_row_scan(function(Image.Image,int,Scanner,int:void) callback)
+ */
 static void f_scanner_nonblocking_row_scan( INT32 args )
 {
   SANE_Parameters p;
@@ -677,11 +745,26 @@ static void f_scanner_nonblocking_row_scan( INT32 args )
   push_int( 0 );
 }
 
+/*! @decl void cancel_scan()
+ */
 static void f_scanner_cancel_scan( INT32 args )
 {
   sane_cancel( THIS->h );
 }
 
+/*! @endclass
+ */
+
+/*! @decl constant FrameGray
+ *! @decl constant FrameRGB
+ *! @decl constant FrameRed
+ *! @decl constant FrameGreen
+ *! @decl cosntant FrameBlue
+ */
+
+/*! @endmodule
+ */
+
 static void init_scanner_struct( struct object *p )
 {
   THIS->h = 0;
@@ -693,7 +776,7 @@ static void exit_scanner_struct( struct object *p )
     sane_close( THIS->h );
 }
 
-void pike_module_init()
+PIKE_MODULE_INIT
 {
   struct program *p;
   add_function( "list_scanners", f_list_scanners,
@@ -740,7 +823,7 @@ void pike_module_init()
   free_program( p );
 }
 
-void pike_module_exit()
+PIKE_MODULE_EXIT
 {
   if( sane_is_inited )
     sane_exit();
@@ -749,7 +832,7 @@ void pike_module_exit()
 }
 
 #else
-#include "module_magic.h"
-void pike_module_init() {}
-void pike_module_exit() {}
+#include "module.h"
+PIKE_MODULE_INIT {}
+PIKE_MODULE_EXIT {}
 #endif
