@@ -1,9 +1,10 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-/**/
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: efuns.c,v 1.120 2004/05/13 20:38:14 mast Exp $
+*/
+
 #include "global.h"
 #include "fdlib.h"
 #include "interpret.h"
@@ -25,7 +26,7 @@
 #include "file_machine.h"
 #include "file.h"
 
-RCSID("$Id: efuns.c,v 1.97 2001/10/02 19:18:32 hubbe Exp $");
+RCSID("$Id: efuns.c,v 1.120 2004/05/13 20:38:14 mast Exp $");
 
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
@@ -75,11 +76,20 @@ RCSID("$Id: efuns.c,v 1.97 2001/10/02 19:18:32 hubbe Exp $");
 #include <process.h>
 #endif
 
+#define sp Pike_sp
+
 /* #define DEBUG_FILE */
 /* #define READDIR_DEBUG */
 
 #ifdef __NT__
 
+#include <winbase.h>
+
+/* Old versions of the headerfiles don't have this constant... */
+#ifndef INVALID_SET_FILE_POINTER
+#define INVALID_SET_FILE_POINTER ((DWORD)-1)
+#endif
+
 /* Dynamic load of functions that doesn't exist in all Windows versions. */
 
 static HINSTANCE kernel32lib;
@@ -89,14 +99,14 @@ static HINSTANCE kernel32lib;
   static PIKE_CONCAT(NAME,type) NAME
 
 LINKFUNC(BOOL, movefileex, (
-  LPCTSTR lpExistingFileName,  // file name
-  LPCTSTR lpNewFileName,       // new file name
-  DWORD dwFlags                // move options
+  LPCTSTR lpExistingFileName,  /* file name     */
+  LPCTSTR lpNewFileName,       /* new file name */
+  DWORD dwFlags                /* move options  */
 ));
 
 #endif
 
-struct array *encode_stat(struct stat *s)
+struct array *encode_stat(PIKE_STAT_T *s)
 {
   struct array *a;
   a=allocate_array(7);
@@ -127,9 +137,35 @@ struct array *encode_stat(struct stat *s)
   return a;
 }
 
+/*! @decl Stdio.Stat file_stat(string path, void|int(0..1) symlink)
+ *!
+ *! Stat a file.
+ *!
+ *! If the argument @[symlink] is @tt{1@} symlinks will not be followed.
+ *!
+ *! @returns
+ *!   If the path specified by @[path] doesn't exist @tt{0@} (zero) will
+ *!   be returned.
+ *!
+ *!   Otherwise an object describing the properties of @[path] will be
+ *!   returned.
+ *!
+ *! @note
+ *!   In Pike 7.0 and earlier this function returned an array with 7 elements.
+ *!   The old behaviour can be simulated with the following function:
+ *!   @code{array(int) file_stat(string path, void|int(0..1) symlink)
+ *!         {
+ *!           File.Stat st = predef::file_stat(path, symlink);
+ *!           if (!st) return 0;
+ *!           return (array(int))st;
+ *!         }@}
+ *!
+ *! @seealso
+ *!   @[Stdio.Stat], @[Stdio.File->stat()]
+ */
 void f_file_stat(INT32 args)
 {
-  struct stat st;
+  PIKE_STAT_T st;
   int i, l;
   struct pike_string *str;
   
@@ -138,10 +174,10 @@ void f_file_stat(INT32 args)
   if(sp[-args].type != T_STRING)
     Pike_error("Bad argument 1 to file_stat()\n");
 
-  VALID_FILE_IO("file_stat","read");
-
   str = sp[-args].u.string;
-  l = (args>1 && !IS_ZERO(sp+1-args))?1:0;
+  l = (args>1 && !UNSAFE_IS_ZERO(sp+1-args))?1:0;
+
+  VALID_FILE_IO("file_stat","read");
 
   if (strlen(str->str) != (size_t)str->len) {
     /* Filenames with NUL are not supported. */
@@ -169,23 +205,43 @@ void f_file_stat(INT32 args)
   }
 }
 
+/*! @decl int file_truncate(string file, int length)
+ *!
+ *! Truncates the file @[file] to the length specified in @[length].
+ *!
+ *! @returns
+ *!   Returns 1 if ok, 0 if failed.
+ */
 void f_file_truncate(INT32 args)
 {
-#ifdef HAVE_LSEEK64
-  long long len;
+#if defined (INT64) || defined (HAVE_TRUNCATE64)
+  INT64 len;
 #else
-  INT32 len;
+  off_t len;
 #endif
   struct pike_string *str;
   int res;
 
-  if(args<1 || sp[-args].type != T_STRING)
-    Pike_error("Bad argument 1 to file_truncate(string filename,int length).\n");
-  if(args<2 || sp[1-args].type != T_INT)
-    Pike_error("Bad argument 2 to file_truncate(string filename,int length).\n");
+  if(args < 2)
+    SIMPLE_TOO_FEW_ARGS_ERROR("file_truncate", 2);
+  if(sp[-args].type != T_STRING)
+    SIMPLE_BAD_ARG_ERROR("file_truncate", 1, "string");
+
+#if defined (INT64) && defined (AUTO_BIGNUM)
+#if defined (HAVE_FTRUNCATE64) || SIZEOF_OFF_T > SIZEOF_INT_TYPE
+  if(is_bignum_object_in_svalue(&Pike_sp[1-args])) {
+    if (!int64_from_bignum(&len, Pike_sp[1-args].u.object))
+      Pike_error ("Bad argument 2 to file_truncate(). Length too large.\n");
+  }
+  else
+#endif
+#endif
+    if(sp[1-args].type != T_INT)
+      SIMPLE_BAD_ARG_ERROR("file_truncate", 2, "int");
+    else
+      len = sp[1-args].u.integer;
 
   str = sp[-args].u.string;
-  len = sp[1-args].u.integer;
 
   if (strlen(str->str) != (size_t)str->len) {
     /* Filenames with NUL are not supported. */
@@ -206,18 +262,35 @@ void f_file_truncate(INT32 args)
       errno = GetLastError();
       res=-1;
     } else {
-      if(SetFilePointer(h, len, NULL, FILE_BEGIN) != 0xffffffff &&
-	 SetEndOfFile(h))
-	res=0;
-      else {
+      LONG high;
+      DWORD err;
+#ifdef INT64
+      high = len >> 32;
+      len &= (1 << 32) - 1;
+#else
+      high = 0;
+#endif
+      if (SetFilePointer(h, DO_NOT_WARN ((LONG) len), &high, FILE_BEGIN) ==
+	  INVALID_SET_FILE_POINTER &&
+	  (err = GetLastError()) != NO_ERROR) {
+	errno = err;
+	res = -1;
+      }
+      else if (!SetEndOfFile(h)) {
 	errno = GetLastError();
 	res=-1;
       }
+      else
+	res = 0;
       CloseHandle(h);
     }
   }
 #else  /* !__NT__ */
+#ifdef HAVE_TRUNCATE64
+  res = truncate64 (str->str, len);
+#else
   res=truncate(str->str, len);
+#endif
 #endif /* __NT__ */
 
   pop_n_elems(args);
@@ -225,6 +298,47 @@ void f_file_truncate(INT32 args)
   push_int(!res);
 }
 
+/*! @decl mapping(string:int) filesystem_stat(string path)
+ *!
+ *! Returns a mapping describing the properties of the filesystem
+ *! containing the path specified by @[path].
+ *!
+ *! @returns
+ *!   If a filesystem cannot be determined @tt{0@} (zero) will be returned.
+ *!
+ *!   Otherwise a mapping(string:int) with the following fields will be
+ *!   returned:
+ *!   @mapping
+ *!     @member int "blocksize"
+ *!       Size in bytes of the filesystem blocks.
+ *!     @member int "blocks"
+ *!       Size of the entire filesystem in blocks.
+ *!     @member int "bfree"
+ *!       Number of free blocks in the filesystem.
+ *!     @member int "bavail"
+ *!       Number of available blocks in the filesystem.
+ *!       This is usually somewhat less than the @tt{"bfree"@} value, and
+ *!       can usually be adjusted with eg tunefs(1M).
+ *!     @member int "files"
+ *!       Total number of files (aka inodes) allowed by this filesystem.
+ *!     @member int "ffree"
+ *!       Number of free files in the filesystem.
+ *!     @member int "favail"
+ *!       Number of available files in the filesystem.
+ *!       This is usually the same as the @tt{"ffree"@} value, and can
+ *!       usually be adjusted with eg tunefs(1M).
+ *!     @member int "fsname"
+ *!       Name assigned to the filesystem.
+ *!     @member int "fstype"
+ *!       Type of filesystem (eg @tt{"nfs"@}).
+ *!   @endmapping
+ *!
+ *! @note
+ *!   Please note that not all members are present on all OSs.
+ *!
+ *! @seealso
+ *!   @[file_stat()]
+ */
 #ifdef __NT__
 
 void f_filesystem_stat( INT32 args )
@@ -329,7 +443,7 @@ void f_filesystem_stat(INT32 args)
 #endif /* HAVE_STRUCT_STATFS */
 #else /* !HAVE_STATFS */
 #ifdef HAVE_USTAT
-  struct stat statbuf;
+  PIKE_STAT_T statbuf;
   struct ustat st;
 #else /* !HAVE_USTAT */
   /* Should not be reached */
@@ -456,6 +570,10 @@ void f_filesystem_stat(INT32 args)
 #endif /* HAVE_STATVFS || HAVE_STATFS || HAVE_USTAT */
 #endif /* __NT__ */
 
+/*! @decl void werror(string msg, mixed ... args)
+ *!
+ *! Write to standard error.
+ */
 void f_werror(INT32 args)
 {
   if(!args)
@@ -475,9 +593,19 @@ void f_werror(INT32 args)
   pop_n_elems(args);
 }
 
+/*! @decl int rm(string f)
+ *!
+ *! Remove a file or directory.
+ *!
+ *! @returns
+ *!   Returns @tt{0@} (zero) on failure, @tt{1@} otherwise.
+ *!
+ *! @seealso
+ *!   @[mkdir()], @[Stdio.recursive_rm()]
+ */
 void f_rm(INT32 args)
 {
-  struct stat st;
+  PIKE_STAT_T st;
   INT32 i;
   struct pike_string *str;
 
@@ -544,6 +672,19 @@ void f_rm(INT32 args)
   push_int(i);
 }
 
+/*! @decl int mkdir(string dirname, void|int mode)
+ *!
+ *! Create a directory.
+ *!
+ *! If @[mode] is specified, it's will be used for the new directory after
+ *! being @tt{&@}'ed with the current umask (on OS'es that support this).
+ *!
+ *! @returns
+ *!   Returns @tt{0@} (zero) on failure, @tt{1@} otherwise.
+ *!
+ *! @seealso
+ *!   @[rm()], @[cd()], @[Stdio.mkdirhier()]
+ */
 void f_mkdir(INT32 args)
 {
   struct pike_string *str;
@@ -584,10 +725,10 @@ void f_mkdir(INT32 args)
      implementations (e.g. Mac OS X) */
   s = str->str;
   s_dup = 0;
-  if (s[strlen(s) - 1] == '/') {
-    if (s_dup = strdup(s)) {
+  if (str->len && s[str->len - 1] == '/') {
+    if ((s_dup = strdup(s))) {
       s = s_dup;
-      s[strlen(s) - 1] = '\0';
+      s[str->len - 1] = '\0';
     }
   }
   
@@ -626,7 +767,11 @@ void f_mkdir(INT32 args)
 	mode = ((mode & 0777) | (statbuf1.st_mode & ~0777)) & ~mask;
 	do {
 	  i = chmod(str->str, mode) != -1;
-	} while (!i && (errno == EINTR));
+	  if (i || errno != EINTR) break;
+	  THREADS_DISALLOW_UID();
+	  check_threads_etc();
+	  THREADS_ALLOW_UID();
+	} while (1);
       }
       if (i) {
 	i = LSTAT(str->str, &statbuf2) != -1;
@@ -654,6 +799,14 @@ void f_mkdir(INT32 args)
 #define HAVE_READDIR_R
 #endif
 
+/*! @decl array(string) get_dir(string dirname)
+ *!
+ *! Returns an array of all filenames in the directory @[dirname], or
+ *! @tt{0@} (zero) if the directory does not exist.
+ *!
+ *! @seealso
+ *!   @[mkdir()], @[cd()]
+ */
 void f_get_dir(INT32 args)
 {
   struct svalue *save_sp=sp;
@@ -839,6 +992,16 @@ void f_get_dir(INT32 args)
     push_int(0);
 }
 
+/*! @decl int cd(string s)
+ *!
+ *! Change the current directory for the whole Pike process.
+ *!
+ *! @returns
+ *!   Returns @tt{1@} for success, @tt{0@} (zero) otherwise.
+ *!
+ *! @seealso
+ *!   @[getcwd()]
+ */
 void f_cd(INT32 args)
 {
   INT32 i;
@@ -867,6 +1030,13 @@ void f_cd(INT32 args)
   push_int(i);
 }
 
+/*! @decl string getcwd()
+ *!
+ *! Returns the current working directory.
+ *!
+ *! @seealso
+ *!   @[cd()]
+ */
 void f_getcwd(INT32 args)
 {
   char *e;
@@ -903,6 +1073,28 @@ void f_getcwd(INT32 args)
   free(e);
 }
 
+/*! @decl int exece(string file, array(string) args)
+ *! @decl int exece(string file, array(string) args, @
+ *!                 mapping(string:string) env)
+ *!
+ *!   This function transforms the Pike process into a process running
+ *!   the program specified in the argument @[file] with the arguments @[args].
+ *!
+ *!   If the mapping @[env] is present, it will completely replace all
+ *!   environment variables before the new program is executed.
+ *!
+ *! @returns
+ *!   This function only returns if something went wrong during @tt{exece(2)@},
+ *!   and in that case it returns @tt{0@} (zero).
+ *!
+ *! @note
+ *!   The Pike driver _dies_ when this function is called. You must either
+ *!   use @[fork()] or @[Process.create_process()] if you wish to execute a
+ *!   program and still run the Pike runtime.
+ *!
+ *! @seealso
+ *!   @[Process.create_process()], @[fork()], @[Stdio.File->pipe()]
+ */
 void f_exece(INT32 args)
 {
   INT32 e;
@@ -1016,6 +1208,26 @@ void f_exece(INT32 args)
   push_int(0);
 }
 
+/*! @decl int mv(string from, string to)
+ *!
+ *! Rename or move a file or directory.
+ *!
+ *! If the destination already exists, it will be replaced.
+ *! Replacement often only works if @[to] is of the same type as
+ *! @[from], i.e. a file can only be replaced by another file and so
+ *! on. Also, a directory will commonly be replaced only if it's
+ *! empty.
+ *!
+ *! On some OSs this function can't move directories, only rename
+ *! them.
+ *!
+ *! @returns
+ *!   Returns @tt{0@} (zero) on failure, @tt{1@} otherwise. Call
+ *!   @[errno()] to get more error info on failure.
+ *!
+ *! @seealso
+ *!   @[rm()]
+ */
 void f_mv(INT32 args)
 {
   INT32 i;
@@ -1024,7 +1236,7 @@ void f_mv(INT32 args)
 #ifdef __NT__
   int orig_errno = errno;
   int err;
-  struct stat st;
+  PIKE_STAT_T st;
 #endif
 
   if(args<2)
@@ -1218,6 +1430,14 @@ no_nt_rename_kludge:
   push_int(!i);
 }
 
+/*! @decl string strerror(int errno)
+ *!
+ *! This function returns a description of an error code. The error
+ *! code is usually obtained from eg @[Stdio.File->errno()].
+ *!
+ *! @note
+ *!   On some platforms the string returned can be somewhat nondescriptive.
+ */
 void f_strerror(INT32 args)
 {
   char *s;
@@ -1248,6 +1468,16 @@ void f_strerror(INT32 args)
   }
 }
 
+/*! @decl int errno()
+ *!
+ *! This function returns the system error from the last file operation.
+ *!
+ *! @note
+ *!   Note that you should normally use @[Stdio.File->errno()] instead.
+ *!
+ *! @seealso
+ *!   @[Stdio.File->errno()], @[strerror()]
+ */
 void f_errno(INT32 args)
 {
   pop_n_elems(args);
@@ -1282,16 +1512,17 @@ void init_files_efuns(void)
   }
 #endif
 
-/* function(string,int|void:int *) */
-  ADD_EFUN("file_stat",f_file_stat,tFunc(tStr tOr(tInt,tVoid),tObj), OPT_EXTERNAL_DEPEND);
+/* function(string,int|void:object) */
+  ADD_EFUN("file_stat",f_file_stat,tFunc(tStr tOr(tInt,tVoid),tObj), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
 
-  ADD_EFUN("file_truncate",f_file_truncate,tFunc(tStr tInt,tInt),0);
+  /* function(string,int:int(0..1)) */
+  ADD_EFUN("file_truncate",f_file_truncate,tFunc(tStr tInt,tInt),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
 
 
 #if defined(HAVE_STATVFS) || defined(HAVE_STATFS) || defined(HAVE_USTAT) || defined(__NT__)
   
 /* function(string:mapping(string:string|int)) */
-  ADD_EFUN("filesystem_stat", f_filesystem_stat,tFunc(tStr,tMap(tStr,tOr(tStr,tInt))), OPT_EXTERNAL_DEPEND);
+  ADD_EFUN("filesystem_stat", f_filesystem_stat,tFunc(tStr,tMap(tStr,tOr(tStr,tInt))), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
 #endif /* HAVE_STATVFS || HAVE_STATFS */
   
 /* function(:int) */
@@ -1310,13 +1541,13 @@ void init_files_efuns(void)
   ADD_EFUN("mv", f_mv,tFunc(tStr tStr,tInt), OPT_SIDE_EFFECT);
   
 /* function(string:string *) */
-  ADD_EFUN("get_dir",f_get_dir,tFunc(tStr,tArr(tStr)),OPT_EXTERNAL_DEPEND);
+  ADD_EFUN("get_dir",f_get_dir,tFunc(tStr,tArr(tStr)),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
   
 /* function(string:int) */
   ADD_EFUN("cd",f_cd,tFunc(tStr,tInt),OPT_SIDE_EFFECT);
   
 /* function(:string) */
-  ADD_EFUN("getcwd",f_getcwd,tFunc(tNone,tStr),OPT_EXTERNAL_DEPEND);
+  ADD_EFUN("getcwd",f_getcwd,tFunc(tNone,tStr),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
   
 /* function(string,mixed*,void|mapping(string:string):int) */
   ADD_EFUN("exece",f_exece,tFunc(tStr tArr(tMix) tOr(tVoid,tMap(tStr,tStr)),tInt),OPT_SIDE_EFFECT); 
