@@ -1,8 +1,15 @@
-  FILE_FUNC("open",file_open,"function(string,string,void|int:int)")
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: file_functions.h,v 1.25 2003/09/12 12:04:29 marcus Exp $
+*/
+
+FILE_FUNC("open",file_open,"function(string,string,void|int:int)")
   FILE_FUNC("close",file_close,"function(string|void:int)")
   FILE_FUNC("read",file_read,"function(int|void,int|void:string)")
 #ifndef __NT__
-  FILE_FUNC("peek",file_peek,"function(void:int)")
+  FILE_FUNC("peek",file_peek,"function(float|int|void:int)")
 #endif
   FILE_FUNC("write",file_write,"function(string|array(string),void|mixed...:int)")
 #ifdef WITH_OOB
@@ -48,6 +55,9 @@
   FILE_FUNC("set_buffer",file_set_buffer,"function(int,string|void:void)")
   FILE_FUNC("open_socket",file_open_socket,"function(int|void,string|void:int)")
   FILE_FUNC("connect",file_connect,"function(string,int:int)|function(string,int,string,int:int)")
+#ifdef HAVE_SYS_UN_H
+  FILE_FUNC("connect_unix",file_connect_unix,"function(string:int)")
+#endif /* HAVE_SYS_UN_H */
   FILE_FUNC("query_address",file_query_address,"function(int|void:string)")
   FILE_FUNC("create",file_create,"function(void|string,void|string:void)")
   FILE_FUNC("`<<",file_lsh,"function(mixed:object)")
@@ -61,6 +71,14 @@
   FILE_FUNC("trylock",file_trylock,"function(void|int:object)")
 #endif
 
+#if !defined(__NT__) && (defined(HAVE_POSIX_OPENPT) || defined(PTY_MASTER_PATHNAME))
+   FILE_FUNC("openpt",file_openpt,"function(string:int)")
+#endif
+
+#if defined(HAVE_GRANTPT) || defined(USE_PT_CHMOD)
+   FILE_FUNC("grantpt",file_grantpt,"function(void:string)")
+#endif
+
 #if defined(HAVE_TERMIOS_H)
    FILE_FUNC("tcgetattr",file_tcgetattr,"function(void:mapping)")
    FILE_FUNC("tcsetattr",file_tcsetattr,"function(mapping,void|string:int)")
