@@ -1,6 +1,11 @@
 /*
- * $Id: sendfile.c,v 1.55 2003/09/08 10:27:32 mast Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: sendfile.c,v 1.63 2003/12/09 09:57:14 grubba Exp $
+*/
+
+/*
  * Sends headers + from_fd[off..off+len-1] + trailers to to_fd asyncronously.
  *
  * Henrik Grubbstrm 1999-04-02
@@ -68,6 +73,8 @@
 #endif /* HAVE_SYS_MMAN_H */
 #endif
 
+#define sp Pike_sp
+
 /* #define SF_DEBUG */
 
 #ifdef SF_DEBUG
@@ -244,7 +251,6 @@ static void sf_call_callback(struct pike_sendfile *this)
 static void call_callback_and_free(struct callback *cb, void *this_, void *arg)
 {
   struct pike_sendfile *this = this_;
-  int sz;
 
   SF_DFPRINTF((stderr, "sendfile: Calling callback...\n"));
 
@@ -340,9 +346,6 @@ void low_do_sendfile(struct pike_sendfile *this)
   SF_DFPRINTF((stderr, "sendfile: Worker started\n"));
 
   if ((this->from_file) && (this->len)) {
-    struct iovec *iov;
-    int iovcnt;
-
 #if defined(HAVE_FREEBSD_SENDFILE) || defined(HAVE_HPUX_SENDFILE)
     off_t sent = 0;
     int len = this->len;
@@ -405,9 +408,9 @@ void low_do_sendfile(struct pike_sendfile *this)
       case EFAULT:
 	/* Bad arguments */
 #ifdef HAVE_FREEBSD_SENDFILE
-	fatal("FreeBSD style sendfile(): EFAULT\n");
+	Pike_fatal("FreeBSD style sendfile(): EFAULT\n");
 #else /* !HAVE_FREEBSD_SENDFILE */
-	fatal("HP/UX style sendfile(): EFAULT\n");
+	Pike_fatal("HP/UX style sendfile(): EFAULT\n");
 #endif /* HAVE_FREEBSD_SENDFILE */
 	break;
       case EBADF:
@@ -461,7 +464,7 @@ void low_do_sendfile(struct pike_sendfile *this)
 #if 0 /* mmap is slower than read/write on most if not all systems */
 #if defined(HAVE_MMAP) && defined(HAVE_MUNMAP)
     {
-      struct stat st;
+      PIKE_STAT_T st;
 
       if (!fd_fstat(this->from_fd, &st) &&
 	  S_ISREG(st.st_mode)) {
@@ -614,7 +617,7 @@ static void worker(void *this_)
 
   low_do_sendfile(this);
 
-  mt_lock_interpreter();
+  low_mt_lock_interpreter();	/* Can run even if threads_disabled. */
 
   /*
    * Unlock the fd's
@@ -624,7 +627,10 @@ static void worker(void *this_)
       this->from->flags &= ~FILE_LOCK_FD;
     } else {
       /* Destructed. */
-      fd_close(this->from_fd);
+      if (this->from_fd >= 0) {
+	set_backend_for_fd(this->from_fd, NULL);
+	fd_close(this->from_fd);
+      }
       /* Paranoia */
       this->from->fd = -1;
     }
@@ -633,7 +639,10 @@ static void worker(void *this_)
     this->to->flags &= ~FILE_LOCK_FD;
   } else {
     /* Destructed */
-    fd_close(this->to_fd);
+    if (this->to_fd >= 0) {
+      set_backend_for_fd(this->to_fd, NULL);
+      fd_close(this->to_fd);
+    }
     /* Paranoia */
     this->to->fd = -1;
   }
@@ -643,6 +652,10 @@ static void worker(void *this_)
    * * Call the callback.
    * * Get rid of extra ref to the object, and free ourselves.
    */
+#ifdef PIKE_DEBUG
+  if (this->backend_callback)
+    Pike_fatal ("Didn't expect a backend callback to be installed already.\n");
+#endif
   this->backend_callback =
     add_backend_callback(call_callback_and_free, this, 0);
 
@@ -694,7 +707,7 @@ static void sf_create(INT32 args)
   MEMSET(&sf, 0, sizeof(struct pike_sendfile));
   sf.callback.type = T_INT;
 
-  get_all_args("sendfile", args, "%A%O%i%i%A%o%*",
+  get_all_args("sendfile", args, "%A%O%l%l%A%o%*",
 	       &(sf.headers), &(sf.from_file), &(sf.offset),
 	       &(sf.len), &(sf.trailers), &(sf.to_file), &cb);
 
@@ -726,7 +739,7 @@ static void sf_create(INT32 args)
 #ifdef PIKE_DEBUG
     if ((sp[5-args].type != T_OBJECT) ||
 	(sp[5-args].u.object != sf.to_file)) {
-      fatal("sendfile: Stack out of sync(1).\n");
+      Pike_fatal("sendfile: Stack out of sync(1).\n");
     }
 #endif /* PIKE_DEBUG */
     sf.to_file = *ob;
@@ -765,7 +778,7 @@ static void sf_create(INT32 args)
 #ifdef PIKE_DEBUG
       if ((sp[1-args].type != T_OBJECT) ||
 	  (sp[1-args].u.object != sf.from_file)) {
-	fatal("sendfile: Stack out of sync(2).\n");
+	Pike_fatal("sendfile: Stack out of sync(2).\n");
       }
 #endif /* PIKE_DEBUG */
       sf.from_file = *ob;
@@ -895,7 +908,7 @@ static void sf_create(INT32 args)
     struct array *a = copy_array(sf.headers);
 #ifdef PIKE_DEBUG
     if ((sp[-args].type != T_ARRAY) || (sp[-args].u.array != sf.headers)) {
-      fatal("sendfile: Stack out of sync(3).\n");
+      Pike_fatal("sendfile: Stack out of sync(3).\n");
     }
 #endif /* PIKE_DEBUG */
     free_array(sf.headers);
@@ -906,7 +919,7 @@ static void sf_create(INT32 args)
     struct array *a = copy_array(sf.trailers);
 #ifdef PIKE_DEBUG
     if ((sp[4-args].type != T_ARRAY) || (sp[4-args].u.array != sf.trailers)) {
-      fatal("sendfile: Stack out of sync(4).\n");
+      Pike_fatal("sendfile: Stack out of sync(4).\n");
     }
 #endif /* PIKE_DEBUG */
     free_array(sf.trailers);
@@ -931,7 +944,6 @@ static void sf_create(INT32 args)
 
   {
     /* Threaded blocking mode possible */
-    THREAD_T th_id;
 
     /* Make sure both file objects are in blocking mode.
      */
