@@ -1,13 +1,16 @@
 /*
- * $Id: udp.c,v 1.22 2003/11/03 20:38:53 mast Exp $
- */
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: udp.c,v 1.42 2003/11/03 21:04:36 mast Exp $
+*/
 
 #define NO_PIKE_SHORTHAND
 #include "global.h"
 
 #include "file_machine.h"
 
-RCSID("$Id: udp.c,v 1.22 2003/11/03 20:38:53 mast Exp $");
+RCSID("$Id: udp.c,v 1.42 2003/11/03 21:04:36 mast Exp $");
 #include "fdlib.h"
 #include "interpret.h"
 #include "svalue.h"
@@ -25,6 +28,7 @@ RCSID("$Id: udp.c,v 1.22 2003/11/03 20:38:53 mast Exp $");
 #include "bignum.h"
 
 #include "module_support.h"
+#include "builtin_functions.h"
 
 #ifdef HAVE_SYS_TYPE_H
 #include <sys/types.h>
@@ -132,27 +136,49 @@ RCSID("$Id: udp.c,v 1.22 2003/11/03 20:38:53 mast Exp $");
 struct udp_storage {
   int fd;
   int my_errno;
+   
+  int type;
+  int protocol;
+
   struct svalue read_callback;
 };
 
+// FIXME: This class does not keep an extra ref while the read
+// callback is registered. It'll probably segfault in
+// udp_read_callback if the object has run out of refs. /mast
+
 #undef THIS
 #define THIS ((struct udp_storage *)Pike_fp->current_storage)
 #define THISOBJ (Pike_fp->current_object)
 #define FD (THIS->fd)
 
+/*! @module Stdio
+ */
+
+/*! @class UDP
+ */
+
 extern void get_inet_addr(struct sockaddr_in *addr,char *name);
 
+/*! @decl object bind(int port)
+ *! @decl object bind(int port, string address)
+ *!
+ *! Binds a port for recieving or transmitting UDP.
+ */
 static void udp_bind(INT32 args)
 {
   struct sockaddr_in addr;
   int o;
   int fd,tmp;
+#if !defined(SOL_IP) && defined(HAVE_GETPROTOBYNAME)
+  static int ip_proto_num = -1;
+#endif /* !SOL_IP && HAVE_GETPROTOBYNAME */
 
   
-  if(args < 1) Pike_error("Too few arguments to dumudp->bind()\n");
+  if(args < 1) Pike_error("Too few arguments to udp->bind()\n");
 
   if(Pike_sp[-args].type != PIKE_T_INT)
-    Pike_error("Bad argument 1 to dumudp->bind()\n");
+    Pike_error("Bad argument 1 to udp->bind()\n");
 
   if(FD != -1)
   {
@@ -161,7 +187,7 @@ static void udp_bind(INT32 args)
     FD = -1;
   }
 
-  fd = fd_socket(AF_INET, SOCK_DGRAM, 0);
+  fd = fd_socket(AF_INET, THIS->type, THIS->protocol);
   if(fd < 0)
   {
     pop_n_elems(args);
@@ -177,9 +203,36 @@ static void udp_bind(INT32 args)
   {
     fd_close(fd);
     THIS->my_errno=errno;
-    Pike_error("UDP.bind: setsockopt failed\n");
+    Pike_error("UDP.bind: setsockopt SO_REUSEADDR failed\n");
   }
 
+#ifndef SOL_IP
+#ifdef HAVE_GETPROTOBYNAME
+  if (ip_proto_num == -1) {
+    /* NOTE: getprotobyname(3N) is not MT-safe. */
+    struct protoent *proto = getprotobyname("ip");
+    if (!proto) {
+      /* No entry for IP, assume zero. */
+      ip_proto_num = 0;
+    } else {
+      ip_proto_num = proto->p_proto;
+    }
+  }
+#define SOL_IP ip_proto_num
+#else /* !HAVE_GETPROTOBYNAME */
+#define SOL_IP 0
+#endif /* HAVE_GETPROTOBYNAME */
+#endif /* SOL_IP */
+
+#ifdef IP_HDRINCL
+  /*  From mtr-0.28:net.c: FreeBSD wants this to avoid sending
+      out packets with protocol type RAW to the network. */
+
+  if (THIS->type==SOCK_RAW && THIS->protocol==255 /* raw */)
+     if(fd_setsockopt(fd, SOL_IP, IP_HDRINCL, (char *)&o, sizeof(int)))
+	Pike_error("UDP.bind: setsockopt IP_HDRINCL failed\n");
+#endif /* IP_HDRINCL */
+
   MEMSET((char *)&addr,0,sizeof(struct sockaddr_in));
 
   if(args > 1 && Pike_sp[1-args].type==PIKE_T_STRING) {
@@ -201,7 +254,8 @@ static void udp_bind(INT32 args)
   {
     fd_close(fd);
     THIS->my_errno=errno;
-    Pike_error("UDP.bind: failed to bind to port %d\n",Pike_sp[-args].u.integer);
+    Pike_error("UDP.bind: failed to bind to port %d\n",
+	       (u_short)Pike_sp[-args].u.integer);
     return;
   }
 
@@ -210,6 +264,19 @@ static void udp_bind(INT32 args)
   ref_push_object(THISOBJ);
 }
 
+/*! @decl int(0..1) enable_broadcast()
+ *!
+ *! Set the broadcast flag. 
+ *! If enabled then sockets receive packets sent  to  a  broadcast
+ *! address  and  they are allowed to send packets to a
+ *! broadcast address.
+ *!
+ *! @returns
+ *!  Returns @tt{1@} on success, @tt{0@} (zero) otherwise.
+ *!
+ *! @note
+ *!  This is normally only avalable to root users.
+ */
 void udp_enable_broadcast(INT32 args)
 {
 #ifdef SO_BROADCAST
@@ -224,7 +291,13 @@ void udp_enable_broadcast(INT32 args)
 #endif /* SO_BROADCAST */
 }
 
-/* int wait(int timeout) */
+/*! @decl int(0..1) wait(int|float timeout)
+ *!
+ *! Check for data and wait max. @[timeout] seconds.
+ *!
+ *! @returns
+ *!  Returns @tt{1@} if data are ready, @tt{0@} (zero) otherwise.
+ */
 void udp_wait(INT32 args)
 {
 #ifdef HAVE_POLL
@@ -306,6 +379,24 @@ void udp_wait(INT32 args)
 
 #define UDP_BUFFSIZE 65536
 
+/*! @decl mapping(string:int|string) read()
+ *! @decl mapping(string:int|string) read(int flag)
+ *!
+ *! Read from the UDP socket.
+ *!
+ *! Flag @[flag] is a bitfield, 1 for out of band data and 2 for peek
+ *!
+ *! @returns
+ *!  mapping(string:int|string) in the form 
+ *!	([
+ *!	   "data" : string recieved data
+ *!	   "ip" : string   recieved from this ip
+ *!	   "port" : int    ...and this port
+ *!	])
+ *!
+ *! @seealso
+ *!   @[set_read_callback()]
+ */
 void udp_read(INT32 args)
 {
   int flags = 0, res=0, fd, e;
@@ -340,7 +431,7 @@ void udp_read(INT32 args)
     e = errno;
     THREADS_DISALLOW();
 
-    check_signals(0, 0, 0);
+    check_threads_etc();
   } while((res==-1) && (e==EINTR));
 
   THIS->my_errno=e;
@@ -368,7 +459,7 @@ void udp_read(INT32 args)
 	  Pike_error("Out of memory\n");
 #ifdef ENOTSOCK
        case ENOTSOCK:
-	  fatal("reading from non-socket fd!!!\n");
+	  Pike_fatal("reading from non-socket fd!!!\n");
 #endif
        case EWOULDBLOCK:
 	  push_int( 0 );
@@ -392,9 +483,21 @@ void udp_read(INT32 args)
   f_aggregate_mapping( 6 );
 }
 
+/*! @decl int send(string to, int port, string message)
+ *! @decl int send(string to, int port, string message, int flags)
+ *!
+ *! Send data to a UDP socket. The recepient address will be @[to]
+ *! and port will be @[port].
+ *!
+ *! Flag @[flag] is a bitfield, 1 for out of band data and
+ *! 2 for don't route flag.
+ *!
+ *! @returns
+ *! The number of bytes that were actually written.
+ */
 void udp_sendto(INT32 args)
 {
-  int flags = 0, i, fd, e;
+  int flags = 0, fd, e;
   ptrdiff_t res = 0;
   struct sockaddr_in to;
   char *str;
@@ -402,7 +505,7 @@ void udp_sendto(INT32 args)
 
   if(FD < 0)
     Pike_error("UDP: not open\n");
-
+  
   check_all_args("send", args,
 		 BIT_STRING, BIT_INT, BIT_STRING, BIT_INT|BIT_VOID, 0);
   
@@ -419,13 +522,16 @@ void udp_sendto(INT32 args)
 #endif /* MSG_DONTROUTE */
     }
     if(Pike_sp[3-args].u.integer & ~3) {
-      Pike_error("Illegal 'flags' value passed to udp->send(string m,string t,int p,[int flags])\n");
+      Pike_error("Illegal 'flags' value passed to "
+		 "udp->send(string to, int port, string message, int flags)\n");
     }
   }
-  if(!args)
+  else if(args != 3)
     Pike_error("Illegal number of arguments to udp->sendto(string to"
-	  ", string message, int port[, int flags])\n");
+	       ", int port, string message, void|int flags)\n");
 
+  if( Pike_sp[2-args].type!=PIKE_T_STRING ) 
+    Pike_error("Illegal type of argument to sendto, got non-string message.\n");
 
   if( Pike_sp[-args].type==PIKE_T_STRING ) 
     get_inet_addr(&to, Pike_sp[-args].u.string->str);
@@ -445,7 +551,7 @@ void udp_sendto(INT32 args)
     e = errno;
     THREADS_DISALLOW();
 
-    check_signals(0, 0, 0);
+    check_threads_etc();
   } while((res == -1) && e==EINTR);
   
   if(res<0)
@@ -484,6 +590,9 @@ void zero_udp(struct object *ignored)
   MEMSET(THIS, 0, sizeof(struct udp_storage));
   THIS->read_callback.type=PIKE_T_INT;
   FD = -1;
+
+  THIS->type=SOCK_DGRAM;
+  THIS->protocol=0;
 }
 
 void exit_udp(struct object *ignored)
@@ -499,13 +608,14 @@ void exit_udp(struct object *ignored)
     THREADS_DISALLOW();
   }
   free_svalue(& THIS->read_callback );
+  THIS->read_callback.type = PIKE_T_INT; /* Avoid uncertainty. */
 }
 
 #define THIS_DATA ((struct udp_storage *)data)
 
 static void udp_read_callback( int fd, void *data )
 {
-  if(IS_ZERO(&THIS_DATA->read_callback))
+  if(UNSAFE_IS_ZERO(&THIS_DATA->read_callback))
     set_read_callback(THIS_DATA->fd, 0, 0);
   else {
     apply_svalue(& THIS_DATA->read_callback, 0);
@@ -521,17 +631,19 @@ static void udp_set_read_callback(INT32 args)
 
   if(args != 1)
     Pike_error("Wrong number of arguments to file->set_read_callback().\n");
-  
-  if(IS_ZERO(& THIS->read_callback))
-    assign_svalue(& THIS->read_callback, Pike_sp-1);
-  else
-    assign_svalue_no_free(& THIS->read_callback, Pike_sp-1);
 
-  if(IS_ZERO(& THIS->read_callback))
-    set_read_callback(FD, 0, 0);
-  else
+  if (UNSAFE_IS_ZERO (Pike_sp - 1)) {
+    free_svalue (&THIS->read_callback);
+    THIS->read_callback.type = PIKE_T_INT;
+    THIS->read_callback.u.integer = 0;
+    set_read_callback (FD, 0, 0);
+  }
+  else {
+    assign_svalue(& THIS->read_callback, Pike_sp-1);
     set_read_callback(FD, udp_read_callback, THIS);
-  pop_n_elems(args);
+  }
+
+  pop_stack();
   ref_push_object(THISOBJ);
 }
 
@@ -548,6 +660,10 @@ static void udp_set_nonblocking(INT32 args)
   ref_push_object(THISOBJ);
 }
 
+/*! @decl object set_blocking()
+ *!
+ *! Sets this object to be blocking.
+ */
 static void udp_set_blocking(INT32 args)
 {
   if (FD < 0) Pike_error("File not open.\n");
@@ -556,6 +672,25 @@ static void udp_set_blocking(INT32 args)
   ref_push_object(THISOBJ);
 }
 
+/*! @decl int(0..1) connect(string address, int port)
+ *!
+ *!   Establish an UDP connection.
+ *!
+ *!   This function connects an UDP socket previously created with
+ *!   @[Stdio.UDP()] to a remote socket. The @[address] is the IP name or
+ *!   number for the remote machine. 
+ *!
+ *! @returns
+ *!   Returns @tt{1@} on success, @tt{0@} (zero) otherwise.
+ *!
+ *! @note
+ *!   If the socket is in nonblocking mode, you have to wait
+ *!   for a write or close callback before you know if the connection
+ *!   failed or not.
+ *!
+ *! @seealso
+ *!   @[bind()], @[query_address()]
+ */
 static void udp_connect(INT32 args)
 {
   struct sockaddr_in addr;
@@ -598,6 +733,12 @@ static void udp_connect(INT32 args)
   }
 }
 
+/*! @decl string query_address()
+ *! 
+ *! Returns the local address of a socket on the form "x.x.x.x port".
+ *! If this file is not a socket, not connected or some other error occurs,
+ *! zero is returned.
+ */
 static void udp_query_address(INT32 args)
 {
   struct sockaddr_in addr;
@@ -631,18 +772,77 @@ static void udp_query_address(INT32 args)
   push_string(make_shared_string(buffer));
 }
 
+/*! @decl int errno()
+ *!
+ *! Returns the error code for the last command on this object.
+ *! Error code is normally cleared when a command is successful.
+ */
 static void udp_errno(INT32 args)
 {
    pop_n_elems(args);
    push_int(THIS->my_errno);
 }
 
+/*! @decl object set_type(int sock_type)
+ *! @decl object set_type(int sock_type, int family)
+ *!
+ *! Sets socket type and protocol family.
+ */
+static void udp_set_type(INT32 args)
+{
+   INT_TYPE type=0;
+   INT_TYPE proto=0;
+   if (args<2)
+      get_all_args("UDP.set_type",args,"%i",&type);
+   else
+      get_all_args("UDP.set_type",args,"%i%i",&type,&proto);
+
+   THIS->type=(int)type;
+   THIS->protocol=(int)proto;
+   
+   pop_n_elems(args);
+   ref_push_object(THISOBJ);
+}
+
+/*! @decl array(int) get_type()
+ *!
+ *! Returns socket type and protocol family.
+ */
+static void udp_get_type(INT32 args)
+{
+   pop_n_elems(args);
+   push_int(THIS->type);
+   push_int(THIS->protocol);
+   f_aggregate(2);
+}
+
+/*! @decl constant MSG_OOB
+ *! @fixme
+ *! Document this constant.
+ */
+
+/*! @decl constant MSG_PEEK
+ *! @fixme
+ *! Document this constant.
+ */
+
+/*! @endclass
+ */
+
+/*! @endmodule
+ */
+
 void init_udp(void)
 {
   start_new_program();
 
   ADD_STORAGE(struct udp_storage);
 
+  ADD_FUNCTION("set_type",udp_set_type,
+	       tFunc(tInt tOr(tVoid,tInt),tObj),0);
+  ADD_FUNCTION("get_type",udp_get_type,
+	       tFunc(tNone,tArr(tInt)),0);
+
   ADD_FUNCTION("bind",udp_bind,
 	       tFunc(tInt tOr(tVoid,tStr),tObj),0);
 
@@ -679,5 +879,63 @@ void init_udp(void)
   set_exit_callback(exit_udp);
 
   end_class("UDP",0);
-}
 
+  start_new_program();
+#ifdef SOCK_STREAM
+  add_integer_constant("STREAM",SOCK_STREAM,0);
+#endif
+#ifdef SOCK_DGRAM
+  add_integer_constant("DGRAM",SOCK_DGRAM,0);
+#endif
+#ifdef SOCK_SEQPACKET
+  add_integer_constant("SEQPACKET",SOCK_SEQPACKET,0);
+#endif
+#ifdef SOCK_RAW
+  add_integer_constant("RAW",SOCK_RAW,0);
+#endif
+#ifdef SOCK_RDM
+  add_integer_constant("RDM",SOCK_RDM,0);
+#endif
+#ifdef SOCK_PACKET
+  add_integer_constant("PACKET",SOCK_PACKET,0);
+#endif
+   push_program(end_program());
+   f_call_function(1);
+   simple_add_constant("SOCK",Pike_sp-1,0);
+   pop_stack();
+
+   start_new_program();
+   add_integer_constant("IP",0,0);       /* Dummy protocol for TCP.  */
+   add_integer_constant("HOPOPTS",0,0);  /* IPv6 Hop-by-Hop options.  */
+   add_integer_constant("ICMP",1,0);     /* Internet Control Message Protocol.  */
+   add_integer_constant("IGMP",2,0);     /* Internet Group Management Protocol. */
+   add_integer_constant("GGP", 3, 0);	 /* Gateway-Gateway Protocol.  */
+   add_integer_constant("IPIP",4,0);     /* IPIP tunnels (older KA9Q tunnels use 94).  */
+   add_integer_constant("TCP",6,0);      /* Transmission Control Protocol.  */
+   add_integer_constant("EGP",8,0);      /* Exterior Gateway Protocol.  */
+   add_integer_constant("PUP",12,0);     /* PUP protocol.  */
+   add_integer_constant("UDP",17,0);     /* User Datagram Protocol.  */
+   add_integer_constant("IDP",22,0);     /* XNS IDP protocol.  */
+   add_integer_constant("RDP", 27, 0);	 /* "Reliable Datagram" Protocol.  */
+   add_integer_constant("TP",29,0);      /* SO Transport Protocol Class 4.  */
+   add_integer_constant("IPV6",41,0);    /* IPv6 header.  */
+   add_integer_constant("ROUTING",43,0); /* IPv6 routing header.  */
+   add_integer_constant("FRAGMENT",44,0);/* IPv6 fragmentation header.  */
+   add_integer_constant("RSVP",46,0);    /* Reservation Protocol.  */
+   add_integer_constant("GRE",47,0);     /* General Routing Encapsulation.  */
+   add_integer_constant("ESP",50,0);     /* encapsulating security payload.  */
+   add_integer_constant("AH",51,0);      /* authentication header.  */
+   add_integer_constant("ICMPV6",58,0);  /* ICMPv6.  */
+   add_integer_constant("NONE",59,0);    /* IPv6 no next header.  */
+   add_integer_constant("DSTOPTS",60,0); /* IPv6 destination options.  */
+   add_integer_constant("MTP",92,0);     /* Multicast Transport Protocol.  */
+   add_integer_constant("ENCAP",98,0);   /* Encapsulation Header.  */
+   add_integer_constant("PIM",103,0);    /* Protocol Independent Multicast.  */
+   add_integer_constant("COMP",108,0);   /* Compression Header Protocol.  */
+   add_integer_constant("RAW",255,0);    /* Raw IP packets.  */
+   push_program(end_program());
+   f_call_function(1);
+   simple_add_constant("IPPROTO",Pike_sp-1,0);
+   pop_stack();
+
+}
