@@ -1,3 +1,10 @@
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: xml.c,v 1.57 2002/12/17 18:06:34 nilsson Exp $
+*/
+
 #include "global.h"
 
 #ifdef HAVE_CONFIG_H
@@ -19,8 +26,8 @@
 #include "block_alloc.h"
 #include "bignum.h"
 
-/* This must be included last! */
-#include "module_magic.h"
+
+#define sp Pike_sp
 
 /* FIXME:
  * recursive parsing will lock make the parser lock or
@@ -35,7 +42,9 @@ struct xmlinput
   PCHARP datap;
   ptrdiff_t len;
   ptrdiff_t pos;
+  struct mapping *callbackinfo;
   struct pike_string *to_free;
+  struct pike_string *entity;
 };
 
 BLOCK_ALLOC(xmlinput, 64)
@@ -59,8 +68,11 @@ struct xmlobj
   int rxml_mode;
 };
 
+
+static struct svalue location_string_svalue;
+
 #undef THIS
-#define THIS ((struct xmlobj *)(fp->current_storage))
+#define THIS ((struct xmlobj *)(Pike_fp->current_storage))
 
 /* FIXME: Make all these functions available inside pike
  */
@@ -565,29 +577,6 @@ static int isExtender(INT32 c)
   return 0;
 }
 
-static int isChar(INT32 c)
-{
- 
-  if(c <= 0xd7ff)
-  {
-    if(c>0x20 || c==0x9 || c==0x0a || c==0x0d) return 1;
-  }else{
-    /* FIXME: The test below is always false!
-     * /grubba 1998-12-05
-     * 
-     * So? This function isn't used anyways...
-     * /Hubbe 1998-12-16
-     */
-    if(c <=0xffd)
-    {
-      if(c>=0xe000) return 1;
-    }else{
-      if(c>=0x10000 && c<=0x10ffff) return 1;
-    }
-  }
-  return 0;
-}
-
 static INLINE int isSpace(INT32 c)
 {
   switch(c)
@@ -666,7 +655,7 @@ ISWRAP(isHexChar)
 
 #define PEEK(X) \
 ( IF_XMLDEBUG(((X) > data->input.len)? \
-    (fatal("PEEK out of bounds (%ld/%ld)\n", \
+    (Pike_fatal("PEEK out of bounds (%ld/%ld)\n", \
            PTRDIFF_T_TO_LONG(X), \
            PTRDIFF_T_TO_LONG(data->input.len)), 0):) \
   INDEX_PCHARP(data->input.datap,X) )
@@ -675,27 +664,26 @@ ISWRAP(isHexChar)
 
 #define POKE(X,Y) string_builder_putchar(&X,Y)
 
-#define PUSH(s) do {				\
-  struct xmlinput *i=alloc_xmlinput();		\
-  *i=data->input;				\
-  data->input.next=i;				\
-  data->input.pos=0;				\
-  data->input.datap=MKPCHARP_STR(s);		\
-  data->input.len=(s)->len;			\
-  copy_shared_string(data->input.to_free,s);   	\
-}while(0)
-
 #define XMLEOF() (data->input.len <= 0)
 
-#define POP() do {						\
-  struct xmlinput *i=data->input.next;				\
-  IF_XMLDEBUG(fprintf(stderr,"SMEG POP\n"));\
-  if(data->input.to_free) free_string(data->input.to_free); 	\
-  data->input=*i;						\
-  really_free_xmlinput(i);					\
- IF_XMLDEBUG(fprintf(stderr,"ptr=%p len=%d pos=%d to_free=%p\n", \
-                     data->input.datap.ptr, data->input.len, \
-                     data->input.pos, data->input.to_free)); \
+#define POP() do {							\
+  struct xmlinput *i=data->input.next;					\
+  IF_XMLDEBUG(fprintf(stderr,"SMEG POP\n"));				\
+  if (data->input.entity) {						\
+    if (data->input.to_free) {						\
+      mapping_string_insert_string(THIS->entities,			\
+				   data->input.entity,			\
+				   data->input.to_free);		\
+    }									\
+    free_string(data->input.entity);					\
+  }									\
+  if(data->input.to_free) free_string(data->input.to_free);		\
+  if(data->input.callbackinfo) free_mapping(data->input.callbackinfo);	\
+  data->input=*i;							\
+  really_free_xmlinput(i);						\
+ IF_XMLDEBUG(fprintf(stderr,"ptr=%p len=%d pos=%d to_free=%p\n",	\
+                     data->input.datap.ptr, data->input.len,		\
+                     data->input.pos, data->input.to_free));		\
 } while(0)
 
 static inline int xmlread(int z,struct xmldata *data, int line)
@@ -892,15 +880,24 @@ static int gobble(struct xmldata *data, char *s)
     if(THIS->entities)							 \
     {									 \
       struct pike_string *name=0;					 \
-      ONERROR tmp3;							 \
+      struct mapping *callbackinfo;                                      \
+      ONERROR tmp3,tmp4;						 \
+									 \
       map_delete_no_free(THIS->entities, sp-1, sp);			 \
       name=dmalloc_touch(struct pike_string *, sp[-1].u.string);	 \
       sp[-1]=*sp;							 \
       SET_ONERROR(tmp3, do_free_string, name);				 \
 									 \
-      do {								 \
+      UPDATE_LOCATION(data->input.pos, data->input.callbackinfo);        \
+      callbackinfo=copy_mapping(data->input.callbackinfo);               \
+      SET_ONERROR(tmp4, do_free_mapping, callbackinfo);                  \
+      push_constant_text("previous");                                    \
+      ref_push_mapping(data->input.callbackinfo);                        \
+      mapping_insert(callbackinfo, sp-2, sp-1);                          \
+      pop_n_elems(2);                                                    \
 									 \
-	if(IS_ZERO(sp-1))						 \
+      do {								 \
+	if(UNSAFE_IS_ZERO(sp-1))					 \
 	{								 \
 	  pop_stack();							 \
           if(index_shared_string(name,0)=='%')				 \
@@ -918,7 +915,7 @@ static int gobble(struct xmldata *data, char *s)
 	  push_int(ATTR);						 \
 	  f_aggregate_mapping(2); /* attributes */			 \
 	  push_int(0); /* no data */					 \
-	  low_sys(data);						 \
+	  very_low_sys(data,callbackinfo);				 \
 	  if(sp[-1].type != T_STRING)				 	 \
 	  {								 \
 	    pop_stack();						 \
@@ -931,6 +928,7 @@ static int gobble(struct xmldata *data, char *s)
 	  if(sp[-1].type!=T_STRING)					 \
 	  {								 \
 	    XMLERROR("XML->__entities value is not a string!");		 \
+	    break;                                        		 \
 	  }								 \
 	}								 \
 									 \
@@ -941,20 +939,21 @@ static int gobble(struct xmldata *data, char *s)
 	  debug_malloc_touch(s);					 \
 	  sp--;								 \
 	  IF_XMLDEBUG(fprintf(stderr,"Entity expands to: %s\n",s->str)); \
-	  SET_ONERROR(tmp2, free_xmldata, &my_tmp);			 \
 	  check_stack(10);						 \
+	  SET_ONERROR(tmp2, free_xmldata, &my_tmp);			 \
 	  my_tmp.input.to_free=s;					 \
 	  my_tmp.input.datap=MKPCHARP_STR(s);				 \
 	  my_tmp.input.len=s->len;					 \
 	  my_tmp.input.pos=0;						 \
 	  my_tmp.input.next=0;						 \
+	  add_ref(my_tmp.input.callbackinfo=callbackinfo);		 \
 	  PARSE_RECURSIVELY;						 \
 	  if(THIS->entities)						 \
 	    mapping_string_insert_string(THIS->entities, name, s);	 \
-	  UNSET_ONERROR(tmp2);						 \
-	  free_string(s);						 \
+	  CALL_AND_UNSET_ONERROR(tmp2);					 \
 	}								 \
       }while(0);							 \
+      CALL_AND_UNSET_ONERROR(tmp4);					 \
       CALL_AND_UNSET_ONERROR(tmp3);					 \
     }									 \
   }while(0)
@@ -1000,7 +999,7 @@ static int gobble(struct xmldata *data, char *s)
     XMLERROR("Missing ';' after parsed entity reference.");	\
   READ(1);							\
   PARSE_REF(ATTR,PARSE_RECURSIVELY);				\
-  DO_IF_DEBUG(if(sp<spsave) fatal("Stack underflow\n");)        \
+  DO_IF_DEBUG(if(sp<spsave) Pike_fatal("Stack underflow\n");)        \
 }while(0)
 
 
@@ -1088,15 +1087,20 @@ static int gobble(struct xmldata *data, char *s)
      SYS();					\
     }  } while (0)
 
+#define UPDATE_LOCATION(P,m) do{		\
+  push_int64((P));				\
+  mapping_insert((m),				\
+		 &location_string_svalue,	\
+		 sp-1);				\
+  pop_stack();					\
+}while(0)
 
-static inline void low_sys(struct xmldata *data)
+static inline void very_low_sys(struct xmldata *data,
+				struct mapping *callbackinfo)
 {
   struct xmlinput *i=&data->input;
   check_stack(1+data->num_extra_args);
-  push_constant_text("location");
-  while(i->next) i=i->next;
-  push_int64(i->pos);
-  f_aggregate_mapping(2);
+  ref_push_mapping(i->callbackinfo);
   assign_svalues_no_free(sp, data->extra_args,
 			  data->num_extra_args,
 			  data->extra_arg_types);
@@ -1104,10 +1108,17 @@ static inline void low_sys(struct xmldata *data)
   apply_svalue(data->func, 5+data->num_extra_args);
 }
 
+static inline void low_sys(struct xmldata *data)
+{
+  struct xmlinput *i=&data->input;
+  UPDATE_LOCATION(i->pos,i->callbackinfo);
+  very_low_sys(data, i->callbackinfo);
+}
+
 static void sys(struct xmldata *data)
 {
   low_sys(data);
-  if(IS_ZERO(sp-1)) pop_stack();
+  if(SAFE_IS_ZERO(sp-1)) pop_stack();
 }
 
 #define SYS() sys(data)
@@ -1131,15 +1142,22 @@ static int read_smeg_pereference(struct xmldata *data)
   {
     XMLERROR("XML->__entities is not a mapping");
   }else{
+    struct mapping *callbackinfo;
     struct pike_string *name=0;
-    ONERROR tmp3;			
+    struct pike_string *full_name=0;
+    int external_entity_value=0;
+    ONERROR tmp3,tmp4,tmp5;
 
     push_constant_text("%");
     SIMPLE_READNAME();
+
     add_ref(name=sp[-1].u.string);
     SET_ONERROR(tmp3, do_free_string, name);
 
     f_add(2);
+    add_ref(full_name=sp[-1].u.string);
+    SET_ONERROR(tmp5, do_free_string, full_name);
+
     if(PEEK(0)!=';')
       XMLERROR("Missing ';' after parsed entity reference.");
     READ(1);
@@ -1150,15 +1168,24 @@ static int read_smeg_pereference(struct xmldata *data)
     free_svalue(sp-1);
     sp[-1]=*sp;
 
+    UPDATE_LOCATION(data->input.pos, data->input.callbackinfo);
+    callbackinfo=copy_mapping(data->input.callbackinfo);
+    SET_ONERROR(tmp4, do_free_mapping, callbackinfo);
+    push_constant_text("previous");
+    ref_push_mapping(data->input.callbackinfo);
+    mapping_insert(callbackinfo, sp-2, sp-1);
+    pop_n_elems(2);
+    
     do {
-      if(IS_ZERO(sp-1))
+      if(UNSAFE_IS_ZERO(sp-1))
       {
+	external_entity_value = 1;
 	pop_stack();
 	push_constant_text("%");
 	ref_push_string(name);
 	f_aggregate_mapping(0);
 	push_int(0); /* no data */
-	low_sys(data);
+	very_low_sys(data, callbackinfo);
 	if(sp[-1].type != T_STRING)
 	{
 	  pop_stack();
@@ -1176,7 +1203,27 @@ static int read_smeg_pereference(struct xmldata *data)
 	IF_XMLDEBUG(fprintf(stderr, "ptr=%p len=%d pos=%d to_free=%p\n",
 			    data->input.datap.ptr, data->input.len,
 			    data->input.pos, data->input.to_free));
-	PUSH(s);
+
+	/* PUSH(s) */
+	{
+	  struct xmlinput *i=alloc_xmlinput();
+	  *i=data->input;
+	  data->input.next=i;
+	  data->input.pos=0;
+	  data->input.datap=MKPCHARP_STR(s);
+	  data->input.len=(s)->len;
+	  free_mapping(data->input.callbackinfo);
+	  data->input.callbackinfo=callbackinfo;
+	  copy_shared_string(data->input.to_free,s);
+	  if (external_entity_value || !full_name) {
+	    data->input.entity = 0;
+	  } else {
+	    copy_shared_string(data->input.entity, full_name);
+	  }
+	}
+	CALL_AND_UNSET_ONERROR(tmp5);
+	UNSET_ONERROR(tmp4);
+
 	READ(0); /* autopop empty strings */
 	pop_stack();
 	CALL_AND_UNSET_ONERROR(tmp3);
@@ -1184,6 +1231,8 @@ static int read_smeg_pereference(struct xmldata *data)
       }
 
     }while(0);
+    CALL_AND_UNSET_ONERROR(tmp5);
+    CALL_AND_UNSET_ONERROR(tmp4);
     CALL_AND_UNSET_ONERROR(tmp3);
   }
   return 0;
@@ -1421,7 +1470,7 @@ static void simple_read_attributes(struct xmldata *data,
     if(is_cdata)
     {
       struct svalue *s=low_mapping_lookup(is_cdata,sp-1);
-      if(s && IS_ZERO(s))
+      if(s && SAFE_IS_ZERO(s))
 	iscd=0;
     }
 
@@ -1529,8 +1578,8 @@ void read_choice_seq_or_name(struct xmldata *data, int maybe_pcdata)
 #endif
 	    {
 #ifdef PIKE_DEBUG
-	      if(!IS_ZERO(sp-num))
-		fatal("num not in sync with arguments on stack!\n");
+	      if(!SAFE_IS_ZERO(sp-num))
+		Pike_fatal("num not in sync with arguments on stack!\n");
 #endif
 	      if(seq == ',')
 		sp[-num].u.string=make_shared_string(",");
@@ -1597,7 +1646,7 @@ void read_choice_seq_or_name(struct xmldata *data, int maybe_pcdata)
 
 #ifdef PIKE_DEBUG
   if(sp != save_sp2+1)
-    fatal("XML: Stack corruption in read_choice_seq_or_name.\n");
+    Pike_fatal("XML: Stack corruption in read_choice_seq_or_name.\n");
 #endif
 }
 
@@ -1612,7 +1661,7 @@ static int really_low_parse_dtd(struct xmldata *data)
   {
 #ifdef PIKE_DEBUG
     if(sp<save_sp)
-      fatal("Stack underflow.\n");
+      Pike_fatal("Stack underflow.\n");
 #endif
     switch(PEEK(0))
     {
@@ -1927,7 +1976,7 @@ static int really_low_parse_dtd(struct xmldata *data)
 		    }
 #ifdef PIKE_DEBUG
 		    if(sp<save)
-		      fatal("Stack underflow.\n");
+		      Pike_fatal("Stack underflow.\n");
 #endif
 		    f_aggregate(DO_NOT_WARN(sp - save));
 		    SKIPSPACE();
@@ -2062,7 +2111,7 @@ static int really_low_parse_dtd(struct xmldata *data)
 		    }
 #ifdef PIKE_DEBUG
 		    if(sp<save)
-		      fatal("Stack underflow.\n");
+		      Pike_fatal("Stack underflow.\n");
 #endif
 		    f_aggregate(DO_NOT_WARN(sp - save));
 		    f_aggregate(2);
@@ -2110,6 +2159,7 @@ static int really_low_parse_dtd(struct xmldata *data)
 		    default:
 		      XMLERROR("Expected PUBLIC or SYSTEM, found something else.");
 		      push_int(0);
+		      
 		  }
 		  SKIPSPACE();
 		  if(PEEK(0)!='>')
@@ -2186,7 +2236,7 @@ static int really_low_parse_dtd(struct xmldata *data)
 	    }
 #ifdef PIKE_DEBUG
 	    if(sp<save_sp)
-	      fatal("Stack underflow.\n");
+	      Pike_fatal("Stack underflow.\n");
 #endif
 	    break;
 
@@ -2241,7 +2291,7 @@ static int low_parse_dtd(struct xmldata *data)
   done=really_low_parse_dtd(data);
 #ifdef PIKE_DEBUG
   if(sp<save_sp)
-    fatal("Stack underflow.\n");
+    Pike_fatal("Stack underflow.\n");
 #endif
 #ifdef VERBOSE_XMLDEBUG
   fprintf(stderr,"Exiting low_parse_dtd %p %p\n",sp,save_sp);
@@ -2592,7 +2642,6 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		
 	    }
 
-	    /*
 	    if(toplevel)
 	    {
 	      done=1;
@@ -2600,8 +2649,6 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 	      if(!XMLEOF())
 		XMLERROR("All data must be in one top-level tag.\n");
 	    }
-	    */
-
 	}
     }
   }
@@ -2621,15 +2668,21 @@ static int low_parse_xml(struct xmldata *data,
   pop_stack();
 #ifdef PIKE_DEBUG
   if(sp<save_sp)
-    fatal("Stack underflow.\n");
+    Pike_fatal("Stack underflow.\n");
 #endif
   f_aggregate(DO_NOT_WARN(sp - save_sp));
   /* There is now one value on the stack */
   return !!end;
 }
 
+/*! @module spider
+ */
+
+/*! @decl array parse_xml(string xml, function cb)
+ */
 static void parse_xml(INT32 args)
 {
+  int d;
   struct svalue tmp;
   struct pike_string *s;
   struct xmldata data;
@@ -2657,16 +2710,22 @@ static void parse_xml(INT32 args)
   data.input.len=s->len;
   data.input.pos=0;
   data.input.to_free=0;
+  data.input.entity=0;
+  data.input.callbackinfo=allocate_mapping(0);
   data.input.next=0;
   data.func=sp+1-args;
   data.extra_args=sp+2-args;
   data.num_extra_args=args-2;
-  data.extra_arg_types = 0xffff; /* FIXME */
+  data.extra_arg_types=0;
+
+  for(d=0;d<data.num_extra_args;d++)
+    data.extra_arg_types |= 1 << data.extra_args[d].type;
+
   data.allow_pesmeg_everywhere=0;
 
   SET_ONERROR(e,free_xmldata, &data);
   low_parse_xml(&data,0,1);
-  UNSET_ONERROR(e);
+  CALL_AND_UNSET_ONERROR(e);
   tmp=*--sp;
   pop_n_elems(args);
   *sp++=tmp;
@@ -2681,8 +2740,20 @@ static void free_xmldata(struct xmldata *data)
     free_string(data->input.to_free);
     data->input.to_free=0;
   }
+  if(data->input.callbackinfo)
+  {
+    free_mapping(data->input.callbackinfo);
+    data->input.callbackinfo=0;
+  }
+    
 }
 
+/*! @class XML
+ *! @appears Parser.XML.Simple
+ */
+
+/*! @decl void define_entity_raw(string entity, string raw)
+ */
 static void define_entity_raw(INT32 args)
 {
   check_all_args("XML->define_entity_raw",args,BIT_STRING, BIT_STRING,0);
@@ -2692,13 +2763,17 @@ static void define_entity_raw(INT32 args)
     THIS->entities=sp[-1].u.mapping;
     sp--;
   }else{
+    /* FIXME: This looks broken. Shouldn't the stack be adjusted? */
     mapping_insert(THIS->entities, sp-2, sp-1);
   }
 }
 
+/*! @decl void define_entity(string entity, string raw, @
+ *!                          function cb, mixed ... extras)
+ */
 static void define_entity(INT32 args)
 {
-  struct svalue tmp;
+  int d;
   struct pike_string *s;
   struct xmldata data;
   ONERROR e;
@@ -2711,10 +2786,14 @@ static void define_entity(INT32 args)
   data.input.len=s->len;
   data.input.pos=0;
   data.input.to_free=0;
+  data.input.entity=0;
+  data.input.callbackinfo=allocate_mapping(0);
   data.func=sp+2-args;
   data.extra_args=sp+3-args;
   data.num_extra_args=args-3;
-  data.extra_arg_types = 0xffff; /* FIXME */
+  for(d=0;d<data.num_extra_args;d++)
+    data.extra_arg_types |= 1 << data.extra_args[d].type;
+
   data.allow_pesmeg_everywhere=0;
     
   SET_ONERROR(e,free_xmldata, &data);
@@ -2726,6 +2805,8 @@ static void define_entity(INT32 args)
   define_entity_raw(2);
 }
 
+/*! @decl void allow_rxml_entities(int(0..1) yes_no)
+ */
 static void allow_rxml_entities(INT32 args)
 {
   check_all_args("XML->allow_rxml_entities", args, BIT_INT, 0);
@@ -2734,6 +2815,8 @@ static void allow_rxml_entities(INT32 args)
   push_int(0);
 }
 
+/*! @decl mixed parse_dtd(string dtd, function cb, mixed ... extras)
+ */
 static void parse_dtd(INT32 args)
 {
   struct svalue tmp;
@@ -2750,6 +2833,8 @@ static void parse_dtd(INT32 args)
   data.input.len=s->len;
   data.input.pos=0;
   data.input.to_free=0;
+  data.input.entity=0;
+  data.input.callbackinfo=allocate_mapping(0);
   data.func=sp+1-args;
   data.extra_args=sp+2-args;
   data.num_extra_args=args-2;
@@ -2765,6 +2850,8 @@ static void parse_dtd(INT32 args)
   *sp++=tmp;
 }
 
+/*! @decl string autoconvert(string xml)
+ */
 static void autoconvert(INT32 args)
 {
   INT32 e;
@@ -2854,7 +2941,7 @@ static void autoconvert(INT32 args)
 
       case 0x3c3f786d: /* ASCII? UTF-8? ISO-8859? */
 	{
-	  int pos,e,encstart,encend;
+	  int pos,encstart;
 
 	  IF_XMLDEBUG(fprintf(stderr,"Extended ASCII detected (assuming UTF8).\n"));
 	  pos=5;
@@ -2971,9 +3058,19 @@ static void init_xml_struct(struct object *o)
   dmalloc_touch_svalue(sp);
 }
 
+/*! @endclass
+ */
+
+/*! @endmodule
+ */
+
 void init_xml(void)
 {
   ptrdiff_t off;
+  init_xmlinput_blocks();
+  push_text("location");
+  location_string_svalue=sp[-1];
+  sp--;
   start_new_program();
   off = ADD_STORAGE(struct xmlobj);
   map_variable("__entities", "mapping", ID_STATIC|ID_PRIVATE,
@@ -3029,4 +3126,5 @@ void init_xml(void)
 void exit_xml(void)
 {
   free_all_xmlinput_blocks();
+  free_svalue(&location_string_svalue);
 }
