@@ -1,6 +1,11 @@
 /*
- * $Id: acconfig.h,v 1.8 2003/05/20 15:12:31 mast Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: acconfig.h,v 1.14 2003/05/20 15:12:40 mast Exp $
+*/
+
+/*
  * System dependant definitions for the system module for Pike
  *
  * Henrik Grubbstrm 1997-01-20
@@ -42,9 +47,18 @@
 /* Define if the get/setrlimit syntax is unknown */
 #undef SETRLIMIT_SYNTAX_UNKNOWN
 
-/* Define if you have the <sys/resource.h> header file.  */
+/* Define if you have a working <sys/resource.h> header file.  */
 #undef HAVE_SYS_RESOURCE_H
 
+/* Define if you have a working <sys/user.h> header file.  */
+#undef HAVE_SYS_USER_H
+
+/* Define if ITIMER_REAL, ITIMER_VIRTUAL and ITIMER_PROF is in tInt02 */
+#undef ITIMER_TYPE_IS_02
+
+/* Define if you have the RDTSC instruction */
+#undef HAVE_RDTSC
+
 /* Define if syslog() might have races which makes it unsafe to run unlocked. */
 #undef MIGHT_HAVE_SYSLOG_RACES
 
