@@ -1,6 +1,11 @@
 /*
- * $Id: nt.c,v 1.38 2004/03/09 15:48:06 grubba Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: nt.c,v 1.57 2003/06/26 12:41:46 grubba Exp $
+*/
+
+/*
  * NT system calls for Pike
  *
  * Fredrik Hubinette, 1998-08-22
@@ -28,7 +33,8 @@
 
 /*
  * Get some wrappers for functions not implemented in old versions
- * of WIN32.
+ * of WIN32. Needs a Platform SDK installed. The SDK included in 
+ * MSVS 6.0 is not enough.
  */
 #define COMPILE_NEWAPIS_STUBS
 /* We want GetLongPathName()... */
@@ -47,6 +53,11 @@
 #include "operators.h"
 #include "stuff.h"
 
+#define sp Pike_sp
+
+/*! @module System
+ */
+
 static void throw_nt_error(char *funcname, int err)
 /*
  *  Give string equivalents to some of the more common NT error codes.
@@ -234,7 +245,7 @@ static void push_regvalue(DWORD type, char* buffer, DWORD len)
  * (On W2k/IA64 HKEY is typedefed to struct HKEY__ *).
  *
  * NOTE: Order must match the values specified with
- * ADD_GLOBAL_INTEGER_CONSTANT() in init_pike_module() below.
+ * ADD_GLOBAL_INTEGER_CONSTANT() in init_nt_system_calls() below.
  */
 static const HKEY hkeys[] = {
   HKEY_CLASSES_ROOT,
@@ -243,6 +254,36 @@ static const HKEY hkeys[] = {
   HKEY_USERS,
 };
 
+/*! @decl string|int|array(string) RegGetValue(int hkey, string key, @
+ *!                                            string index)
+ *!
+ *!   Get a single value from the register.
+ *!
+ *! @param hkey
+ *!   One of the following:
+ *!   @int
+ *!     @value HKEY_CLASSES_ROOT
+ *!     @value HKEY_LOCAL_MACHINE
+ *!     @value HKEY_CURRENT_USER
+ *!     @value HKEY_USERS
+ *!   @endint
+ *!
+ *! @param key
+ *!   Registry key.
+ *!
+ *! @param index
+ *!   Value name.
+ *!
+ *! @returns
+ *!   Returns the value stored at the specified location in the register
+ *!   if any. Throws errors on failure.
+ *!
+ *! @note
+ *!   This function is only available on Win32 systems.
+ *!
+ *! @seealso
+ *!   @[RegGetValues()], @[RegGetKeyNames()]
+ */
 void f_RegGetValue(INT32 args)
 {
   long ret;
@@ -280,6 +321,32 @@ static void do_regclosekey(HKEY key)
   RegCloseKey(key);
 }
 
+/*! @decl array(string) RegGetKeyNames(int hkey, string key)
+ *!
+ *!   Get a list of value key names from the register.
+ *!
+ *! @param hkey
+ *!   One of the following:
+ *!   @int
+ *!     @value HKEY_CLASSES_ROOT
+ *!     @value HKEY_LOCAL_MACHINE
+ *!     @value HKEY_CURRENT_USER
+ *!     @value HKEY_USERS
+ *!   @endint
+ *!
+ *! @param key
+ *!   Registry key.
+ *!
+ *! @returns
+ *!   Returns an array of value keys stored at the specified location iof any.
+ *!   Throws errors on failure.
+ *!
+ *! @note
+ *!   This function is only available on Win32 systems.
+ *!
+ *! @seealso
+ *!   @[RegGetValue()], @[RegGetValues()]
+ */
 void f_RegGetKeyNames(INT32 args)
 {
   INT_TYPE hkey_num;
@@ -329,6 +396,33 @@ void f_RegGetKeyNames(INT32 args)
   f_aggregate(i);
 }
 
+/*! @decl mapping(string:string|int|array(string)) RegGetValues(int hkey, @
+ *!                                                             string key)
+ *!
+ *!   Get multiple values from the register.
+ *!
+ *! @param hkey
+ *!   One of the following:
+ *!   @int
+ *!     @value HKEY_CLASSES_ROOT
+ *!     @value HKEY_LOCAL_MACHINE
+ *!     @value HKEY_CURRENT_USER
+ *!     @value HKEY_USERS
+ *!   @endint
+ *!
+ *! @param key
+ *!   Registry key.
+ *!
+ *! @returns
+ *!   Returns a mapping with all the values stored at the specified location
+ *!   in the register if any. Throws errors on failure.
+ *!
+ *! @note
+ *!   This function is only available on Win32 systems.
+ *!
+ *! @seealso
+ *!   @[RegGetValue()], @[RegGetKeyNames()]
+ */
 void f_RegGetValues(INT32 args)
 {
   INT_TYPE hkey_num;
@@ -386,7 +480,7 @@ void f_RegGetValues(INT32 args)
 
 static struct program *token_program;
 
-#define THIS_TOKEN (*(HANDLE *)(fp->current_storage))
+#define THIS_TOKEN (*(HANDLE *)(Pike_fp->current_storage))
 
 typedef BOOL (WINAPI *logonusertype)(LPSTR,LPSTR,LPSTR,DWORD,DWORD,PHANDLE);
 typedef DWORD (WINAPI *getlengthsidtype)(PSID);
@@ -415,7 +509,7 @@ LINKFUNC(BOOL,addauditaccessace, (PACL,DWORD,DWORD,PSID,BOOL,BOOL) );
 
 HINSTANCE advapilib;
 
-#define THIS_PSID (*(PSID *)fp->current_storage)
+#define THIS_PSID (*(PSID *)Pike_fp->current_storage)
 static struct program *sid_program;
 static void init_sid(struct object *o)
 {
@@ -501,7 +595,45 @@ static void f_sid_account(INT32 args)
   
 }
 
-
+/*! @decl object LogonUser(string username, string|int(0..0) domain, @
+ *!                        string password, int|void logon_type, @
+ *!                        int|void logon_provider)
+ *!
+ *!   Logon a user.
+ *!
+ *! @param username
+ *!   User name of the user to login.
+ *!
+ *! @param domain
+ *!   Domain to login on, or zero if local logon.
+ *!
+ *! @param password
+ *!   Password to login with.
+ *!
+ *! @param logon_type
+ *!   One of the following values:
+ *!   @int
+ *!     @value LOGON32_LOGON_BATCH
+ *!     @value LOGON32_LOGON_INTERACTIVE
+ *!     @value LOGON32_LOGON_SERVICE
+ *!
+ *!     @value LOGON32_LOGON_NETWORK
+ *!       This is the default.
+ *!   @endint
+ *!
+ *! @param logon_provider
+ *!   One of the following values:
+ *!   @int
+ *!     @value LOGON32_PROVIDER_DEFAULT
+ *!       This is the default.
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns a login token object on success, and zero on failure.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ */
 void f_LogonUser(INT32 args)
 {
   LPTSTR username, domain, pw;
@@ -950,6 +1082,46 @@ static netapibufferfreetype netapibufferfree;
 HINSTANCE netapilib;
 
 
+/*! @decl string|array(string|int) NetUserGetInfo(string username, @
+ *!                                               string|int(0..0) server, @
+ *!                                               int|void level)
+ *!
+ *!   Get information about a network user.
+ *!
+ *! @param username
+ *!   User name of the user to get information about.
+ *!
+ *! @param server
+ *!   Server the user exists on.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!     @value 2
+ *!     @value 3
+ *!     @value 10
+ *!     @value 11
+ *!     @value 20
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserEnum()], @[NetGroupEnum()]
+ *!   @[NetLocalGroupEnum()], @[NetUserGetGroups()],
+ *!   @[NetUserGetLocalGroups()], @[NetGroupGetUsers()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetUserGetInfo(INT32 args)
 {
   char *to_free1=NULL,*to_free2=NULL;
@@ -1006,13 +1178,50 @@ void f_NetUserGetInfo(INT32 args)
     throw_nt_error("NetGetUserInfo", ret);
 }
 
+/*! @decl array(string|array(string|int)) @
+ *!     NetUserEnum(string|int(0..0)|void server, @
+ *!                 int|void level, int|void filter)
+ *!
+ *!   Get information about network users.
+ *!
+ *! @param server
+ *!   Server the users exist on.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!     @value 2
+ *!     @value 3
+ *!     @value 10
+ *!     @value 11
+ *!     @value 20
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetGroupEnum()]
+ *!   @[NetLocalGroupEnum()], @[NetUserGetGroups()],
+ *!   @[NetUserGetLocalGroups()], @[NetGroupGetUsers()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetUserEnum(INT32 args)
 {
   char *to_free1=NULL;
   DWORD level=0;
   DWORD filter=0;
   LPWSTR server=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
   DWORD resume=0;
   NET_API_STATUS ret;
@@ -1048,7 +1257,7 @@ void f_NetUserEnum(INT32 args)
 
   do
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e;
     LPBYTE buf=0,ptr;
 
     THREADS_ALLOW();
@@ -1090,12 +1299,44 @@ void f_NetUserEnum(INT32 args)
 }
 
 
+/*! @decl array(string|array(string|int)) @
+ *!     NetGroupEnum(string|int(0..0)|void server, int|void level)
+ *!
+ *!   Get information about network groups.
+ *!
+ *! @param server
+ *!   Server the groups exist on.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!     @value 2
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetLocalGroupEnum()], @[NetUserGetGroups()],
+ *!   @[NetUserGetLocalGroups()], @[NetGroupGetUsers()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetGroupEnum(INT32 args)
 {
   char *to_free1=NULL, *tmp_server=NULL;
   DWORD level=0;
   LPWSTR server=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
   DWORD resume=0;
   NET_API_STATUS ret;
@@ -1120,7 +1361,7 @@ void f_NetGroupEnum(INT32 args)
 
   do
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e;
     LPBYTE buf=0,ptr;
 
     THREADS_ALLOW();
@@ -1161,12 +1402,43 @@ void f_NetGroupEnum(INT32 args)
   if(to_free1) free(to_free1);
 }
 
+/*! @decl array(array(string|int)) @
+ *!     NetLocalGroupEnum(string|int(0..0)|void server, int|void level)
+ *!
+ *!   Get information about local network groups.
+ *!
+ *! @param server
+ *!   Server the groups exist on.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetUserGetGroups()],
+ *!   @[NetUserGetLocalGroups()], @[NetGroupGetUsers()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetLocalGroupEnum(INT32 args)
 {
   char *to_free1=NULL, *tmp_server=NULL;
   DWORD level=0;
   LPWSTR server=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
   DWORD resume=0;
   NET_API_STATUS ret;
@@ -1191,7 +1463,7 @@ void f_NetLocalGroupEnum(INT32 args)
 
   do
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e;
     LPBYTE buf=0,ptr;
 
     THREADS_ALLOW();
@@ -1231,15 +1503,49 @@ void f_NetLocalGroupEnum(INT32 args)
   if(to_free1) free(to_free1);
 }
 
+/*! @decl array(array(string|int)) @
+ *!     NetUserGetGroups(string|int(0..0) server, string user, int|void level)
+ *!
+ *!   Get information about group membership for a network user.
+ *!
+ *! @param server
+ *!   Server the groups exist on.
+ *!
+ *! @param user
+ *!   User to retrieve groups for.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetLocalGroupEnum()],
+ *!   @[NetUserGetLocalGroups()], @[NetGroupGetUsers()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetUserGetGroups(INT32 args)
 {
-  char *to_free1=NULL, *to_free2=NULL, *tmp_server=NULL, *tmp_user;
+  char *to_free1=NULL, *to_free2=NULL, *tmp_server=NULL;
   DWORD level=0;
   LPWSTR server=NULL;
   LPWSTR user=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
-  DWORD read=0, total=0;
+  DWORD read=0, total=0, e;
   NET_API_STATUS ret;
   LPBYTE buf=0,ptr;
 
@@ -1302,16 +1608,56 @@ void f_NetUserGetGroups(INT32 args)
   }
 }
 
+/*! @decl array(string) @
+ *!     NetUserGetLocalGroups(string|int(0..0) server, @
+ *!                           string user, int|void level, int|void flags)
+ *!
+ *!   Get information about group membership for a local network user.
+ *!
+ *! @param server
+ *!   Server the groups exist on.
+ *!
+ *! @param user
+ *!   User to retrieve groups for.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!   @endint
+ *!
+ *! @param flags
+ *!   Zero, of one of the following:
+ *!   @int
+ *!     @value LG_INCLUDE_INDIRECT
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetLocalGroupEnum()],
+ *!   @[NetUserGetGroups()], @[NetGroupGetUsers()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetUserGetLocalGroups(INT32 args)
 {
-  char *to_free1=NULL, *to_free2=NULL, *tmp_server=NULL, *tmp_user;
+  char *to_free1=NULL, *to_free2=NULL, *tmp_server=NULL;
   DWORD level=0;
   DWORD flags=0;
   LPWSTR server=NULL;
   LPWSTR user=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
-  DWORD read=0, total=0;
+  DWORD read=0, total=0, e;
   NET_API_STATUS ret;
   LPBYTE buf=0,ptr;
 
@@ -1379,13 +1725,48 @@ void f_NetUserGetLocalGroups(INT32 args)
   if(to_free2) free(to_free2);
 }
 
+/*! @decl array(string|array(int|string)) @
+ *!     NetGroupGetUsers(string|int(0..0) server, @
+ *!                      string group, int|void level)
+ *!
+ *!   Get information about group membership for a network group.
+ *!
+ *! @param server
+ *!   Server the groups exist on.
+ *!
+ *! @param group
+ *!   Group to retrieve members for.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetLocalGroupEnum()],
+ *!   @[NetUserGetGroups()], @[NetUserGetLocalGroups()],
+ *!   @[NetLocalGroupGetMembers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetGroupGetUsers(INT32 args)
 {
   char *to_free1=NULL, *to_free2=NULL, *tmp_server=NULL;
   DWORD level=0;
   LPWSTR server=NULL;
   LPWSTR group=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
   DWORD resume=0;
 
@@ -1412,7 +1793,7 @@ void f_NetGroupGetUsers(INT32 args)
 
   while(1)
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e;
     NET_API_STATUS ret;
     LPBYTE buf=0,ptr;
 
@@ -1459,13 +1840,50 @@ void f_NetGroupGetUsers(INT32 args)
   if(to_free2) free(to_free2);
 }
 
+/*! @decl array(string|array(int|string)) @
+ *!     NetLocalGroupGetMembers(string|int(0..0) server, @
+ *!                             string group, int|void level)
+ *!
+ *!   Get information about group membership for a network group.
+ *!
+ *! @param server
+ *!   Server the groups exist on.
+ *!
+ *! @param group
+ *!   Group to retrieve members for.
+ *!
+ *! @param level
+ *!   Information level. One of:
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!     @value 2
+ *!     @value 3
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns an array on success. Throws errors on failure.
+ *!
+ *! @fixme
+ *!   Document the return value.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetLocalGroupEnum()],
+ *!   @[NetUserGetGroups()], @[NetUserGetLocalGroups()],
+ *!   @[NetGroupGetUsers()], @[NetGetDCName()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetLocalGroupGetMembers(INT32 args)
 {
   char *to_free1=NULL, *to_free2=NULL, *tmp_server=NULL;
   DWORD level=0;
   LPWSTR server=NULL;
   LPWSTR group=NULL;
-  INT32 pos=0,e;
+  INT32 pos=0;
   struct array *a=0;
   DWORD resume=0;
 
@@ -1492,7 +1910,7 @@ void f_NetLocalGroupGetMembers(INT32 args)
 
   while(1)
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e;
     NET_API_STATUS ret;
     LPBYTE buf=0,ptr;
 
@@ -1546,11 +1964,34 @@ void f_NetLocalGroupGetMembers(INT32 args)
   if(to_free2) free(to_free2);
 }
 
+/*! @decl string NetGetDCName(string|int(0..0) server, string domain)
+ *!
+ *!   Get name of the domain controller from a server.
+ *!
+ *! @param server
+ *!   Server the domain exists on.
+ *!
+ *! @param domain
+ *!   Domain to get the domain controller for.
+ *!
+ *! @returns
+ *!   Returns the name of the domain controller on success.
+ *!   Throws errors on failure.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetLocalGroupEnum()],
+ *!   @[NetUserGetGroups()], @[NetUserGetLocalGroups()],
+ *!   @[NetGroupGetUsers()], @[NetLocalGroupGetMembers()],
+ *!   @[NetGetAnyDCName()]
+ */
 void f_NetGetDCName(INT32 args)
 {
   char *to_free1=NULL,*to_free2=NULL;
   BYTE *tmp=0;
-  DWORD level;
   LPWSTR server, domain;
   NET_API_STATUS ret;
 
@@ -1592,11 +2033,34 @@ void f_NetGetDCName(INT32 args)
   }
 }
 
+/*! @decl string NetGetAnyDCName(string|int(0..0) server, string domain)
+ *!
+ *!   Get name of a domain controller from a server.
+ *!
+ *! @param server
+ *!   Server the domain exists on.
+ *!
+ *! @param domain
+ *!   Domain to get a domain controller for.
+ *!
+ *! @returns
+ *!   Returns the name of a domain controller on success.
+ *!   Throws errors on failure.
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[NetUserGetInfo()], @[NetUserEnum()],
+ *!   @[NetGroupEnum()], @[NetLocalGroupEnum()],
+ *!   @[NetUserGetGroups()], @[NetUserGetLocalGroups()],
+ *!   @[NetGroupGetUsers()], @[NetLocalGroupGetMembers()],
+ *!   @[NetGetDCName()]
+ */
 void f_NetGetAnyDCName(INT32 args)
 {
   char *to_free1=NULL,*to_free2=NULL;
   BYTE *tmp=0;
-  DWORD level;
   LPWSTR server, domain;
   NET_API_STATUS ret;
 
@@ -1805,10 +2269,28 @@ static int sizeof_wkstauser_info(int level)
   }
 }
 
-
+/*! @decl array(int|string) NetSessionEnum(string|int(0..0) server, @
+ *!                                        string|int(0..0) client, @
+ *!                                        string|int(0..0) user, int level)
+ *!
+ *!   Get session information.
+ *!
+ *! @param level
+ *!   One of
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!     @value 2
+ *!     @value 10
+ *!     @value 502
+ *!   @endint
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ */
 static void f_NetSessionEnum(INT32 args)
 {
-  INT32 pos=0,e;
   LPWSTR server, client, user;
   DWORD level;
   DWORD resume = 0;
@@ -1838,7 +2320,7 @@ static void f_NetSessionEnum(INT32 args)
   
   while(1)
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e, pos = 0;
     NET_API_STATUS ret;
     LPBYTE buf=0,ptr;
 
@@ -1885,9 +2367,21 @@ static void f_NetSessionEnum(INT32 args)
 
 /* End netsessionenum */
 
+/*! @decl array(mixed) NetWkstaUserEnum(string|int(0..0) server, int level)
+ *!
+ *! @param level
+ *!   One of
+ *!   @int
+ *!     @value 0
+ *!     @value 1
+ *!   @endint
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ */
 static void f_NetWkstaUserEnum(INT32 args)
 {
-  INT32 pos=0,e;
   LPWSTR server;
   DWORD level;
   DWORD resume = 0;
@@ -1906,7 +2400,7 @@ static void f_NetWkstaUserEnum(INT32 args)
   
   while(1)
   {
-    DWORD read=0, total=0;
+    DWORD read=0, total=0, e, pos = 0;
     NET_API_STATUS ret;
     LPBYTE buf=0,ptr;
 
@@ -1954,7 +2448,7 @@ static void f_NetWkstaUserEnum(INT32 args)
  *!   Normalize an NT filesystem path.
  *!
  *!   The following transformations are currently done:
- *!   @dl
+ *!   @ul
  *!     @item
  *!       Trailing slashes are removed.
  *!     @item
@@ -1963,10 +2457,14 @@ static void f_NetWkstaUserEnum(INT32 args)
  *!       Short filenames are expanded to their corresponding long
  *!       variants.
  *!     @item
+ *!       Forward slashes ('/') are converted to backward slashes ('\').
+ *!     @item
+ *!       Current- and parent-directory paths are removed ("." and "..").
+ *!     @item
  *!       Relative paths are expanded to absolute paths.
  *!     @item
  *!       Case-information is restored.
- *!   @enddl
+ *!   @endul
  *!
  *! @returns
  *!   A normalized absolute path without trailing slashes.
@@ -2032,6 +2530,16 @@ static void f_normalize_path(INT32 args)
   push_string(finish_string_builder(&res));
 }
 
+/*! @decl int GetFileAttributes(string filename)
+ *!
+ *!   Get the file attributes for the specified file.
+ *!
+ *! @note
+ *!   This function is only available on Win32 systems.
+ *!
+ *! @seealso
+ *!   @[SetFileAttributes()]
+ */
 static void f_GetFileAttributes(INT32 args)
 {
   char *file;
@@ -2043,6 +2551,16 @@ static void f_GetFileAttributes(INT32 args)
   push_int(ret);
 }
 
+/*! @decl int SetFileAttributes(string filename)
+ *!
+ *!   Set the file attributes for the specified file.
+ *!
+ *! @note
+ *!   This function is only available on Win32 systems.
+ *!
+ *! @seealso
+ *!   @[GetFileAttributes()]
+ */
 static void f_SetFileAttributes(INT32 args)
 {
   char *file;
@@ -2056,6 +2574,12 @@ static void f_SetFileAttributes(INT32 args)
   push_int(ret);
 }
 
+/*! @decl array(mixed) LookupAccountName(string|int(0..0) sys, string account)
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ */
 static void f_LookupAccountName(INT32 args)
 {
   LPCTSTR sys=0;
@@ -2125,8 +2649,8 @@ static struct array *encode_acl(PACL acl)
   ACL_SIZE_INFORMATION tmp;
   if(GetAclInformation(acl, &tmp, sizeof(tmp), AclSizeInformation))
   {
-    int e;
-    check_stack(tmp.AceCount + 4);
+    unsigned int e;
+    check_stack(((ptrdiff_t)(tmp.AceCount + 4)));
     for(e=0;e<tmp.AceCount;e++)
     {
       void *ace;
@@ -2255,6 +2779,15 @@ static PACL decode_acl(struct array *arr)
  * Note, this function does not use errno!!
  * (Time to learn how to autodoc... /Hubbe)
  */
+/*! @decl array(mixed) SetNamedSecurityInfo(string name, @
+ *!                                         mapping(string:mixed) options)
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[GetNamedSecurityInfo()]
+ */
 static void f_SetNamedSecurityInfo(INT32 args)
 {
   struct svalue *sval;
@@ -2324,6 +2857,16 @@ static void f_SetNamedSecurityInfo(INT32 args)
   push_int(ret);
 }
 
+/*! @decl mapping(mixed:mixed) GetNamedSecurityInfo(string name, @
+ *!                                                 int|void type, @
+ *!                                                 int|void flags)
+ *!
+ *! @note
+ *!   This function is only available on some Win32 systems.
+ *!
+ *! @seealso
+ *!   @[SetNamedSecurityInfo()]
+ */
 static void f_GetNamedSecurityInfo(INT32 args)
 {
   PSID owner=0, group=0;
@@ -2390,6 +2933,189 @@ static void f_GetNamedSecurityInfo(INT32 args)
   if(desc) LocalFree(desc);
 }
 
+/* Implementation of uname() for Win32. */
+static void f_nt_uname(INT32 args)
+{
+  /* More info could be added here */
+
+  char buf[1024];
+  char *machine = "unknown";
+  OSVERSIONINFO osversion;
+  SYSTEM_INFO sysinfo;
+  int n=0;
+
+  pop_n_elems(args);
+
+  osversion.dwOSVersionInfoSize=sizeof(osversion);
+  if(!GetVersionEx(&osversion))
+    Pike_error("GetVersionEx failed with errno %d\n",GetLastError());
+
+  GetSystemInfo(&sysinfo);
+
+  n+=2;
+  push_text("architecture");
+  switch(sysinfo.wProcessorArchitecture)
+  {
+    case PROCESSOR_ARCHITECTURE_INTEL:
+      sprintf(buf, "i%d", sysinfo.dwProcessorType);
+      push_text(buf);
+      machine = "i86pc";
+      break;
+
+    case PROCESSOR_ARCHITECTURE_MIPS:
+      push_text("mips");
+      machine = "mips";
+      break;
+
+    case PROCESSOR_ARCHITECTURE_ALPHA:
+      push_text("alpha");
+      machine = "alpha";
+      break;
+
+    case PROCESSOR_ARCHITECTURE_PPC:
+      push_text("ppc");
+      machine = "ppc";
+      break;
+
+#ifdef PROCESSOR_ARCHITECTURE_SHX
+    case PROCESSOR_ARCHITECTURE_SHX:
+      machine = "shx";
+      switch (sysinfo.dwProcessorType) {
+      case PROCESSOR_HITACHI_SH3:
+      case PROCESSOR_SHx_SH3:
+        push_text("sh3");
+	break;
+      case PROCESSOR_HITACHI_SH3E:
+        push_text("sh3e");
+	break;
+      case PROCESSOR_HITACHI_SH4:
+      case PROCESSOR_SHx_SH4:
+        push_text("sh4");
+	break;
+      default:
+	push_text("shx");
+	break;
+      }
+      break;
+#endif /* PROCESSOR_ARCHITECTURE_SHX */
+
+#ifdef PROCESSOR_ARCHITECTURE_ARM
+    case PROCESSOR_ARCHITECTURE_ARM:
+      machine = "arm";
+      switch (sysinfo.dwProcessorType) {
+      case PROCESSOR_STRONGARM:
+	push_text("strongarm");
+	break;
+      case PROCESSOR_ARM720:
+	push_text("arm720");
+	break;
+      case PROCESSOR_ARM820:
+	push_text("arm820");
+	break;
+      case PROCESSOR_ARM920:
+	push_text("arm920");
+	break;
+      case PROCESSOR_ARM_7TDMI:
+	push_text("arm7tdmi");
+	break;
+      default:
+	push_text("arm");
+	break;
+      }
+      break;
+#endif /* PROCESSOR_ARCHITECTURE_ARM */
+
+#ifdef PROCESSOR_ARCHITECTURE_IA64
+    case PROCESSOR_ARCHITECTURE_IA64:
+      machine = "ia64";
+      push_text("ia64");
+      break;
+#endif /* PROCESSOR_ARCHITECTURE_IA64 */
+
+#ifdef PROCESSOR_ARCHITECTURE_ALPHA64
+    case PROCESSOR_ARCHITECTURE_ALPHA64:
+      machine = "alpha64";
+      push_text("alpha64");
+      break;
+#endif /* PROCESSOR_ARCHITECTURE_ALPHA64 */
+
+#ifdef PROCESSOR_ARCHITECTURE_MSIL
+    case PROCESSOR_ARCHITECTURE_MSIL:
+      machine = "msil";
+      push_text("msil");
+      break;
+#endif /* PROCESSOR_ARCHITECTURE_MSIL */
+
+    default:
+    case PROCESSOR_ARCHITECTURE_UNKNOWN:
+      machine = "unknown";
+      push_text("unknown");
+      break;
+  }
+
+  n+=2;
+  push_text("machine");
+  push_text(machine);
+
+  n+=2;
+  push_text("sysname");
+  switch(osversion.dwPlatformId)
+  {
+    case VER_PLATFORM_WIN32s:
+      push_text("Win32s");
+      break;
+
+    case VER_PLATFORM_WIN32_WINDOWS:
+      push_text("Win32");
+      if(osversion.dwMajorVersion)
+      {
+	sprintf(buf,"Windows 95 %d.%d.%d",
+		osversion.dwMajorVersion,
+		osversion.dwMinorVersion,
+		osversion.dwBuildNumber & 0xffff);
+		
+      }else{
+	sprintf(buf,"Windows 98 %d.%d.%d",
+		osversion.dwMajorVersion,
+		osversion.dwMinorVersion,
+		osversion.dwBuildNumber & 0xffff);
+		
+      }
+      break;
+
+    case VER_PLATFORM_WIN32_NT:
+      push_text("Win32");
+      sprintf(buf,"Windows NT %d.%d.%d",
+	      osversion.dwMajorVersion,
+	      osversion.dwMinorVersion,
+	      osversion.dwBuildNumber);
+      break;
+
+    default:
+      push_text("Win32");
+      sprintf(buf,"Windows ?? %d.%d.%d",
+	      osversion.dwMajorVersion,
+	      osversion.dwMinorVersion,
+	      osversion.dwBuildNumber);
+      break;
+  }
+
+  n+=2;
+  push_text("release");
+  push_text(buf);
+
+  n+=2;
+  push_text("version");
+  push_text(osversion.szCSDVersion);
+
+  n+=2;
+  push_text("nodename");
+  gethostname(buf, sizeof(buf));
+  push_text(buf);
+
+  f_aggregate_mapping(n);
+}
+
 
 /**************************/
 /* start Security context */
@@ -2428,7 +3154,7 @@ struct sctx_storage {
   int        lastError;
 };
 
-#define THIS_SCTX ((struct sctx_storage *)fp->current_storage)
+#define THIS_SCTX ((struct sctx_storage *)Pike_fp->current_storage)
 static struct program *sctx_program;
 static void init_sctx(struct object *o)
 {
@@ -2783,6 +3509,9 @@ void init_nt_system_calls(void)
   ADD_EFUN("RegGetKeyNames", f_RegGetKeyNames, tFunc(tInt tStr, tArr(tStr)),
 	   OPT_EXTERNAL_DEPEND);
 
+  ADD_EFUN("uname", f_nt_uname,tFunc(tNone,tMapping), OPT_TRY_OPTIMIZE);
+  ADD_FUNCTION2("uname", f_nt_uname,tFunc(tNone,tMapping), 0, OPT_TRY_OPTIMIZE);
+
   ADD_FUNCTION("normalize_path", f_normalize_path, tFunc(tStr, tStr), 0);
 
   /* LogonUser only exists on NT, link it dynamically */
@@ -3129,3 +3858,5 @@ void exit_nt_system_calls(void)
 
 #endif
 
+/*! @endmodule
+ */
