@@ -1,6 +1,11 @@
 /*
- * $Id: system.h,v 1.4 1998/03/03 10:51:38 mast Exp $
- *
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: system.h,v 1.9 2003/10/10 19:12:53 nilsson Exp $
+*/
+
+/*
  * Prototypes for the Pike system-module
  *
  * Henrik Grubbstrm 1997-01-28
@@ -42,3 +47,21 @@ void f_syslog(INT32 args);
 void f_closelog(INT32 args);
 
 #endif /* PIKE_MODULES_SYSTEM_H */
+
+/*
+ * memory.c
+ */
+
+struct memory_storage
+{
+   unsigned char *p;
+   size_t size;
+   
+#define MEM_READ        0x01
+#define MEM_WRITE       0x02
+#define MEM_FREE_FREE   0x10
+#define MEM_FREE_MUNMAP 0x20   
+#define MEM_FREE_SHMDEL 0x40
+   unsigned long flags;
+   void *extra;
+};
