@@ -1,9 +1,10 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-/**/
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: opcodes.c,v 1.130 2003/08/26 17:36:33 grubba Exp $
+*/
+
 #include "global.h"
 #include <errno.h>
 #include <math.h>
@@ -27,7 +28,9 @@
 #include "bignum.h"
 #include "operators.h"
 
-RCSID("$Id: opcodes.c,v 1.101 2001/05/28 18:11:50 marcus Exp $");
+#define sp Pike_sp
+
+RCSID("$Id: opcodes.c,v 1.130 2003/08/26 17:36:33 grubba Exp $");
 
 void index_no_free(struct svalue *to,struct svalue *what,struct svalue *ind)
 {
@@ -50,7 +53,12 @@ void index_no_free(struct svalue *to,struct svalue *what,struct svalue *ind)
       index_no_free(to, what, ind);
       if(IS_UNDEFINED(to)) {
 	if (val) {
-	  Pike_error("Indexing the integer %d with an unknown method.\n", val);
+	  if (ind->type == T_STRING)
+	    Pike_error("Indexing the integer %"PRINTPIKEINT"d "
+		       "with unknown method \"%s\".\n", val, ind->u.string->str);
+	  else
+	    Pike_error("Indexing the integer %"PRINTPIKEINT"d "
+		       "with an unknown method.\n", val);
 	} else {
           if(ind->type == T_STRING)
             Pike_error("Indexing the NULL value with \"%s\".\n", ind->u.string->str);
@@ -77,7 +85,7 @@ void index_no_free(struct svalue *to,struct svalue *what,struct svalue *ind)
   case T_MULTISET:
     i=multiset_member(what->u.multiset, ind);
     to->type=T_INT;
-    to->subtype=i ? NUMBER_UNDEFINED : 0;
+    to->subtype=i ? 0 : NUMBER_UNDEFINED;
     to->u.integer=i;
     break;
 
@@ -133,10 +141,247 @@ void o_index(void)
   sp++;
 }
 
-void o_cast(struct pike_string *type, INT32 run_time_type)
+/* Special case for casting to int. */
+void o_cast_to_int(void)
 {
-  INT32 i;
+  switch(sp[-1].type)
+  {
+  case T_OBJECT:
+    {
+      struct pike_string *s;
+      MAKE_CONSTANT_SHARED_STRING(s, "int");
+      push_string(s);
+      if(!sp[-2].u.object->prog)
+	Pike_error("Cast called on destructed object.\n");
+      if(FIND_LFUN(sp[-2].u.object->prog,LFUN_CAST) == -1)
+	Pike_error("No cast method in object.\n");
+      apply_lfun(sp[-2].u.object, LFUN_CAST, 1);
+      free_svalue(sp-2);
+      sp[-2]=sp[-1];
+      sp--;
+      dmalloc_touch_svalue(sp);
+    }
+    if(sp[-1].type != PIKE_T_INT)
+    {
+      if(sp[-1].type == T_OBJECT && sp[-1].u.object->prog)
+      {
+	int f=FIND_LFUN(sp[-1].u.object->prog, LFUN__IS_TYPE);
+	if( f != -1)
+	{
+	  struct pike_string *s;
+	  MAKE_CONSTANT_SHARED_STRING(s, "int");
+	  push_string(s);
+	  apply_low(sp[-2].u.object, f, 1);
+	  f=!UNSAFE_IS_ZERO(sp-1);
+	  pop_stack();
+	  if(f) return;
+	}
+      }
+      Pike_error("Cast failed, wanted int, got %s\n",
+		 get_name_of_type(sp[-1].type));
+    }
+    break;
 
+  case T_FLOAT:
+    {
+      int i=DO_NOT_WARN((int)(sp[-1].u.float_number));
+#ifdef AUTO_BIGNUM
+      if((i < 0 ? -i : i) < floor(fabs(sp[-1].u.float_number)))
+      {
+	/* Note: This includes the case when i = 0x80000000, i.e.
+	   the absolute value is not computable. */
+	convert_stack_top_to_bignum();
+	return;   /* FIXME: OK to return? Cast tests below indicates
+		     we have to do this, at least for now... /Noring */
+	/* Yes, it is ok to return, it is actually an optimization :)
+	 * /Hubbe
+	 */
+      }
+      else
+#endif /* AUTO_BIGNUM */
+      {
+	sp[-1].type=T_INT;
+	sp[-1].u.integer=i;
+      }
+    }
+    break;
+      
+  case T_STRING:
+    /* This can be here independently of AUTO_BIGNUM. Besides,
+       we really want to reduce the number of number parsers
+       around here. :) /Noring */
+#ifdef AUTO_BIGNUM
+
+    /* The generic function is rather slow, so I added this
+     * code for benchmark purposes. :-) /per
+     */
+    if( sp[-1].u.string->len < 10 &&
+	!sp[-1].u.string->size_shift )
+    {
+      int i=atoi(sp[-1].u.string->str);
+      free_string(sp[-1].u.string);
+      sp[-1].type=T_INT;
+      sp[-1].u.integer=i;
+    }
+    else
+      convert_stack_top_string_to_inumber(10);
+    return;   /* FIXME: OK to return? Cast tests below indicates
+		 we have to do this, at least for now... /Noring */
+    /* Yes, it is ok to return, it is actually an optimization :)
+     * /Hubbe
+     */
+#else
+    {
+      int i=STRTOL(sp[-1].u.string->str,0,10);
+      free_string(sp[-1].u.string);
+      sp[-1].type=T_INT;
+      sp[-1].u.integer=i;
+    }
+#endif /* AUTO_BIGNUM */
+    break;
+
+  case PIKE_T_INT:
+    break;
+	    
+  default:
+    Pike_error("Cannot cast %s to int.\n", get_name_of_type(sp[-1].type));
+  }
+}
+
+void o_cast_to_string(void)
+{
+  char buf[200];
+  switch(sp[-1].type)
+  {
+  case PIKE_T_STRING:
+    return;
+
+  case T_OBJECT:
+    {
+      struct pike_string *s;
+      MAKE_CONSTANT_SHARED_STRING(s, "string");
+      push_string(s);
+      if(!sp[-2].u.object->prog)
+	Pike_error("Cast called on destructed object.\n");
+      if(FIND_LFUN(sp[-2].u.object->prog,LFUN_CAST) == -1)
+	Pike_error("No cast method in object.\n");
+      apply_lfun(sp[-2].u.object, LFUN_CAST, 1);
+      free_svalue(sp-2);
+      sp[-2]=sp[-1];
+      sp--;
+      dmalloc_touch_svalue(sp);
+    }
+    if(sp[-1].type != PIKE_T_STRING)
+    {
+      if(sp[-1].type == T_OBJECT && sp[-1].u.object->prog)
+      {
+	int f=FIND_LFUN(sp[-1].u.object->prog, LFUN__IS_TYPE);
+	if( f != -1)
+	{
+	  struct pike_string *s;
+	  MAKE_CONSTANT_SHARED_STRING(s, "string");
+	  push_string(s);
+	  apply_low(sp[-2].u.object, f, 1);
+	  f=!UNSAFE_IS_ZERO(sp-1);
+	  pop_stack();
+	  if(f) return;
+	}
+      }
+      Pike_error("Cast failed, wanted string, got %s\n",
+		 get_name_of_type(sp[-1].type));
+    }
+    return;
+
+  case T_ARRAY:
+    {
+      int i;
+      struct array *a = sp[-1].u.array;
+      struct pike_string *s;
+      int shift = 0;
+
+      for(i = a->size; i--; ) {
+	unsigned INT32 val;
+	if (a->item[i].type != T_INT) {
+	  Pike_error("cast: Item %d is not an integer.\n", i);
+	}
+	val = (unsigned INT32)a->item[i].u.integer;
+	if (val > 0xff) {
+	  shift = 1;
+	  if (val > 0xffff) {
+	    shift = 2;
+	    while(i--)
+	      if (a->item[i].type != T_INT)
+		Pike_error("cast: Item %d is not an integer.\n", i);
+	    break;
+	  }
+	  while(i--) {
+	    if (a->item[i].type != T_INT) {
+	      Pike_error("cast: Item %d is not an integer.\n", i);
+	    }
+	    val = (unsigned INT32)a->item[i].u.integer;
+	    if (val > 0xffff) {
+	      shift = 2;
+	      while(i--)
+		if (a->item[i].type != T_INT)
+		  Pike_error("cast: Item %d is not an integer.\n", i);
+	      break;
+	    }
+	  }
+	  break;
+	}
+      }
+      s = begin_wide_shared_string(a->size, shift);
+      switch(shift) {
+      case 0:
+	for(i = a->size; i--; ) {
+	  s->str[i] = a->item[i].u.integer;
+	}
+	break;
+      case 1:
+	{
+	  p_wchar1 *str1 = STR1(s);
+	  for(i = a->size; i--; ) {
+	    str1[i] = a->item[i].u.integer;
+	  }
+	}
+	break;
+      case 2:
+	{
+	  p_wchar2 *str2 = STR2(s);
+	  for(i = a->size; i--; ) {
+	    str2[i] = a->item[i].u.integer;
+	  }
+	}
+	break;
+      default:
+	free_string(end_shared_string(s));
+	Pike_fatal("cast: Bad shift: %d.\n", shift);
+	break;
+      }
+      s = end_shared_string(s);
+      pop_stack();
+      push_string(s);
+    }
+    return;
+	    
+  case T_INT:
+    sprintf(buf, "%ld", (long)sp[-1].u.integer);
+    break;
+	    
+  case T_FLOAT:
+    sprintf(buf, "%f", (double)sp[-1].u.float_number);
+    break;
+
+  default:
+    Pike_error("Cannot cast %s to string.\n", get_name_of_type(sp[-1].type));
+  }
+	
+  sp[-1].type = PIKE_T_STRING;
+  sp[-1].u.string = make_shared_string(buf);
+}
+
+void o_cast(struct pike_type *type, INT32 run_time_type)
+{
   if(run_time_type != sp[-1].type)
   {
     if(run_time_type == T_MIXED)
@@ -238,56 +483,14 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	}
 	break;
 	
-      case T_INT:
-	switch(sp[-1].type)
-	{
-	  case T_FLOAT:
-	    i=DO_NOT_WARN((int)(sp[-1].u.float_number));
-#ifdef AUTO_BIGNUM
-	    if((i < 0 ? -i : i) < floor(fabs(sp[-1].u.float_number)))
-	    {
-	      /* Note: This includes the case when i = 0x80000000, i.e.
-		 the absolute value is not computable. */
-	      convert_stack_top_to_bignum();
-	      return;   /* FIXME: OK to return? Cast tests below indicates
-			   we have to do this, at least for now... /Noring */
-	      /* Yes, it is ok to return, it is actually an optimization :)
-	       * /Hubbe
-	       */
-	    }
-	    else
-#endif /* AUTO_BIGNUM */
-	    {
-	      sp[-1].type=T_INT;
-	      sp[-1].u.integer=i;
-	    }
-	    break;
-	    
-	  case T_STRING:
-	    /* This can be here independently of AUTO_BIGNUM. Besides,
-	       we really want to reduce the number of number parsers
-	       around here. :) /Noring */
-#ifdef AUTO_BIGNUM
-	    convert_stack_top_string_to_inumber(10);
-	    return;   /* FIXME: OK to return? Cast tests below indicates
-			 we have to do this, at least for now... /Noring */
-	      /* Yes, it is ok to return, it is actually an optimization :)
-	       * /Hubbe
-	       */
-#else
-	    i=STRTOL(sp[-1].u.string->str,0,10);
-	    free_string(sp[-1].u.string);
-	    sp[-1].type=T_INT;
-	    sp[-1].u.integer=i;
-#endif /* AUTO_BIGNUM */
-	    break;
-	    
-	  default:
-	    Pike_error("Cannot cast %s to int.\n",get_name_of_type(sp[-1].type));
-	}
-	
-	break;
+    case T_INT:
+      o_cast_to_int();
+      return;
 	
+    case T_STRING:
+      o_cast_to_string();
+      return;
+
       case T_FLOAT:
       {
 	FLOAT_TYPE f = 0.0;
@@ -299,8 +502,10 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	    break;
 	    
 	  case T_STRING:
-	    f=STRTOD_PCHARP(MKPCHARP(sp[-1].u.string->str,
-				     sp[-1].u.string->size_shift),0);
+	    f =
+	      (FLOAT_TYPE)STRTOD_PCHARP(MKPCHARP(sp[-1].u.string->str,
+						 sp[-1].u.string->size_shift),
+					0);
 	    free_string(sp[-1].u.string);
 	    break;
 	    
@@ -313,114 +518,21 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	break;
       }
       
-      case T_STRING:
-      {
-	char buf[200];
-	switch(sp[-1].type)
-	{
-	  case T_INT:
-	    sprintf(buf,"%ld",(long)sp[-1].u.integer);
-	    break;
-	    
-	  case T_FLOAT:
-	    sprintf(buf,"%f",(double)sp[-1].u.float_number);
-	    break;
-
-	  case T_ARRAY:
-	    {
-	      int i;
-	      struct array *a = sp[-1].u.array;
-	      struct pike_string *s;
-	      int shift = 0;
-
-	      for(i = a->size; i--; ) {
-		unsigned INT32 val;
-		if (a->item[i].type != T_INT) {
-		  Pike_error("cast: Item %d is not an integer.\n", i);
-		}
-		val = (unsigned INT32)a->item[i].u.integer;
-		if (val > 0xff) {
-		  shift = 1;
-		  if (val > 0xffff) {
-		    shift = 2;
-		    while(i--)
-		      if (a->item[i].type != T_INT)
-			Pike_error("cast: Item %d is not an integer.\n", i);
-		    break;
-		  }
-		  while(i--) {
-		    if (a->item[i].type != T_INT) {
-		      Pike_error("cast: Item %d is not an integer.\n", i);
-		    }
-		    val = (unsigned INT32)a->item[i].u.integer;
-		    if (val > 0xffff) {
-		      shift = 2;
-		      while(i--)
-			if (a->item[i].type != T_INT)
-			  Pike_error("cast: Item %d is not an integer.\n", i);
-		      break;
-		    }
-		  }
-		  break;
-		}
-	      }
-	      s = begin_wide_shared_string(a->size, shift);
-	      switch(shift) {
-	      case 0:
-		for(i = a->size; i--; ) {
-		  s->str[i] = a->item[i].u.integer;
-		}
-		break;
-	      case 1:
-		{
-		  p_wchar1 *str1 = STR1(s);
-		  for(i = a->size; i--; ) {
-		    str1[i] = a->item[i].u.integer;
-		  }
-		}
-		break;
-	      case 2:
-		{
-		  p_wchar2 *str2 = STR2(s);
-		  for(i = a->size; i--; ) {
-		    str2[i] = a->item[i].u.integer;
-		  }
-		}
-		break;
-	      default:
-		free_string(end_shared_string(s));
-		fatal("cast: Bad shift: %d.\n", shift);
-		break;
-	      }
-	      s = end_shared_string(s);
-	      pop_stack();
-	      push_string(s);
-	      return;
-	    }
-	    break;
-	    
-	  default:
-	    Pike_error("Cannot cast %s to string.\n",get_name_of_type(sp[-1].type));
-	}
-	
-	sp[-1].type=T_STRING;
-	sp[-1].u.string=make_shared_string(buf);
-	break;
-      }
-      
       case T_OBJECT:
 	switch(sp[-1].type)
 	{
-	  case T_STRING:
-	    if(Pike_fp->pc)
-	    {
-	      INT32 lineno;
-	      push_text(get_line(Pike_fp->pc, Pike_fp->context.prog, &lineno));
+	  case T_STRING: {
+	    struct pike_string *file;
+	    INT32 lineno;
+	    if(Pike_fp->pc &&
+	       (file = low_get_line(Pike_fp->pc, Pike_fp->context.prog, &lineno))) {
+	      push_string(file);
 	    }else{
 	      push_int(0);
 	    }
 	    APPLY_MASTER("cast_to_object",2);
 	    return;
+	  }
 	    
 	  case T_FUNCTION:
 	    if (Pike_sp[-1].subtype == FUNCTION_BUILTIN) {
@@ -438,16 +550,18 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
       case T_PROGRAM:
       switch(sp[-1].type)
       {
-	case T_STRING:
-	  if(Pike_fp->pc)
-	  {
-	    INT32 lineno;
-	    push_text(get_line(Pike_fp->pc, Pike_fp->context.prog, &lineno));
+	case T_STRING: {
+	  struct pike_string *file;
+	  INT32 lineno;
+	  if(Pike_fp->pc &&
+	     (file = low_get_line(Pike_fp->pc, Pike_fp->context.prog, &lineno))) {
+	    push_string(file);
 	  }else{
 	    push_int(0);
 	  }
 	  APPLY_MASTER("cast_to_program",2);
 	  return;
+	}
 	  
 	case T_FUNCTION:
 	{
@@ -479,7 +593,7 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
       {
 	push_text(get_name_of_type(run_time_type));
 	apply_low(sp[-2].u.object, f, 1);
-	f=!IS_ZERO(sp-1);
+	f=!UNSAFE_IS_ZERO(sp-1);
 	pop_stack();
 	if(f) goto emulated_type_ok;
       }
@@ -497,11 +611,11 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
   {
     case T_ARRAY:
     {
-      struct pike_string *itype;
+      struct pike_type *itype;
       INT32 run_time_itype;
 
-      push_string(itype=index_type(type,int_type_string,0));
-      run_time_itype=compile_type_to_runtime_type(itype);
+      push_type_value(itype = index_type(type, int_type_string, 0));
+      run_time_itype = compile_type_to_runtime_type(itype);
 
       if(run_time_itype != T_MIXED)
       {
@@ -513,7 +627,7 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	{
 	  ref_push_array(a);
 	}else{
-	  INT32 e,i;
+	  INT32 e;
 #ifdef PIKE_DEBUG
 	  struct svalue *save_sp=sp+1;
 #endif
@@ -529,7 +643,7 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	  }
 #ifdef PIKE_DEBUG
 	  if(save_sp!=sp)
-	    fatal("o_cast left stack droppings.\n");
+	    Pike_fatal("o_cast left stack droppings.\n");
 #endif
 	}
 	END_CYCLIC();
@@ -542,27 +656,56 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 
     case T_MULTISET:
     {
-      struct pike_string *itype;
+      struct pike_type *itype;
       INT32 run_time_itype;
 
-      push_string(itype=key_type(type,0));
-      run_time_itype=compile_type_to_runtime_type(itype);
+      push_type_value(itype = key_type(type, 0));
+      run_time_itype = compile_type_to_runtime_type(itype);
 
       if(run_time_itype != T_MIXED)
       {
 	struct multiset *m;
+#ifdef PIKE_NEW_MULTISETS
+	struct multiset *tmp=sp[-2].u.multiset;
+#else
 	struct array *tmp=sp[-2].u.multiset->ind;
+#endif
 	DECLARE_CYCLIC();
 	
 	if((m=(struct multiset *)BEGIN_CYCLIC(tmp,0)))
 	{
 	  ref_push_multiset(m);
 	}else{
-	  INT32 e,i;
-	  struct array *a;
 #ifdef PIKE_DEBUG
 	  struct svalue *save_sp=sp+1;
 #endif
+
+#ifdef PIKE_NEW_MULTISETS
+	  ptrdiff_t nodepos;
+	  if (multiset_indval (tmp))
+	    Pike_error ("FIXME: Casting not implemented for multisets with values.\n");
+	  push_multiset (m = allocate_multiset (multiset_sizeof (tmp),
+						multiset_get_flags (tmp),
+						multiset_get_cmp_less (tmp)));
+
+	  SET_CYCLIC_RET(m);
+
+	  if ((nodepos = multiset_first (tmp)) >= 0) {
+	    ONERROR uwp;
+	    SET_ONERROR (uwp, do_sub_msnode_ref, tmp);
+	    do {
+	      push_multiset_index (tmp, nodepos);
+	      o_cast(itype, run_time_itype);
+	      multiset_insert_2 (m, sp - 1, NULL, 0);
+	      pop_stack();
+	    } while ((nodepos = multiset_next (tmp, nodepos)) >= 0);
+	    UNSET_ONERROR (uwp);
+	    sub_msnode_ref (tmp);
+	  }
+
+#else  /* PIKE_NEW_MULTISETS */
+	  INT32 e;
+	  struct array *a;
 	  push_multiset(m=allocate_multiset(a=allocate_array(tmp->size)));
 	  
 	  SET_CYCLIC_RET(m);
@@ -574,11 +717,13 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	    array_set_index(a,e,sp-1);
 	    pop_stack();
 	  }
+	  order_multiset(m);
+#endif
+
 #ifdef PIKE_DEBUG
 	  if(save_sp!=sp)
-	    fatal("o_cast left stack droppings.\n");
+	    Pike_fatal("o_cast left stack droppings.\n");
 #endif
-	  order_multiset(m);
 	}
 	END_CYCLIC();
 	assign_svalue(sp-3,sp-1);
@@ -590,15 +735,15 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 
     case T_MAPPING:
     {
-      struct pike_string *itype,*vtype;
+      struct pike_type *itype, *vtype;
       INT32 run_time_itype;
       INT32 run_time_vtype;
 
-      push_string(itype=key_type(type,0));
-      run_time_itype=compile_type_to_runtime_type(itype);
+      push_type_value(itype = key_type(type, 0));
+      run_time_itype = compile_type_to_runtime_type(itype);
 
-      push_string(vtype=index_type(type,mixed_type_string,0));
-      run_time_vtype=compile_type_to_runtime_type(vtype);
+      push_type_value(vtype = index_type(type, mixed_type_string, 0));
+      run_time_vtype = compile_type_to_runtime_type(vtype);
 
       if(run_time_itype != T_MIXED ||
 	 run_time_vtype != T_MIXED)
@@ -611,7 +756,7 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	{
 	  ref_push_mapping(m);
 	}else{
-	  INT32 e,i;
+	  INT32 e;
 	  struct keypair *k;
 #ifdef PIKE_DEBUG
 	  struct svalue *save_sp=sp+1;
@@ -631,7 +776,7 @@ void o_cast(struct pike_string *type, INT32 run_time_type)
 	  }
 #ifdef PIKE_DEBUG
 	  if(save_sp!=sp)
-	    fatal("o_cast left stack droppings.\n");
+	    Pike_fatal("o_cast left stack droppings.\n");
 #endif
 	}
 	END_CYCLIC();
@@ -648,14 +793,15 @@ PMOD_EXPORT void f_cast(void)
 {
 #ifdef PIKE_DEBUG
   struct svalue *save_sp=sp;
-  if(sp[-2].type != T_STRING)
-    fatal("Cast expression destroyed stack or left droppings!\n");
+  if(sp[-2].type != T_TYPE)
+    Pike_fatal("Cast expression destroyed stack or left droppings! (Type:%d)\n",
+	  sp[-2].type);
 #endif
-  o_cast(sp[-2].u.string,
-	 compile_type_to_runtime_type(sp[-2].u.string));
+  o_cast(sp[-2].u.type,
+	 compile_type_to_runtime_type(sp[-2].u.type));
 #ifdef PIKE_DEBUG
   if(save_sp != sp)
-    fatal("Internal error: o_cast() left droppings on stack.\n");
+    Pike_fatal("Internal error: o_cast() left droppings on stack.\n");
 #endif
   free_svalue(sp-2);
   sp[-2]=sp[-1];
@@ -796,8 +942,8 @@ CHAROPT(							\
     order=get_switch_order(set->a);				\
     for(e=0;e<(size_t)set->a->size;e+=2)			\
     {								\
-      if(order[e]+1 != order[e+1])				\
-      {								\
+      if(order[e]+1 != order[e+1] &&				\
+         order[e+1]+1 != order[e]) {				\
         free_array(set->a);					\
         set->a=0;						\
         free((char *)order);					\
@@ -844,6 +990,8 @@ static INLINE FLOAT_TYPE low_parse_IEEE_float(char *b, int sz)
   int s, e;
   unsigned char x[4];
   double r;
+  DECLARE_INF
+  DECLARE_NAN
 
   x[0] = EXTRACT_UCHAR(b);
   x[1] = EXTRACT_UCHAR(b+1);
@@ -880,30 +1028,10 @@ static INLINE FLOAT_TYPE low_parse_IEEE_float(char *b, int sz)
   if(e>=9999)
     if(f||extra_f) {
       /* NAN */
-#ifdef HAVE_INFNAN
-      return (FLOAT_TYPE)infnan(EDOM);
-#else
-#ifdef HAVE_NAN
-      /* C99 provides a portable way of generating NaN */
-      return (FLOAT_TYPE)nan("");
-#else
-      return (FLOAT_TYPE)0.0;
-#endif /* HAVE_NAN */
-#endif /* HAVE_INFNAN */
+      return (FLOAT_TYPE)MAKE_NAN();
     } else {
       /* +/- Infinity */
-#ifdef HAVE_INFNAN
-      return (FLOAT_TYPE)infnan(s? -ERANGE:ERANGE);
-#else
-#ifdef HUGE_VAL
-      return (FLOAT_TYPE)(s? -HUGE_VAL:HUGE_VAL);
-#else
-      /* This number is infinite enough...  :) */
-      e = 1024;
-      f = 1;
-      extra_f = 0;
-#endif /* HUGE_VAL */
-#endif /* HAVE_INFNAN */
+      return (FLOAT_TYPE)MAKE_INF(s? -1:1);
     }
 
   r = (double)f;
@@ -1036,7 +1164,6 @@ static INT32 PIKE_CONCAT4(very_low_sscanf_,INPUT_SHIFT,_,MATCH_SHIFT)(	 \
   ptrdiff_t cnt, eye, e, field_length = 0;				 \
   int no_assign = 0, minus_flag = 0;					 \
   struct sscanf_set set;						 \
-  struct svalue *argp;							 \
 									 \
 									 \
   set.a = 0;								 \
@@ -1073,7 +1200,7 @@ static INT32 PIKE_CONCAT4(very_low_sscanf_,INPUT_SHIFT,_,MATCH_SHIFT)(	 \
     DO_IF_DEBUG(							 \
     if(match[cnt]!='%' || match[cnt+1]=='%')				 \
     {									 \
-      fatal("Error in sscanf.\n");					 \
+      Pike_fatal("Error in sscanf.\n");					 \
     }									 \
     );									 \
 									 \
@@ -1165,7 +1292,9 @@ static INT32 PIKE_CONCAT4(very_low_sscanf_,INPUT_SHIFT,_,MATCH_SHIFT)(	 \
 									 \
 	case 'c':							 \
         {								 \
+CHAROPT2(								 \
           int e;							 \
+)									 \
 	  sval.type=T_INT;						 \
 	  sval.subtype=NUMBER_NUMBER;					 \
           if(field_length == -1)					 \
@@ -1201,7 +1330,6 @@ CHAROPT2(								 \
 									 \
 	     while(--field_length >= 0)					 \
 	     {								 \
-	       int lshfun, orfun;					 \
 	       x = input[eye];						 \
 									 \
                DO_IF_BIGNUM(						 \
@@ -1305,8 +1433,9 @@ CHAROPT2(								 \
 	    chars_matched[0]=eye;					 \
 	    return matches;						 \
 	  }								 \
-	  sval.u.float_number=STRTOD_PCHARP(MKPCHARP(input+eye,		 \
-						     INPUT_SHIFT),&t2);	 \
+	  sval.u.float_number =						 \
+	    (FLOAT_TYPE)STRTOD_PCHARP(MKPCHARP(input+eye,		 \
+					       INPUT_SHIFT),&t2);	 \
 	  t = (PIKE_CONCAT(p_wchar, INPUT_SHIFT) *)(t2.ptr);		 \
 	  if(input + eye == t)						 \
 	  {								 \
@@ -1429,7 +1558,6 @@ CHAROPT2(								 \
 		  MEMSET(set.c, 1, sizeof(set.c));			 \
 		  for(e='0';e<='9';e++) set.c[e]=0;			 \
 		  set.c['-']=0;						 \
-		  set.c['x']=0;						 \
 		  goto match_set;					 \
 									 \
 		case 'f':						 \
@@ -1630,7 +1758,7 @@ void o_sscanf(INT32 args)
 #ifdef PIKE_DEBUG
   extern int t_flag;
 #endif
-  INT32 e, i=0;
+  INT32 i=0;
   int x;
   ptrdiff_t matched_chars;
   struct svalue *save_sp=sp;
@@ -1752,12 +1880,21 @@ void o_sscanf(INT32 args)
   push_int(i);
 }
 
+/*! @decl array array_sscanf(string data, string format)
+ *!
+ *! This function works just like @[sscanf()], but returns the matched
+ *! results in an array instead of assigning them to lvalues. This is often
+ *! useful for user-defined sscanf strings.
+ *!
+ *! @seealso
+ *!   @[sscanf()], @[`/()]
+ */
 PMOD_EXPORT void f_sscanf(INT32 args)
 {
 #ifdef PIKE_DEBUG
   extern int t_flag;
 #endif
-  INT32 e,i;
+  INT32 i;
   int x;
   ptrdiff_t matched_chars;
   struct svalue *save_sp=sp;
@@ -1859,3 +1996,8 @@ PMOD_EXPORT void f_sscanf(INT32 args)
   push_array(a);
 }
 
+
+void o_breakpoint(void)
+{
+  /* Does nothing */
+}
