@@ -1,3 +1,10 @@
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: peep.c,v 1.88 2004/05/11 15:47:39 grubba Exp $
+*/
+
 #include "global.h"
 #include "language.h"
 #include "stralloc.h"
@@ -14,13 +21,18 @@
 #include "stuff.h"
 #include "bignum.h"
 #include "opcodes.h"
+#include "builtin_functions.h"
+#include "constants.h"
+#include "interpret.h"
+#include "pikecode.h"
 
-RCSID("$Id: peep.c,v 1.48 2003/10/10 01:17:44 mast Exp $");
+RCSID("$Id: peep.c,v 1.88 2004/05/11 15:47:39 grubba Exp $");
 
 static void asm_opt(void);
 
 dynamic_buffer instrbuf;
 
+#ifdef PIKE_DEBUG
 static int hasarg(int opcode)
 {
   return instrs[opcode-F_OFFSET].flags & I_HASARG;
@@ -31,7 +43,6 @@ static int hasarg2(int opcode)
   return instrs[opcode-F_OFFSET].flags & I_HASARG2;
 }
 
-#ifdef PIKE_DEBUG
 static void dump_instr(p_instr *p)
 {
   if(!p) return;
@@ -50,7 +61,7 @@ static void dump_instr(p_instr *p)
 
 void init_bytecode(void)
 {
-  low_reinit_buf(&instrbuf);
+  initialize_buf(&instrbuf);
 }
 
 void exit_bytecode(void)
@@ -76,7 +87,7 @@ ptrdiff_t insert_opcode2(unsigned int f,
 
 #ifdef PIKE_DEBUG
   if(!hasarg2(f) && c)
-    fatal("hasarg2(%d) is wrong!\n",f);
+    Pike_fatal("hasarg2(%d) is wrong!\n",f);
 #endif
 
   p=(p_instr *)low_make_buf_space(sizeof(p_instr), &instrbuf);
@@ -84,7 +95,7 @@ ptrdiff_t insert_opcode2(unsigned int f,
 
 #ifdef PIKE_DEBUG
   if(!instrbuf.s.len)
-    fatal("Low make buf space failed!!!!!!\n");
+    Pike_fatal("Low make buf space failed!!!!!!\n");
 #endif
 
   p->opcode=f;
@@ -103,7 +114,7 @@ ptrdiff_t insert_opcode1(unsigned int f,
 {
 #ifdef PIKE_DEBUG
   if(!hasarg(f) && b)
-    fatal("hasarg(%d) is wrong!\n",f);
+    Pike_fatal("hasarg(%d) is wrong!\n",f);
 #endif
 
   return insert_opcode2(f,b,0,current_line,current_file);
@@ -113,7 +124,7 @@ ptrdiff_t insert_opcode0(int f,int current_line, struct pike_string *current_fil
 {
 #ifdef PIKE_DEBUG
   if(hasarg(f))
-    fatal("hasarg(%d) is wrong!\n",f);
+    Pike_fatal("hasarg(%d) is wrong!\n",f);
 #endif
   return insert_opcode1(f,0,current_line, current_file);
 }
@@ -124,112 +135,163 @@ void update_arg(int instr,INT32 arg)
   p_instr *p;
 #ifdef PIKE_DEBUG
   if(instr > (long)instrbuf.s.len / (long)sizeof(p_instr) || instr < 0)
-    fatal("update_arg outside known space.\n");
+    Pike_fatal("update_arg outside known space.\n");
 #endif  
   p=(p_instr *)instrbuf.s.str;
   p[instr].arg=arg;
 }
 
+#ifndef FLUSH_CODE_GENERATOR_STATE
+#define FLUSH_CODE_GENERATOR_STATE()
+#endif
 
 /**** Bytecode Generator *****/
 
-void ins_f_byte(unsigned int b)
+INT32 assemble(int store_linenumbers)
 {
-#ifdef PIKE_DEBUG
-  if(store_linenumbers && b<F_MAX_OPCODE)
-    ADD_COMPILED(b);
-#endif /* PIKE_DEBUG */
-
-  b-=F_OFFSET;
-#ifdef PIKE_DEBUG
-  if(b>255)
-    Pike_error("Instruction too big %d\n",b);
-#endif
-  add_to_program((unsigned char)b);
-}
-
-static void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
-{
-  switch(b >> 8)
-  {
-  case 0 : break;
-  case 1 : ins_f_byte(F_PREFIX_256); break;
-  case 2 : ins_f_byte(F_PREFIX_512); break;
-  case 3 : ins_f_byte(F_PREFIX_768); break;
-  case 4 : ins_f_byte(F_PREFIX_1024); break;
-  default:
-    if( b < 256*256)
-    {
-      ins_f_byte(F_PREFIX_CHARX256);
-      add_to_program((unsigned char)(b>>8));
-    }else if(b < 256*256*256) {
-      ins_f_byte(F_PREFIX_WORDX256);
-      add_to_program((unsigned char)(b>>16));
-      add_to_program((unsigned char)(b>>8));
-    }else{
-      ins_f_byte(F_PREFIX_24BITX256);
-      add_to_program((unsigned char)(b>>24));
-      add_to_program((unsigned char)(b>>16));
-      add_to_program((unsigned char)(b>>8));
-    }
-  }
-  ins_f_byte(a);
-  add_to_program((unsigned char)b);
-}
-
-static void ins_f_byte_with_2_args(unsigned int a,
-				   unsigned INT32 c,
-				   unsigned INT32 b)
-{
-
-  switch(b >> 8)
-  {
-  case 0 : break;
-  case 1 : ins_f_byte(F_PREFIX2_256); break;
-  case 2 : ins_f_byte(F_PREFIX2_512); break;
-  case 3 : ins_f_byte(F_PREFIX2_768); break;
-  case 4 : ins_f_byte(F_PREFIX2_1024); break;
-  default:
-    if( b < 256*256)
-    {
-      ins_f_byte(F_PREFIX2_CHARX256);
-      add_to_program((unsigned char)(b>>8));
-    }else if(b < 256*256*256) {
-      ins_f_byte(F_PREFIX2_WORDX256);
-      add_to_program((unsigned char)(b>>16));
-      add_to_program((unsigned char)(b>>8));
-    }else{
-      ins_f_byte(F_PREFIX2_24BITX256);
-      add_to_program((unsigned char)(b>>24));
-      add_to_program((unsigned char)(b>>16));
-      add_to_program((unsigned char)(b>>8));
-    }
-  }
-  ins_f_byte_with_arg(a,c);
-  add_to_program((unsigned char)b);
-}
-
-void assemble(void)
-{
-  INT32 d,max_label,tmp;
+  INT32 entry_point;
+  INT32 max_label=-1,tmp;
   INT32 *labels, *jumps, *uses;
   ptrdiff_t e, length;
   p_instr *c;
-  int reoptimize=1;
+  int reoptimize=!(debug_options & NO_PEEP_OPTIMIZING);
+#ifdef PIKE_PORTABLE_BYTECODE
+  struct pike_string *tripples = NULL;
+#endif /* PIKE_PORTABLE_BYTECODE */
+#ifdef PIKE_DEBUG
+  INT32 max_pointer=-1;
+  int synch_depth = 0;
+  size_t fun_start = Pike_compiler->new_program->num_program;
+#endif
 
   c=(p_instr *)instrbuf.s.str;
   length=instrbuf.s.len / sizeof(p_instr);
 
-  max_label=-1;
-  for(e=0;e<length;e++,c++)
-    if(c->opcode == F_LABEL)
+#ifdef PIKE_DEBUG
+  if((a_flag > 1 && store_linenumbers) || a_flag > 2)
+  {
+    for (e = 0; e < length; e++) {
+      if (c[e].opcode == F_POP_SYNCH_MARK) synch_depth--;
+      fprintf(stderr, "~~~%4d %4lx %*s", c[e].line,
+	      DO_NOT_WARN((unsigned long)e), synch_depth, "");
+      dump_instr(c+e);
+      fprintf(stderr,"\n");
+      if (c[e].opcode == F_SYNCH_MARK) synch_depth++;
+    }
+    if (synch_depth) {
+      Pike_fatal("Unbalanced sync_mark/pop_sync_mark: %d\n",
+		 synch_depth);
+    }
+  }
+#endif
+
+#ifdef PIKE_PORTABLE_BYTECODE
+  /* No need to do this for constant evaluations. */
+  if (store_linenumbers) {
+    p_wchar2 *current_tripple;
+    struct pike_string *previous_file = NULL;
+    int previous_line = 0;
+    ptrdiff_t num_linedirectives = 0;
+
+    /* Count the number of F_FILENAME/F_LINE pseudo-ops we need to add. */
+    for (e=0; e < length; e++) {
+      if (c[e].file != previous_file) {
+	previous_file = c[e].file;
+	num_linedirectives++;
+      }
+      if (c[e].line != previous_line) {
+	previous_line = c[e].line;
+	num_linedirectives++;
+      }
+    }
+
+    /* fprintf(stderr, "length:%d directives:%d\n",
+     *         length, num_linedirectives);
+     */
+      
+    if (!(tripples = begin_wide_shared_string(3*(length+num_linedirectives),
+					      2))) {
+      Pike_fatal("Failed to allocate wide string of length %d 3*(%d + %d).\n",
+		 3*(length+num_linedirectives), length, num_linedirectives);
+    }
+    previous_file = NULL;
+    previous_line = 0;
+    current_tripple = STR2(tripples);
+    for (e = 0; e < length; e++) {
+      if (c[e].file != previous_file) {
+	current_tripple[0] = F_FILENAME;
+	current_tripple[1] = store_prog_string(c[e].file);
+	current_tripple[2] = 0;
+	current_tripple += 3;
+	previous_file = c[e].file;
+      }
+      if (c[e].line != previous_line) {
+	current_tripple[0] = F_LINE;
+	current_tripple[1] = c[e].line;
+	current_tripple[2] = 0;
+	current_tripple += 3;
+	previous_line = c[e].line;
+      }
+      current_tripple[0] = c[e].opcode;
+      current_tripple[1] = c[e].arg;
+      current_tripple[2] = c[e].arg2;
+      current_tripple += 3;
+    }
+#ifdef PIKE_DEBUG
+    if (current_tripple != STR2(tripples) + 3*(length + num_linedirectives)) {
+      Pike_fatal("Tripple length mismatch %d != %d 3*(%d + %d)\n",
+		 current_tripple - STR2(tripples),
+		 3*(length + num_linedirectives),
+		 length, num_linedirectives);
+    }
+#endif /* PIKE_DEBUG */
+    tripples = end_shared_string(tripples);
+  }
+  
+#endif /* PIKE_PORTABLE_BYTECODE */
+
+  for(e=0;e<length;e++,c++) {
+    if(c->opcode == F_LABEL) {
       if(c->arg > max_label)
 	max_label = c->arg;
+    }
+#ifdef PIKE_DEBUG
+    else if (instrs[c->opcode - F_OFFSET].flags & I_POINTER) {
+      if (c->arg > max_pointer)
+	max_pointer = c->arg;
+    }
+#endif /* PIKE_DEBUG */
+  }
 
+#ifdef PIKE_DEBUG
+  if (max_pointer > max_label) {
+    fprintf(stderr,
+	    "Reference to undefined label %d > %d\n"
+	    "Bad instructions are marked with '***':\n",
+	    max_pointer, max_label);
+    c=(p_instr *)instrbuf.s.str;
+    for(e=0;e<length;e++,c++) {
+      if (c->opcode == F_POP_SYNCH_MARK) synch_depth--;
+      fprintf(stderr, " * %4d %4lx ",
+	      c->line, DO_NOT_WARN((unsigned long)e));
+      dump_instr(c);
+      if ((instrs[c->opcode - F_OFFSET].flags & I_POINTER) &&
+	  (c->arg > max_label)) {
+	fprintf(stderr, " ***\n");
+      } else {
+	fprintf(stderr, "\n");
+      }
+      if (c->opcode == F_SYNCH_MARK) synch_depth++;
+    }
+    
+    Pike_fatal("Reference to undefined label %d > %d\n",
+	       max_pointer, max_label);
+  }
+#endif /* PIKE_DEBUG */
 
-  labels=(INT32 *)xalloc(sizeof(INT32) * (max_label+2));
-  jumps=(INT32 *)xalloc(sizeof(INT32) * (max_label+2));
-  uses=(INT32 *)xalloc(sizeof(INT32) * (max_label+2));
+  labels=(INT32 *)xalloc(sizeof(INT32) * 3 * (max_label+2));
+  jumps = labels + max_label + 2;
+  uses = jumps + max_label + 2;
 
   while(reoptimize)
   {
@@ -243,8 +305,9 @@ void assemble(void)
     c=(p_instr *)instrbuf.s.str;
     length=instrbuf.s.len / sizeof(p_instr);
     for(e=0;e<length;e++)
-      if(c[e].opcode == F_LABEL && c[e].arg>=0)
+      if(c[e].opcode == F_LABEL && c[e].arg>=0) {
 	labels[c[e].arg]=DO_NOT_WARN((INT32)e);
+      }
     
     for(e=0;e<length;e++)
     {
@@ -252,7 +315,7 @@ void assemble(void)
       {
 	while(1)
 	{
-	  int tmp,tmp2;
+	  int tmp;
 	  tmp=labels[c[e].arg];
 	  
 	  while(tmp<length &&
@@ -300,7 +363,7 @@ void assemble(void)
 	}
 #ifdef PIKE_DEBUG
 	if (c[e].arg < 0 || c[e].arg > max_label)
-	  fatal ("Invalid index into uses: %d\n", c[e].arg);
+	  Pike_fatal ("Invalid index into uses: %d\n", c[e].arg);
 #endif
 	uses[c[e].arg]++;
       }
@@ -318,29 +381,68 @@ void assemble(void)
     if(!reoptimize) break;
     
     asm_opt();
+#if 1
+#ifdef PIKE_DEBUG
+    if (a_flag > 3)
+      fprintf(stderr, "Rerunning optimizer.\n");
+#endif
+#else /* !1 */
     reoptimize=0;
+#endif /* 1 */
   }
 
+  /* Time to create the actual bytecode. */
+
   c=(p_instr *)instrbuf.s.str;
   length=instrbuf.s.len / sizeof(p_instr);
 
   for(e=0;e<=max_label;e++) labels[e]=jumps[e]=-1;
   
-  c=(p_instr *)instrbuf.s.str;
+#ifdef ALIGN_PIKE_FUNCTION_BEGINNINGS
+  while( ( (((INT32) PIKE_PC)+2) & (ALIGN_PIKE_JUMPS-1)))
+    ins_byte(0);
+#endif
+
+#ifdef PIKE_PORTABLE_BYTECODE
+  if (store_linenumbers) {
+    ins_data(store_prog_string(tripples));
+    free_string(tripples);
+  } else {
+    ins_data(NULL);
+  }
+#endif /* PIKE_PORTABLE_BYTECODE */
+
+  entry_point = PIKE_PC;
+
+#ifdef PIKE_DEBUG
+  synch_depth = 0;
+#endif
+  FLUSH_CODE_GENERATOR_STATE();
   for(e=0;e<length;e++)
   {
 #ifdef PIKE_DEBUG
+    if (c != (((p_instr *)instrbuf.s.str)+e)) {
+      Pike_fatal("Instruction loop deviates. 0x%04x != 0x%04x\n",
+		 e, DO_NOT_WARN((INT32)(c - ((p_instr *)instrbuf.s.str))));
+    }
     if((a_flag > 2 && store_linenumbers) || a_flag > 3)
     {
-      fprintf(stderr, "===%3d %4lx ", c->line,
-	      DO_NOT_WARN((unsigned long)PIKE_PC));
+      if (c->opcode == F_POP_SYNCH_MARK) synch_depth--;
+      fprintf(stderr, "===%4d %4lx %*s", c->line,
+	      DO_NOT_WARN((unsigned long)PIKE_PC), synch_depth, "");
       dump_instr(c);
       fprintf(stderr,"\n");
+      if (c->opcode == F_SYNCH_MARK) synch_depth++;
     }
 #endif
 
-    if(store_linenumbers)
+    if(store_linenumbers) {
       store_linenumber(c->line, c->file);
+#ifdef PIKE_DEBUG
+      if (c->opcode < F_MAX_OPCODE)
+	ADD_COMPILED(c->opcode);
+#endif /* PIKE_DEBUG */
+    }
 
     switch(c->opcode)
     {
@@ -349,51 +451,145 @@ void assemble(void)
     case F_START_FUNCTION:
       break;
     case F_ALIGN:
-      while(PIKE_PC % c->arg) add_to_program(0);
+      ins_align(c->arg);
       break;
 
     case F_BYTE:
-      add_to_program((unsigned char)(c->arg));
+      ins_byte((unsigned char)(c->arg));
       break;
 
     case F_DATA:
-      add_relocated_int_to_program(c->arg);
+      ins_data(c->arg);
+      break;
+
+    case F_ENTRY:
+#ifdef INS_ENTRY
+      INS_ENTRY();
+#endif /* INS_ENTRY */
       break;
 
     case F_LABEL:
       if(c->arg == -1) break;
 #ifdef PIKE_DEBUG
       if(c->arg > max_label || c->arg < 0)
-	fatal("max_label calculation failed!\n");
+	Pike_fatal("max_label calculation failed!\n");
 
       if(labels[c->arg] != -1)
-	fatal("Duplicate label!\n");
+	Pike_fatal("Duplicate label!\n");
 #endif
+      FLUSH_CODE_GENERATOR_STATE();
       labels[c->arg] = DO_NOT_WARN((INT32)PIKE_PC);
+      if ((e+1 < length) &&
+	  (c[1].opcode != F_LABEL) &&
+	  (c[1].opcode != F_BYTE) &&
+	  (c[1].opcode != F_DATA)) {
+	/* Don't add redundant code before labels or raw data. */
+	UPDATE_PC();
+      }
+
+      break;
+
+    case F_VOLATILE_RETURN:
+      ins_f_byte(F_RETURN);
+      break;
+
+    case F_POINTER:
+#ifdef PIKE_DEBUG
+      if(c->arg > max_label || c->arg < 0)
+	Pike_fatal("Jump to unknown label?\n");
+#endif
+      tmp = DO_NOT_WARN((INT32)PIKE_PC);
+      ins_pointer(jumps[c->arg]);
+      jumps[c->arg]=tmp;
       break;
 
     default:
-      switch(instrs[c->opcode - F_OFFSET].flags)
+      switch(instrs[c->opcode - F_OFFSET].flags & I_IS_MASK)
       {
       case I_ISJUMP:
+#ifdef INS_F_JUMP
+	tmp=INS_F_JUMP(c->opcode, (labels[c->arg] != -1));
+	if(tmp != -1)
+	{
+	  UPDATE_F_JUMP(tmp, jumps[c->arg]);
+	  jumps[c->arg]=~tmp;
+	  break;
+	}
+#endif
+
 	ins_f_byte(c->opcode);
 
-      case I_ISPOINTER:
 #ifdef PIKE_DEBUG
-	if(c->arg > max_label || c->arg < 0) fatal("Jump to unknown label?\n");
+	if(c->arg > max_label || c->arg < 0)
+	  Pike_fatal("Jump to unknown label?\n");
 #endif
 	tmp = DO_NOT_WARN((INT32)PIKE_PC);
-	add_relocated_int_to_program(jumps[c->arg]);
+	ins_pointer(jumps[c->arg]);
 	jumps[c->arg]=tmp;
 	break;
 
-	case I_TWO_ARGS:
-	  ins_f_byte_with_2_args(c->opcode, c->arg,c->arg2);
+      case I_ISJUMPARGS:
+#ifdef INS_F_JUMP_WITH_TWO_ARGS
+	tmp = INS_F_JUMP_WITH_TWO_ARGS(c->opcode, c->arg, c->arg2,
+				       (labels[c[1].arg] != -1));
+	if(tmp != -1)
+	{
+
+	  /* Step ahead to the pointer instruction, and inline it. */
+#ifdef PIKE_DEBUG
+	  if (c[1].opcode != F_POINTER) {
+	    Pike_fatal("Expected opcode %s to be followed by a pointer\n",
+		       instrs[c->opcode - F_OFFSET].name);
+	  }
+#endif /* PIKE_DEBUG */
+	  c++;
+	  e++;
+	  UPDATE_F_JUMP(tmp, jumps[c->arg]);
+	  jumps[c->arg]=~tmp;
 	  break;
-	  
-	case I_HASARG:
-	  ins_f_byte_with_arg(c->opcode, c->arg);
+	}
+#endif /* INS_F_JUMP_WITH_TWO_ARGS */
+
+	/* FALL_THROUGH
+	 *
+	 * Note that the pointer in this case will be handled by the
+	 * next turn through the loop.
+	 */
+
+      case I_TWO_ARGS:
+	ins_f_byte_with_2_args(c->opcode, c->arg, c->arg2);
+	break;
+
+      case I_ISJUMPARG:
+#ifdef INS_F_JUMP_WITH_ARG
+	tmp = INS_F_JUMP_WITH_ARG(c->opcode, c->arg, (labels[c[1].arg] != -1));
+	if(tmp != -1)
+	{
+
+	  /* Step ahead to the pointer instruction, and inline it. */
+#ifdef PIKE_DEBUG
+	  if (c[1].opcode != F_POINTER) {
+	    Pike_fatal("Expected opcode %s to be followed by a pointer\n",
+		       instrs[c->opcode - F_OFFSET].name);
+	  }
+#endif /* PIKE_DEBUG */
+	  c++;
+	  e++;
+	  UPDATE_F_JUMP(tmp, jumps[c->arg]);
+	  jumps[c->arg]=~tmp;
 	  break;
+	}
+#endif /* INS_F_JUMP_WITH_ARG */
+
+	/* FALL_THROUGH
+	 *
+	 * Note that the pointer in this case will be handled by the
+	 * next turn through the loop.
+	 */
+
+      case I_HASARG:
+	ins_f_byte_with_arg(c->opcode, c->arg);
+	break;
 
       case 0:
 	ins_f_byte(c->opcode);
@@ -401,37 +597,108 @@ void assemble(void)
 
 #ifdef PIKE_DEBUG
       default:
-	fatal("Unknown instruction type.\n");
+	Pike_fatal("Unknown instruction type.\n");
 #endif
       }
     }
+
+
+#ifdef PIKE_DEBUG
+    if (instrs[c->opcode - F_OFFSET].flags & I_HASPOINTER) {
+      if ((e+1 >= length) || (c[1].opcode != F_POINTER)) {
+	Pike_fatal("Expected instruction %s to be followed by a pointer.\n"
+		   "Got %s (%d != %d)\n.",
+		   instrs[c->opcode - F_OFFSET].name,
+		   (e+1 < length)?instrs[c[1].opcode - F_OFFSET].name:"EOI",
+		   (e+1 < length)?c[1].opcode:0, F_POINTER);
+      }
+    }
+#endif /* PIKE_DEBUG */
+
+#ifdef ALIGN_PIKE_JUMPS
+    if(e+1 < length)
+    {
+      /* FIXME: Note that this code won't work for opcodes of type
+       *        I_ISJUMPARG or I_ISJUMPARGS, since c may already
+       *        have been advanced to the corresponding F_POINTER.
+       *        With the current opcode set this is a non-issue, but...
+       * /grubba 2002-11-02
+       */
+      switch(c->opcode)
+      {
+	case F_RETURN:
+	case F_VOLATILE_RETURN:
+	case F_BRANCH:
+	case F_RETURN_0:
+	case F_RETURN_1:
+	case F_RETURN_LOCAL:
+	  
+#define CALLS(X) \
+      case PIKE_CONCAT3(F_,X,_AND_RETURN): \
+      case PIKE_CONCAT3(F_MARK_,X,_AND_RETURN):
+	  
+	  CALLS(APPLY)
+	    CALLS(CALL_FUNCTION)
+	    CALLS(CALL_LFUN)
+	    CALLS(CALL_BUILTIN)
+	    while( ((INT32) PIKE_PC & (ALIGN_PIKE_JUMPS-1) ) )
+	      ins_byte(0);
+      }
+    }
+#endif
     
     c++;
   }
 
   for(e=0;e<=max_label;e++)
   {
-    int tmp2=labels[e];
+    INT32 tmp2=labels[e];
 
     while(jumps[e]!=-1)
     {
 #ifdef PIKE_DEBUG
       if(labels[e]==-1)
-	fatal("Hyperspace error: unknown jump point %ld at %d (pc=%x).\n",
-	      PTRDIFF_T_TO_LONG(e), labels[e], jumps[e]);
+	Pike_fatal("Hyperspace error: unknown jump point %ld(%ld) at %d (pc=%x).\n",
+		   PTRDIFF_T_TO_LONG(e), PTRDIFF_T_TO_LONG(max_label),
+		   labels[e], jumps[e]);
 #endif
-      tmp=read_int(jumps[e]);
-      upd_int(jumps[e], tmp2 - jumps[e]);
+#ifdef INS_F_JUMP
+      if(jumps[e] < 0)
+      {
+	tmp = READ_F_JUMP(~jumps[e]);
+	UPDATE_F_JUMP(~jumps[e], tmp2);
+	jumps[e]=tmp;
+	continue;
+      }
+#endif
+
+      tmp = read_pointer(jumps[e]);
+      upd_pointer(jumps[e], tmp2 - jumps[e]);
       jumps[e]=tmp;
     }
   }
 
   free((char *)labels);
-  free((char *)jumps);
-  free((char *)uses);
 
+#ifdef PIKE_DEBUG
+  if (a_flag > 6) {
+    size_t len = (Pike_compiler->new_program->num_program - fun_start)*
+      sizeof(PIKE_OPCODE_T);
+    fprintf(stderr, "Code at offset %d through %d:\n",
+	    fun_start, Pike_compiler->new_program->num_program-1);
+#ifdef DISASSEMBLE_CODE
+    DISASSEMBLE_CODE(Pike_compiler->new_program->program + fun_start, len);
+#else /* !DISASSEMBLE_CODE */
+    {
+      /* FIXME: Hexdump here. */
+    }
+#endif /* DISASSEMBLE_CODE */
+  }
+#endif /* PIKE_DEBUG */
 
   exit_bytecode();
+
+  return entry_point;
 }
 
 /**** Peephole optimizer ****/
@@ -448,7 +715,7 @@ static INLINE ptrdiff_t insopt2(int f, INT32 a, INT32 b,
 
 #ifdef PIKE_DEBUG
   if(!hasarg2(f) && b)
-    fatal("hasarg(%d) is wrong!\n",f);
+    Pike_fatal("hasarg2(%d /*%s */) is wrong!\n",f,get_f_name(f));
 #endif
 
   p=(p_instr *)low_make_buf_space(sizeof(p_instr), &instrbuf);
@@ -461,7 +728,7 @@ static INLINE ptrdiff_t insopt2(int f, INT32 a, INT32 b,
 
 #ifdef PIKE_DEBUG
   if(!instrbuf.s.len)
-    fatal("Low make buf space failed!!!!!!\n");
+    Pike_fatal("Low make buf space failed!!!!!!\n");
 #endif
 
   p->opcode=f;
@@ -477,7 +744,7 @@ static INLINE ptrdiff_t insopt1(int f, INT32 a, int cl, struct pike_string *cf)
 {
 #ifdef PIKE_DEBUG
   if(!hasarg(f) && a)
-    fatal("hasarg(%d) is wrong!\n",f);
+    Pike_fatal("hasarg(%d /* %s */) is wrong!\n",f,get_f_name(f));
 #endif
 
   return insopt2(f,a,0,cl, cf);
@@ -487,7 +754,7 @@ static INLINE ptrdiff_t insopt0(int f, int cl, struct pike_string *cf)
 {
 #ifdef PIKE_DEBUG
   if(hasarg(f))
-    fatal("hasarg(%d) is wrong!\n",f);
+    Pike_fatal("hasarg(%d /* %s */) is wrong!\n",f,get_f_name(f));
 #endif
   return insopt2(f,0,0,cl, cf);
 }
@@ -498,14 +765,14 @@ static void debug(void)
     fifo_len=(long)instrbuf.s.len / (long)sizeof(p_instr);
 #ifdef PIKE_DEBUG
   if(eye < 0)
-    fatal("Popped beyond start of code.\n");
+    Pike_fatal("Popped beyond start of code.\n");
 
   if(instrbuf.s.len)
   {
     p_instr *p;
     p=(p_instr *)low_make_buf_space(0, &instrbuf);
     if(!p[-1].file)
-      fatal("No file name on last instruction!\n");
+      Pike_fatal("No file name on last instruction!\n");
   }
 #endif
 }
@@ -581,7 +848,7 @@ static void pop_n_opcodes(int n)
     if(d>fifo_len) d=fifo_len;
 #ifdef PIKE_DEBUG
     if((long)d > (long)instrbuf.s.len / (long)sizeof(p_instr))
-      fatal("Popping out of instructions.\n");
+      Pike_fatal("Popping out of instructions.\n");
 #endif
 
     /* FIXME: It looks like the fifo could be optimized.
@@ -645,42 +912,48 @@ static void pop_n_opcodes(int n)
 static void do_optimization(int topop, ...)
 {
   va_list arglist;
-  int q=-1;
+  int q=0;
+  int oplen;
 
   DO_OPTIMIZATION_PREQUEL(topop);
 
   va_start(arglist, topop);
   
-  while(1)
+  while((oplen = va_arg(arglist, int)))
   {
     q++;
-    switch(va_arg(arglist, int))
+    switch(oplen)
     {
-      case 0:
+#ifdef PIKE_DEBUG
+      default:
+	Pike_fatal("Unsupported argument number: %d\n", oplen);
 	break;
+#endif /* PIKE_DEBUG */
+
       case 1:
-      {
-	int i=va_arg(arglist, int);
-	insopt0(i,cl,cf);
-	continue;
-      }
+	{
+	  int i=va_arg(arglist, int);
+	  insopt0(i,cl,cf);
+	}
+	break;
+
       case 2:
-      {
-	int i=va_arg(arglist, int);
-	int j=va_arg(arglist, int);
-	insopt1(i,j,cl,cf);
-	continue;
-      }
+	{
+	  int i=va_arg(arglist, int);
+	  int j=va_arg(arglist, int);
+	  insopt1(i,j,cl,cf);
+	}
+	break;
+
       case 3:
-      {
-	int i=va_arg(arglist, int);
-	int j=va_arg(arglist, int);
-	int k=va_arg(arglist, int);
-	insopt2(i,j,k,cl,cf);
-	continue;
-      }
+	{
+	  int i=va_arg(arglist, int);
+	  int j=va_arg(arglist, int);
+	  int k=va_arg(arglist, int);
+	  insopt2(i,j,k,cl,cf);
+	}
+	break;
     }
-    break;
   }
 
   va_end(arglist);
@@ -696,16 +969,19 @@ static void asm_opt(void)
   {
     p_instr *c;
     ptrdiff_t e, length;
+    int synch_depth = 0;
 
     c=(p_instr *)instrbuf.s.str;
     length=instrbuf.s.len / sizeof(p_instr);
 
-    fprintf(stderr,"Optimization begins: \n");
+    fprintf(stderr,"Before peep optimization:\n");
     for(e=0;e<length;e++,c++)
     {
-      fprintf(stderr,"---%3d: ",c->line);
+      if (c->opcode == F_POP_SYNCH_MARK) synch_depth--;
+      fprintf(stderr,"<<<%4d: %*s",c->line,synch_depth,"");
       dump_instr(c);
       fprintf(stderr,"\n");
+      if (c->opcode == F_SYNCH_MARK) synch_depth++;
     }
   }
 #endif
@@ -719,18 +995,20 @@ static void asm_opt(void)
   {
     p_instr *c;
     ptrdiff_t e, length;
+    int synch_depth = 0;
 
     c=(p_instr *)instrbuf.s.str;
     length=instrbuf.s.len / sizeof(p_instr);
 
-    fprintf(stderr,"Optimization begins: \n");
+    fprintf(stderr,"After peep optimization:\n");
     for(e=0;e<length;e++,c++)
     {
-      fprintf(stderr,">>>%3d: ",c->line);
+      if (c->opcode == F_POP_SYNCH_MARK) synch_depth--;
+      fprintf(stderr,">>>%4d: %*s",c->line,synch_depth,"");
       dump_instr(c);
       fprintf(stderr,"\n");
+      if (c->opcode == F_SYNCH_MARK) synch_depth++;
     }
   }
 #endif
 }
-
