@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: preprocessor.h,v 1.56 2004/04/22 15:45:02 grubba Exp $
+|| $Id: preprocessor.h,v 1.57 2004/06/30 17:57:16 grubba Exp $
 */
 
 /*
@@ -964,8 +964,8 @@ static ptrdiff_t lower_cpp(struct cpp *this,
 		SKIPWHITE();
 		if(data[pos]==')')
 		{
-		  if(d->varargs && arg + 1 == d->args)
-		  {
+ 		  if((d->varargs && arg + 1 == d->args) ||
+		     (!arg && (d->args == 1))) {
 		    arguments[arg].arg = MKPCHARP(data + pos, SHIFT);
 		    arguments[arg].len=0;
 		    continue;
