@@ -1,12 +1,10 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-
 /*
- * $Id: program.h,v 1.119 2002/05/06 14:35:33 mast Exp $
- */
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: program.h,v 1.181 2004/03/23 13:41:56 grubba Exp $
+*/
+
 #ifndef PROGRAM_H
 #define PROGRAM_H
 
@@ -16,7 +14,8 @@
 #include "pike_macros.h"
 #include "svalue.h"
 #include "time_stuff.h"
-
+#include "program_id.h"
+#include "block_alloc_h.h"
 
 #define STRUCT
 #include "compilation.h"
@@ -27,12 +26,33 @@
 /* Needed to support dynamic loading on NT */
 PMOD_PROTO extern struct program_state * Pike_compiler;
 
-
-#ifdef PIKE_DEBUG
-#define PROGRAM_LINE_ARGS int line, char *file
-#else
-#define PROGRAM_LINE_ARGS void
-#endif
+/* Compilation flags */
+#define COMPILATION_CHECK_FINAL		0x01
+	/* This flag is set when resolve functions should force the lookup so
+	 * that we don't get a placeholder back. Used for inherits. */
+#define COMPILATION_FORCE_RESOLVE	0x02
+/* #define FORCE_RESOLVE_DEBUG */
+/* Helper macros for force_resolve */
+#ifdef FORCE_RESOLVE_DEBUG
+#define DO_IF_FRD(X)	X
+#else /* !FORCE_RESOLVE_DEBUG */
+#define DO_IF_FRD(X)
+#endif /* FORCE_RESOLVE_DEBUG */
+#define SET_FORCE_RESOLVE(OLD) do {					\
+    int tmp_ = (OLD) = Pike_compiler->flags;				\
+    Pike_compiler->flags |= COMPILATION_FORCE_RESOLVE;			\
+    DO_IF_FRD(fprintf(stderr,						\
+		      "Force resolve on. Flags:0x%04x (0x%04x)\n",	\
+		      Pike_compiler->flags, tmp_));			\
+  } while(0)
+#define UNSET_FORCE_RESOLVE(OLD) do {					\
+    int tmp_ = (Pike_compiler->flags & ~COMPILATION_FORCE_RESOLVE) |	\
+      ((OLD) & COMPILATION_FORCE_RESOLVE);				\
+    DO_IF_FRD(fprintf(stderr,						\
+		      "Force resolve unset. Flags:0x%04x (0x%04x)\n",	\
+		      tmp_, Pike_compiler->flags));			\
+    Pike_compiler->flags = tmp_;					\
+  } while(0)
 
 extern struct pike_string *this_program_string;
 
@@ -79,11 +99,14 @@ extern struct pike_string *this_program_string;
 #define LFUN__SPRINTF 40
 #define LFUN__EQUAL 41
 #define LFUN__M_DELETE 42
+#define LFUN__GET_ITERATOR 43
 
-#define NUM_LFUNS 43
+#define NUM_LFUNS 44
 
 extern char *lfun_names[];
 
+extern struct pike_string *lfun_strings[];
+
 #ifndef STRUCT_SVALUE_DECLARED
 #define STRUCT_SVALUE_DECLARED
 struct svalue;
@@ -99,6 +122,34 @@ struct node_s;
 struct object;
 #endif
 
+/* Byte-code method identification. */
+#define PIKE_BYTECODE_PORTABLE	-1	/* Only used by the codec. */
+#define PIKE_BYTECODE_DEFAULT	0
+#define PIKE_BYTECODE_GOTO	1
+#define PIKE_BYTECODE_SPARC	2
+#define PIKE_BYTECODE_IA32	3
+#define PIKE_BYTECODE_PPC32     4
+
+#if PIKE_BYTECODE_METHOD == PIKE_BYTECODE_IA32
+#define PIKE_OPCODE_T unsigned INT8
+#elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_SPARC
+#define PIKE_OPCODE_T unsigned INT32
+#elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_PPC32
+#define PIKE_OPCODE_T unsigned INT32
+#elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_GOTO
+#define PIKE_OPCODE_T void *
+#define PIKE_INSTR_T void *
+#else
+#define PIKE_OPCODE_T unsigned INT8
+#endif
+
+#ifndef PIKE_INSTR_T
+/* The type for an opcode instruction identifier (not packed). In all
+ * cases but PIKE_BYTECODE_GOTO, this is n - F_OFFSET where n is the
+ * number in the Pike_opcodes enum. */
+#define PIKE_INSTR_T unsigned int
+#endif
+
 /* I need:
  * a) one type that can point to a callable function.
  *    (C function, or object->fun)
@@ -127,26 +178,36 @@ union idptr
   ptrdiff_t offset;
 };
 
+#define IDENTIFIER_VARIABLE 0
 #define IDENTIFIER_PIKE_FUNCTION 1
 #define IDENTIFIER_C_FUNCTION 2
 #define IDENTIFIER_FUNCTION 3
 #define IDENTIFIER_CONSTANT 4
-#define IDENTIFIER_VARARGS 8
-#define IDENTIFIER_PROTOTYPED 16
+#define IDENTIFIER_TYPE_MASK 7
+
+#define IDENTIFIER_VARARGS 8	/* Used for functions only. */
+#define IDENTIFIER_HAS_BODY 16	/* Function has a body (set already in pass 1). */
 #define IDENTIFIER_SCOPED 32   /* This is used for local functions only */
 #define IDENTIFIER_SCOPE_USED 64 /* contains scoped local functions */
+#define IDENTIFIER_ALIAS 128   /* This identifier is an alias. */
 
 #define IDENTIFIER_IS_FUNCTION(X) ((X) & IDENTIFIER_FUNCTION)
 #define IDENTIFIER_IS_PIKE_FUNCTION(X) ((X) & IDENTIFIER_PIKE_FUNCTION)
+#define IDENTIFIER_IS_C_FUNCTION(X) ((X) & IDENTIFIER_C_FUNCTION)
 #define IDENTIFIER_IS_CONSTANT(X) ((X) & IDENTIFIER_CONSTANT)
 #define IDENTIFIER_IS_VARIABLE(X) (!((X) & (IDENTIFIER_FUNCTION | IDENTIFIER_CONSTANT)))
+#define IDENTIFIER_IS_ALIAS(X)	((X) & IDENTIFIER_ALIAS)
 
-#define IDENTIFIER_MASK 127
+#define IDENTIFIER_MASK 255
 
+/*
+ * Every constant, class, function and variable
+ * gets exactly one of these.
+ */
 struct identifier
 {
   struct pike_string *name;
-  struct pike_string *type;
+  struct pike_type *type;
   unsigned INT8 identifier_flags;	/* IDENTIFIER_??? */
   unsigned INT8 run_time_type;		/* PIKE_T_??? */
   unsigned INT16 opt_flags;		/* OPT_??? */
@@ -158,6 +219,11 @@ struct identifier
   union idptr func;
 };
 
+/*
+ * This is used to store constants, both
+ * inline constants and those defined explicitly with 
+ * the constant keyword.
+ */
 struct program_constant
 {
   struct svalue sval;
@@ -174,20 +240,30 @@ struct program_constant
 #define ID_STATIC          0x01	/* Symbol is not visible by indexing */
 #define ID_PRIVATE         0x02	/* Symbol is not visible by inherit */
 #define ID_NOMASK          0x04	/* Symbol may not be overloaded */
-#define ID_PUBLIC          0x08
-#define ID_PROTECTED       0x10
-#define ID_INLINE          0x20
-#define ID_HIDDEN          0x40	/* Symbol is locally hidden. Used e.g.
-				 * for static references like Foo::bar. */
-#define ID_INHERITED       0x80
+#define ID_PUBLIC          0x08 /* Anti private */
+#define ID_PROTECTED       0x10 /* Not currently used at all */
+#define ID_INLINE          0x20 /* Same as local */
+#define ID_HIDDEN          0x40	/* Symbols that are private and inherited one step later */
+#define ID_INHERITED       0x80 /* Symbol is inherited */
 #define ID_OPTIONAL       0x100	/* Symbol is not required by the interface */
 #define ID_EXTERN         0x200	/* Symbol is defined later */
 #define ID_VARIANT	  0x400 /* Function is overloaded by argument. */
+#define ID_ALIAS	  0x800 /* Variable is an overloaded alias. */
 
-#define ID_MODIFIER_MASK 0x07ff
+#define ID_MODIFIER_MASK 0x0fff
 
 #define ID_STRICT_TYPES  0x8000	/* #pragma strict_types */
+#define ID_SAVE_PARENT  0x10000 /* #pragma save_parent */
+#define ID_DONT_SAVE_PARENT 0x20000 /* #pragma dont_save_parent */
 
+
+/*
+ * All identifiers in this program
+ * and all identifiers in inherited programs
+ * needs to have a 'struct reference' in this
+ * program. When we overload a function, we simply
+ * change the reference to point to the new 'struct identifier'.
+ */
 struct reference
 {
   unsigned INT16 inherit_offset;
@@ -195,6 +271,18 @@ struct reference
   INT16 id_flags; /* static, private etc.. */
 };
 
+/* Magic value used as identifier reference integer to refer to this. */
+#define IDREF_MAGIC_THIS -1
+
+/*
+ * Each program has an array of these,
+ * the first entry points to itself, the
+ * rest are from inherited programs.
+ * Note that when a program is inherited,
+ * all 'struct inherit' from that program is
+ * copied, so the whole tree of inherits is
+ * represented.
+ */
 struct inherit
 {
   INT16 inherit_level;
@@ -208,6 +296,11 @@ struct inherit
   struct pike_string *name;
 };
 
+
+/*
+ * Storage struct for a trampoline object
+ * (not a part of the program type)
+ */
 struct pike_trampoline
 {
   struct pike_frame *frame;
@@ -215,6 +308,7 @@ struct pike_trampoline
 };
 
 
+
 /* program parts have been realloced into one block */
 #define PROGRAM_OPTIMIZED 1
 
@@ -232,52 +326,84 @@ struct pike_trampoline
 /* Program will be destructed as soon at it runs out of references.
  * Normally only used for mutex lock keys and similar
  */
-#define PROGRAM_DESTRUCT_IMMEDIATE 16
+#define PROGRAM_DESTRUCT_IMMEDIATE 0x10
 
 /* Self explanatory, automatically detected */
-#define PROGRAM_HAS_C_METHODS 32
+#define PROGRAM_HAS_C_METHODS 0x20
 
-/* All non-static functions are inlinable */
-#define PROGRAM_CONSTANT 64
+/* Objects created from this program are constant and shareable */
+#define PROGRAM_CONSTANT 0x40
 
 /* */
-#define PROGRAM_USES_PARENT 128
+#define PROGRAM_USES_PARENT 0x80
 
-/* Objects should not be destructed even when they only has weak
+/* Objects should not be destructed even when they only have weak
  * references left. */
-#define PROGRAM_NO_WEAK_FREE 256
+#define PROGRAM_NO_WEAK_FREE 0x100
 
 /* Objects should not be destructed by f_destruct(). */
-#define PROGRAM_NO_EXPLICIT_DESTRUCT 512
+#define PROGRAM_NO_EXPLICIT_DESTRUCT 0x200
 
 /* Program is in an inconsistant state */
-#define PROGRAM_AVOID_CHECK 1024
+#define PROGRAM_AVOID_CHECK 0x400
 
+/* Program has not yet been used for compilation */
+#define PROGRAM_VIRGIN 0x800
+
+/* */
+#define PROGRAM_NEEDS_PARENT 0x1000
+
+/* Using define instead of enum allows for ifdefs - Hubbe */
+#define PROG_EVENT_INIT 0
+#define PROG_EVENT_EXIT 1
+#define PROG_EVENT_GC_RECURSE 2
+#define PROG_EVENT_GC_CHECK 3
+#define NUM_PROG_EVENTS 4
+
+/* These macros should only be used if (p->flags & PROGRAM_USES_PARENT)
+ * is true
+ */
+#define LOW_PARENT_INFO(O,P) ((struct parent_info *)(PIKE_OBJ_STORAGE((O)) + (P)->parent_info_storage))
+#define PARENT_INFO(O) LOW_PARENT_INFO( (O), (O)->prog)
+
+/*
+ * Objects which needs to access their parent
+ * have to allocate one of these structs in
+ * the object data area.
+ * The parent_info_storage member of the program
+ * struct tells us where in the object to find this
+ * data.
+ */
+struct parent_info
+{
+  struct object *parent;
+  INT16 parent_identifier;
+};
 
 struct program
 {
-  INT32 refs;
-#ifdef PIKE_SECURITY
-  struct object *prot;
-#endif
+  PIKE_MEMORY_OBJECT_MEMBERS; /* Must be first */
+
   INT32 id;             /* used to identify program in caches */
-  INT32 parent_program_id;
   /* storage_needed - storage needed in object struct
    * the first inherit[0].storage_offset bytes are not used and are
    * subtracted when inheriting.
    */
   ptrdiff_t storage_needed; /* storage needed in the object struct */
+  ptrdiff_t xstorage; /* Non-inherited storage */
+  ptrdiff_t parent_info_storage;
+
   INT16 flags;          /* PROGRAM_* */
   unsigned INT8 alignment_needed;
   struct timeval timestamp;
 
   struct program *next;
   struct program *prev;
+  struct program *parent;
+  
+  struct node_s *(*optimize)(struct node_s *n);
 
-  void (*init)(struct object *);
-  void (*exit)(struct object *);
-  void (*gc_recurse_func)(struct object *);
-  void (*gc_check_func)(struct object *);
+  void (*event_handler)(int);
 #ifdef PIKE_DEBUG
   unsigned INT32 checksum;
 #endif
@@ -304,15 +430,22 @@ struct program
 #define ID_FROM_INT(P,X) ID_FROM_PTR(P,(P)->identifier_references+(X))
 
 #define FIND_LFUN(P,N) ( dmalloc_touch(struct program *,(P))->flags & PROGRAM_FIXED?((P)->lfuns[(N)]):low_find_lfun((P), (N)) )
+#define QUICK_FIND_LFUN(P,N) (dmalloc_touch(struct program *,(P))->lfuns[N])
 
-#define free_program(p) do{ struct program *_=(p); debug_malloc_touch(_); if(!--_->refs) really_free_program(_); }while(0)
+#define free_program(p) do{ struct program *_=(p); debug_malloc_touch(_); if(!sub_ref(_)) really_free_program(_); }while(0)
+
+BLOCK_ALLOC_FILL_PAGES(program, n/a)
 
 
 extern struct object *error_handler;
+extern struct object *compat_handler;
 
 extern struct program *first_program;
+extern struct program *null_program;
 extern struct program *pike_trampoline_program;
 extern struct program *gc_internal_program;
+extern struct program *placeholder_program;
+extern struct object *placeholder_object;
 
 extern int compilation_depth;
 
@@ -329,6 +462,24 @@ extern int compilation_depth;
 #define FOO(NUMTYPE,TYPE,NAME) void PIKE_CONCAT(add_to_,NAME(TYPE ARG));
 #include "program_areas.h"
 
+
+typedef int supporter_callback (void *, int);
+struct Supporter
+{
+#ifdef PIKE_DEBUG
+  int magic;
+#endif
+  struct Supporter *previous;
+  struct Supporter *depends_on;
+  struct Supporter *dependants;
+  struct Supporter *next_dependant;
+  supporter_callback *fun;
+  void *data;
+  struct program *prog;
+};
+
+
+
 /* Prototypes begin here */
 void ins_int(INT32 i, void (*func)(char tmp));
 void ins_short(INT16 i, void (*func)(char tmp));
@@ -337,20 +488,27 @@ void use_module(struct svalue *s);
 void unuse_modules(INT32 howmany);
 struct node_s *find_module_identifier(struct pike_string *ident,
 				      int see_inherit);
+struct node_s *resolve_identifier(struct pike_string *ident);
+struct node_s *program_magic_identifier (struct program_state *state,
+					 int state_depth, int inherit_num,
+					 struct pike_string *ident,
+					 int colon_colon_ref);
 struct program *parent_compilation(int level);
 struct program *id_to_program(INT32 id);
 void optimize_program(struct program *p);
-int program_function_index_compare(const void *a,const void *b);
-char *find_program_name(struct program *p, INT32 *line);
-int override_identifier (struct reference *ref, struct pike_string *name, int cur_id);
+void fsort_program_identifier_index(unsigned short *start,
+				    unsigned short *end,
+				    struct program *p);
+struct pike_string *find_program_name(struct program *p, INT32 *line);
+int override_identifier (struct reference *ref, struct pike_string *name);
 void fixate_program(void);
 struct program *low_allocate_program(void);
 void low_start_new_program(struct program *p,
+			   int pass,
 			   struct pike_string *name,
 			   int flags,
 			   int *idp);
-PMOD_EXPORT void debug_start_new_program(PROGRAM_LINE_ARGS);
-PMOD_EXPORT void really_free_program(struct program *p);
+PMOD_EXPORT void debug_start_new_program(int line, const char *file);
 void dump_program_desc(struct program *p);
 int sizeof_variable(int run_time_type);
 void dump_program_tables (struct program *p, int indent);
@@ -359,17 +517,19 @@ struct program *end_first_pass(int finish);
 PMOD_EXPORT struct program *debug_end_program(void);
 PMOD_EXPORT size_t low_add_storage(size_t size, size_t alignment,
 				   ptrdiff_t modulo_orig);
-PMOD_EXPORT void set_init_callback(void (*init)(struct object *));
-PMOD_EXPORT void set_exit_callback(void (*exit)(struct object *));
+PMOD_EXPORT void set_init_callback(void (*init_callback)(struct object *));
+PMOD_EXPORT void set_exit_callback(void (*exit_callback)(struct object *));
 PMOD_EXPORT void set_gc_recurse_callback(void (*m)(struct object *));
 PMOD_EXPORT void set_gc_check_callback(void (*m)(struct object *));
+void pike_set_prog_event_callback(void (*cb)(int));
+void pike_set_prog_optimize_callback(node *(*opt)(node *));
 int low_reference_inherited_identifier(struct program_state *q,
 				       int e,
 				       struct pike_string *name,
 				       int flags);
 int find_inherit(struct program *p, struct pike_string *name);
 node *reference_inherited_identifier(struct pike_string *super_name,
-				   struct pike_string *function_name);
+				     struct pike_string *function_name);
 void rename_last_inherit(struct pike_string *n);
 void low_inherit(struct program *p,
 		 struct object *parent,
@@ -383,62 +543,63 @@ PMOD_EXPORT void do_inherit(struct svalue *s,
 void compiler_do_inherit(node *n,
 			 INT32 flags,
 			 struct pike_string *name);
+int call_handle_inherit(struct pike_string *s);
 void simple_do_inherit(struct pike_string *s,
 		       INT32 flags,
 		       struct pike_string *name);
 int isidentifier(struct pike_string *s);
 int low_define_variable(struct pike_string *name,
-			struct pike_string *type,
+			struct pike_type *type,
 			INT32 flags,
 			size_t offset,
 			INT32 run_time_type);
-PMOD_EXPORT int map_variable(char *name,
-		 char *type,
+PMOD_EXPORT int map_variable(const char *name,
+		 const char *type,
 		 INT32 flags,
 		 size_t offset,
 		 INT32 run_time_type);
-PMOD_EXPORT int quick_map_variable(char *name,
+PMOD_EXPORT int quick_map_variable(const char *name,
 		       int name_length,
 		       size_t offset,
-		       char *type,
+		       const char *type,
 		       int type_length,
 		       INT32 run_time_type,
 		       INT32 flags);
 int define_variable(struct pike_string *name,
-		    struct pike_string *type,
+		    struct pike_type *type,
 		    INT32 flags);
-PMOD_EXPORT int simple_add_variable(char *name,
-			char *type,
+PMOD_EXPORT int simple_add_variable(const char *name,
+			const char *type,
 			INT32 flags);
 PMOD_EXPORT int add_constant(struct pike_string *name,
 		 struct svalue *c,
 		 INT32 flags);
-PMOD_EXPORT int simple_add_constant(char *name,
+PMOD_EXPORT int simple_add_constant(const char *name,
 			struct svalue *c,
 			INT32 flags);
-PMOD_EXPORT int add_integer_constant(char *name,
+PMOD_EXPORT int add_integer_constant(const char *name,
 			 INT32 i,
 			 INT32 flags);
-PMOD_EXPORT int quick_add_integer_constant(char *name,
+PMOD_EXPORT int quick_add_integer_constant(const char *name,
 			       int name_length,
 			       INT32 i,
 			       INT32 flags);
-PMOD_EXPORT int add_float_constant(char *name,
+PMOD_EXPORT int add_float_constant(const char *name,
 			 double f,
 			 INT32 flags);
-PMOD_EXPORT int add_string_constant(char *name,
-			char *str,
+PMOD_EXPORT int add_string_constant(const char *name,
+			const char *str,
 			INT32 flags);
-PMOD_EXPORT int add_program_constant(char *name,
+PMOD_EXPORT int add_program_constant(const char *name,
 			 struct program *p,
 			 INT32 flags);
-PMOD_EXPORT int add_object_constant(char *name,
+PMOD_EXPORT int add_object_constant(const char *name,
 			struct object *o,
 			INT32 flags);
-PMOD_EXPORT int add_function_constant(char *name, void (*cfun)(INT32), char * type, INT16 flags);
-PMOD_EXPORT int debug_end_class(char *name, ptrdiff_t namelen, INT32 flags);
+PMOD_EXPORT int add_function_constant(const char *name, void (*cfun)(INT32), const char * type, INT16 flags);
+PMOD_EXPORT int debug_end_class(const char *name, ptrdiff_t namelen, INT32 flags);
 INT32 define_function(struct pike_string *name,
-		      struct pike_string *type,
+		      struct pike_type *type,
 		      unsigned INT16 flags,
 		      unsigned INT8 function_flags,
 		      union idptr *func,
@@ -447,12 +608,13 @@ int really_low_find_shared_string_identifier(struct pike_string *name,
 					     struct program *prog,
 					     int flags);
 int low_find_lfun(struct program *p, ptrdiff_t lfun);
+int lfun_lookup_id(struct pike_string *lfun_name);
 int low_find_shared_string_identifier(struct pike_string *name,
 				      struct program *prog);
 struct ff_hash;
 int find_shared_string_identifier(struct pike_string *name,
 				  struct program *prog);
-PMOD_EXPORT int find_identifier(char *name,struct program *prog);
+PMOD_EXPORT int find_identifier(const char *name,struct program *prog);
 int store_prog_string(struct pike_string *str);
 int store_constant(struct svalue *foo,
 		   int equal,
@@ -462,25 +624,48 @@ struct array *program_values(struct program *p);
 void program_index_no_free(struct svalue *to, struct program *p,
 			   struct svalue *ind);
 int get_small_number(char **q);
+void ext_store_program_line (struct program *prog, INT32 line, struct pike_string *file);
 void start_line_numbering(void);
 void store_linenumber(INT32 current_line, struct pike_string *current_file);
-PMOD_EXPORT char *get_line(unsigned char *pc,struct program *prog,INT32 *linep);
+PMOD_EXPORT struct pike_string *low_get_program_line(struct program *prog,
+						     INT32 *linep);
+PMOD_EXPORT struct pike_string *get_program_line(struct program *prog,
+						 INT32 *linep);
+char *debug_get_line(PIKE_OPCODE_T *pc, struct program *prog, INT32 *linep);
+PMOD_EXPORT struct pike_string *low_get_line(PIKE_OPCODE_T *pc,
+					     struct program *prog, INT32 *linep);
+PMOD_EXPORT struct pike_string *get_line(PIKE_OPCODE_T *pc,
+					 struct program *prog, INT32 *linep);
 void my_yyerror(char *fmt,...)  ATTRIBUTE((format(printf,1,2)));
-struct program *compile(struct pike_string *prog,
-			struct object *handler,
-			int major,
-			int minor);
-PMOD_EXPORT int pike_add_function2(char *name, void (*cfun)(INT32),
-		       char *type, unsigned INT8 flags,
-		       unsigned INT16 opt_flags);
-PMOD_EXPORT int quick_add_function(char *name,
+struct pike_string *format_exception_for_error_msg (struct svalue *thrown);
+void yy_describe_exception(struct svalue *thrown);
+struct supporter_marker;
+void verify_supporters(void);
+void init_supporter(struct Supporter *s,
+		    supporter_callback *fun,
+		    void *data);
+int unlink_current_supporter(struct Supporter *c);
+int call_dependants(struct Supporter *s, int finish);
+int report_compiler_dependency(struct program *p);
+struct compilation;
+void run_pass2(struct compilation *c);
+struct program *compile(struct pike_string *aprog,
+			struct object *ahandler,
+			int amajor, int aminor,
+			struct program *atarget,
+			struct object *aplaceholder);
+PMOD_EXPORT int pike_add_function2(const char *name, void (*cfun)(INT32),
+				   const char *type, unsigned INT8 flags,
+				   unsigned INT16 opt_flags);
+PMOD_EXPORT int quick_add_function(const char *name,
 				   int name_length,
 				   void (*cfun)(INT32),
-				   char *type,
+				   const char *type,
 				   int type_length,
 				   unsigned INT8 flags,
 				   unsigned INT16 opt_flags);
 void check_all_programs(void);
+void placeholder_index(INT32 args);
 void init_program(void);
 void cleanup_program(void);
 void gc_mark_program_as_referenced(struct program *p);
@@ -491,16 +676,16 @@ void gc_mark_all_programs(void);
 void gc_cycle_check_all_programs(void);
 void gc_zap_ext_weak_refs_in_programs(void);
 void gc_free_all_unreferenced_programs(void);
-void count_memory_in_programs(INT32 *num_, INT32 *size_);
 void push_compiler_frame(int lexical_scope);
+void low_pop_local_variables(int level);
 void pop_local_variables(int level);
 void pop_compiler_frame(void);
 ptrdiff_t low_get_storage(struct program *o, struct program *p);
 PMOD_EXPORT char *get_storage(struct object *o, struct program *p);
 struct program *low_program_from_function(struct program *p,
 					  INT32 i);
-PMOD_EXPORT struct program *program_from_function(struct svalue *f);
-PMOD_EXPORT struct program *program_from_svalue(struct svalue *s);
+PMOD_EXPORT struct program *program_from_function(const struct svalue *f);
+PMOD_EXPORT struct program *program_from_svalue(const struct svalue *s);
 struct find_child_cache_s;
 int find_child(struct program *parent, struct program *child);
 void yywarning(char *fmt, ...) ATTRIBUTE((format(printf,1,2)));
@@ -510,8 +695,16 @@ PMOD_EXPORT int is_compatible(struct program *a, struct program *b);
 int yyexplain_not_implements(struct program *a, struct program *b, int flags);
 PMOD_EXPORT void *parent_storage(int depth);
 PMOD_EXPORT void change_compiler_compatibility(int major, int minor);
+void make_program_executable(struct program *p);
 /* Prototypes end here */
 
+void really_free_program(struct program *);
+void count_memory_in_programs(INT32*,INT32*);
+
+#ifndef PIKE_USE_MACHINE_CODE
+#define make_program_executable(X)
+#endif
+
 #define ADD_FUNCTION(NAME, FUNC, TYPE, FLAGS) \
   quick_add_function(NAME, CONSTANT_STRLEN(NAME), FUNC, TYPE,\
                      CONSTANT_STRLEN(TYPE), FLAGS, \
@@ -555,6 +748,9 @@ PMOD_EXPORT void change_compiler_compatibility(int major, int minor);
 #define add_function pike_add_function
 #endif
 
+#define ADD_INHERIT(PROGRAM, FLAGS) \
+  low_inherit((PROGRAM), 0, 0, 0, (FLAGS), 0)
+
 #define START_NEW_PROGRAM_ID(ID) do { \
     start_new_program();  \
     Pike_compiler->new_program->id=PIKE_CONCAT3(PROG_,ID,_ID); \
@@ -569,11 +765,7 @@ PMOD_EXPORT void change_compiler_compatibility(int major, int minor);
 #endif
 
 
-#ifdef PIKE_DEBUG
 #define start_new_program() debug_start_new_program(__LINE__,__FILE__)
-#else
-#define start_new_program() debug_start_new_program()
-#endif
 
 #define gc_cycle_check_program(X, WEAK) \
   gc_cycle_enqueue((gc_cycle_check_cb *) real_gc_cycle_check_program, (X), (WEAK))
