@@ -1,9 +1,10 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-/**/
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: stralloc.c,v 1.155 2004/05/18 08:13:15 grubba Exp $
+*/
+
 #include "global.h"
 #include "stralloc.h"
 #include "pike_macros.h"
@@ -16,17 +17,62 @@
 #include "bignum.h"
 #include "interpret.h"
 #include "block_alloc.h"
+#include "operators.h"
 
 #include <errno.h>
 #include <float.h>
 #include <ctype.h>
 #include <math.h>
 
-#ifndef HUGE
-#define HUGE HUGE_VAL
-#endif /*!HUGE*/
+RCSID("$Id: stralloc.c,v 1.155 2004/05/18 08:13:15 grubba Exp $");
 
-RCSID("$Id: stralloc.c,v 1.121 2001/09/28 23:14:09 hubbe Exp $");
+/* #define STRALLOC_USE_PRIMES */
+
+#ifdef STRALLOC_USE_PRIMES 
+
+#define SET_HSIZE(X) htable_size=hashprimes[(X)]
+#define HMODULO(X) ((X) % htable_size)
+
+#else
+
+#define SET_HSIZE(X) htable_mask=(htable_size=(1<<(X)))-1
+#define HMODULO(X) ((X) & (htable_mask))
+
+unsigned INT32 htable_mask;
+
+#endif
+
+
+#if PIKE_RUN_UNLOCKED
+/* Make this bigger when we get lightweight threads */
+#define BUCKET_LOCKS 2048
+static PIKE_MUTEX_T *bucket_locks;
+
+#define BUCKETLOCK(HVAL) \
+ (bucket_locks + (HMODULO(hval__) & (BUCKET_LOCKS-1)))
+
+#define LOCK_BUCKET(HVAL) do {						    \
+  size_t hval__=(HVAL);							    \
+  PIKE_MUTEX_T *bucket_lock;						    \
+  while(1)								    \
+  {									    \
+    bucket_lock=BUCKETLOCK(hval__);                                         \
+    mt_lock(bucket_lock);						    \
+    if(bucket_lock == BUCKETLOCK(hval__))                                   \
+      break;								    \
+    mt_unlock(bucket_lock);						    \
+  }									    \
+}while(0)
+
+#define UNLOCK_BUCKET(HVAL) do {			\
+  size_t hval__=(HVAL);					\
+  mt_unlock(BUCKETLOCK(hval__));                        \
+}while(0)
+
+#else
+#define LOCK_BUCKET(HVAL)
+#define UNLOCK_BUCKET(HVAL)
+#endif
 
 #define BEGIN_HASH_SIZE 997
 #define MAX_AVG_LINK_LENGTH 3
@@ -40,19 +86,20 @@ static unsigned int need_more_hash_prefix=0;
 unsigned INT32 htable_size=0;
 static unsigned int hashprimes_entry=0;
 static struct pike_string **base_table=0;
-static size_t full_hash_value;
 unsigned INT32 num_strings=0;
+PMOD_EXPORT struct pike_string *empty_pike_string = 0;
 
 /*** Main string hash function ***/
 
-
 #define StrHash(s,len) low_do_hash(s,len,0)
 
-static INLINE size_t low_do_hash(const void *s, ptrdiff_t len__,
+static INLINE size_t low_do_hash(const void *s,
+				 ptrdiff_t len__,
 				 int size_shift)
 {
-  DO_HASHMEM(full_hash_value, s, len__<<size_shift, HASH_PREFIX<<size_shift);
-  return full_hash_value % htable_size;
+  size_t h;
+  DO_HASHMEM(h, s, len__<<size_shift, HASH_PREFIX<<size_shift);
+  return h;
 }
 
 static INLINE size_t do_hash(struct pike_string *s)
@@ -101,7 +148,7 @@ static INLINE unsigned INT32 generic_extract (const void *str, int size, ptrdiff
     case 1: return ((unsigned INT16 *)str)[pos];
     case 2: return ((unsigned INT32 *)str)[pos];
   }
-  fatal("Illegal shift size!\n");
+  Pike_fatal("Illegal shift size!\n");
   return 0;
 }
 
@@ -111,11 +158,11 @@ PMOD_EXPORT INLINE unsigned INT32 index_shared_string(struct pike_string *s,
 #ifdef PIKE_DEBUG
   if(pos > s->len || pos<0) {
     if (s->len) {
-      fatal("String index %ld is out of range [0 - %ld]!\n",
+      Pike_fatal("String index %ld is out of range [0 - %ld]!\n",
 	    DO_NOT_WARN((long)pos),
 	    DO_NOT_WARN((long)s->len-1));
     } else {
-      fatal("Attempt to index the empty string with %ld!\n",
+      Pike_fatal("Attempt to index the empty string with %ld!\n",
 	    DO_NOT_WARN((long)pos));
     }
   }
@@ -128,10 +175,10 @@ PMOD_EXPORT INLINE void low_set_index(struct pike_string *s, ptrdiff_t pos,
 {
 #ifdef PIKE_DEBUG
   if(pos > s->len || pos<0)
-    fatal("string index out of range!\n");
+    Pike_fatal("string index out of range!\n");
   
   if(pos == s->len && value)
-    fatal("string zero termination foul!\n");
+    Pike_fatal("string zero termination foul!\n");
 #endif
   switch(s->size_shift)
   {
@@ -139,7 +186,7 @@ PMOD_EXPORT INLINE void low_set_index(struct pike_string *s, ptrdiff_t pos,
     case 1: STR1(s)[pos]=value; break;
     case 2: STR2(s)[pos]=value; break;
     default:
-      fatal("Illegal shift size!\n");
+      Pike_fatal("Illegal shift size!\n");
   }
 }
 
@@ -147,7 +194,7 @@ PMOD_EXPORT INLINE void low_set_index(struct pike_string *s, ptrdiff_t pos,
 PMOD_EXPORT INLINE struct pike_string *debug_check_size_shift(struct pike_string *a,int shift)
 {
   if(a->size_shift != shift)
-    fatal("Wrong STRX macro used!\n");
+    Pike_fatal("Wrong STRX macro used!\n");
   return a;
 }
 #endif
@@ -191,7 +238,7 @@ PMOD_EXPORT void generic_memcpy(PCHARP to,
 {
 #ifdef PIKE_DEBUG
   if(len<0)
-    fatal("Cannot copy %ld bytes!\n",
+    Pike_fatal("Cannot copy %ld bytes!\n",
 	  DO_NOT_WARN((long)len));
 #endif
 
@@ -252,6 +299,7 @@ static void locate_problem(int (*isproblem)(struct pike_string *))
 
   for(e=0;e<htable_size;e++)
   {
+    LOCK_BUCKET(e);
     for(s=base_table[e];s;s=s->next)
     {
       if(isproblem(s))
@@ -263,6 +311,7 @@ static void locate_problem(int (*isproblem)(struct pike_string *))
 	DM(add_marks_to_memhdr(no,s));
       }
     }
+    UNLOCK_BUCKET(e);
   }
 
   DM(fprintf(stderr,"Plausible problem location(s):\n"));
@@ -272,13 +321,18 @@ static void locate_problem(int (*isproblem)(struct pike_string *))
   DM(dump_memhdr_locations(yes,no,0));
 }
 
+static int bad_pointer(struct pike_string *s)
+{
+  return (((ptrdiff_t)s)&(sizeof(struct pike_string *)-1));
+}
+
 static int has_zero_refs(struct pike_string *s)
 {
   return s->refs<=0;
 }
 static int wrong_hash(struct pike_string *s)
 {
-  return (s->hval % htable_size) != do_hash(s);
+  return s->hval != do_hash(s);
 }
 static int improper_zero_termination(struct pike_string *s)
 {
@@ -294,13 +348,15 @@ static int improper_zero_termination(struct pike_string *s)
 static INLINE struct pike_string *internal_findstring(const char *s,
 						      ptrdiff_t len,
 						      int size_shift,
-						      ptrdiff_t h)
+						      size_t hval)
 {
   struct pike_string *curr,**prev, **base;
 #ifndef HASH_PREFIX
   unsigned int depth=0;
 #endif
-
+  size_t h;
+  LOCK_BUCKET(hval);
+  h=HMODULO(hval);
   for(base = prev = base_table + h;( curr=*prev ); prev=&curr->next)
   {
 #ifdef PIKE_DEBUG
@@ -308,12 +364,12 @@ static INLINE struct pike_string *internal_findstring(const char *s,
     {
       debug_dump_pike_string(curr, 70);
       locate_problem(has_zero_refs);
-      fatal("String with no references.\n");
+      Pike_fatal("String with no references.\n");
     }
 #endif
     debug_malloc_touch(curr);
 
-    if (full_hash_value == curr->hval &&
+    if (hval == curr->hval &&
 	len==curr->len &&
 	size_shift==curr->size_shift &&
 	( curr->str == s ||
@@ -322,6 +378,7 @@ static INLINE struct pike_string *internal_findstring(const char *s,
       *prev = curr->next;
       curr->next = *base;
       *base = curr;
+      UNLOCK_BUCKET(hval);
       return curr;		/* pointer to string */
     }
 #ifndef HASH_PREFIX
@@ -339,6 +396,7 @@ static INLINE struct pike_string *internal_findstring(const char *s,
       need_more_hash_prefix--;
   }
 #endif
+  UNLOCK_BUCKET(hval);
   return 0; /* not found */
 }
 
@@ -375,7 +433,7 @@ static struct pike_string *propagate_shared_string(const struct pike_string *s,
     {
       debug_dump_pike_string(curr, 70);
       locate_problem(has_zero_refs);
-      fatal("String with no references.\n");
+      Pike_fatal("String with no references.\n");
     }
 #endif
   }
@@ -389,12 +447,12 @@ static void rehash_string_backwards(struct pike_string *s)
   ptrdiff_t h;
   if(!s) return;
   rehash_string_backwards(s->next);
-  h = s->hval % htable_size;
+  h = HMODULO(s->hval);
   s->next=base_table[h];
   base_table[h]=s;
 }
 
-static void rehash(void)
+static void stralloc_rehash(void)
 {
   int h,old;
   struct pike_string **old_base;
@@ -402,7 +460,24 @@ static void rehash(void)
   old=htable_size;
   old_base=base_table;
 
-  htable_size=hashprimes[++hashprimes_entry];
+#ifdef PIKE_RUN_UNLOCKED
+  mt_lock(bucket_locks);
+  if(old != htable_size)
+  {
+    /* Someone got here before us */
+    mt_lock(bucket_locks);
+    return;
+  }
+
+  /* Now that we have bucket zero, the hash table
+   * cannot change, go ahead and lock ALL buckets.
+   * NOTE: bucket zero is already locked
+   */
+  for(h=1;h<BUCKET_LOCKS;h++) mt_lock(bucket_locks+h);
+#endif
+
+  SET_HSIZE( ++hashprimes_entry );
+
   base_table=(struct pike_string **)xalloc(sizeof(struct pike_string *)*htable_size);
   MEMSET((char *)base_table,0,sizeof(struct pike_string *)*htable_size);
 
@@ -411,6 +486,10 @@ static void rehash(void)
 
   if(old_base)
     free((char *)old_base);
+
+#ifdef PIKE_RUN_UNLOCKED
+  for(h=0;h<BUCKET_LOCKS;h++) mt_unlock(bucket_locks + h);
+#endif
 }
 
 /* Allocation of strings */
@@ -448,7 +527,7 @@ BLOCK_ALLOC(short_pike_string2, SHORT_STRING_BLOCK)
        really_free_short_pike_string1((struct short_pike_string1 *)s); \
      DO_IF_DEBUG( \
      } else if (s->size_shift != 2) { \
-       fatal("Unsupported string shift: %d\n", s->size_shift); \
+       Pike_fatal("Unsupported string shift: %d\n", s->size_shift); \
      ) \
      } else { \
        really_free_short_pike_string2((struct short_pike_string2 *)s); \
@@ -485,15 +564,20 @@ PMOD_EXPORT struct pike_string *debug_begin_shared_string(size_t len)
   return t;
 }
 
-static void link_pike_string(struct pike_string *s, size_t h)
+static void link_pike_string(struct pike_string *s, size_t hval)
 {
+  size_t h;
+  LOCK_BUCKET(hval);
+  h=HMODULO(hval);
   s->refs = 0;
   s->next = base_table[h];
   base_table[h] = s;
-  s->hval=full_hash_value;
+  s->hval=hval;
   num_strings++;
+  UNLOCK_BUCKET(hval);
+
   if(num_strings > MAX_AVG_LINK_LENGTH * htable_size)
-    rehash();
+    stralloc_rehash();
 
 #ifndef HASH_PREFIX
   /* These heuruistics might require tuning! /Hubbe */
@@ -502,8 +586,18 @@ static void link_pike_string(struct pike_string *s, size_t h)
     /* This could in theory have a pretty ugly complexity */
     /* /Hubbe
      */
-    size_t save_full_hash_value = full_hash_value;
-    
+
+#ifdef PIKE_RUN_UNLOCKED
+    mt_lock(bucket_locks);
+    if(need_more_hash_prefix <= ( htable_size >> 4))
+    {
+      /* Someone got here before us */
+      mt_lock(bucket_locks);
+      return;
+    }
+    for(h=1;h<BUCKET_LOCKS;h++) mt_lock(bucket_locks+h);
+#endif
+
     need_more_hash_prefix=0;
     HASH_PREFIX=HASH_PREFIX*2;
 /*    fprintf(stderr,"Doubling HASH_PREFIX to %d and rehashing\n",HASH_PREFIX); */
@@ -518,14 +612,16 @@ static void link_pike_string(struct pike_string *s, size_t h)
 	struct pike_string *tmp2=tmp; /* First unlink */
 	tmp=tmp2->next;
 
-	h2=do_hash(tmp2);	      /* compute new hash value */
-	tmp2->hval=full_hash_value;
+	tmp2->hval=do_hash(tmp2); /* compute new hash value */
+	h2=HMODULO(tmp2->hval);
 
 	tmp2->next=base_table[h2];    /* and re-hash */
 	base_table[h2]=tmp2;
       }
     }
-    full_hash_value = save_full_hash_value;
+#ifdef PIKE_RUN_UNLOCKED
+    for(h=0;h<BUCKET_LOCKS;h++) mt_unlock(bucket_locks + h);
+#endif
   }
 #endif
 }
@@ -545,7 +641,7 @@ PMOD_EXPORT struct pike_string *debug_begin_wide_shared_string(size_t len, int s
       t = (struct pike_string *)alloc_short_pike_string1();
 #ifdef PIKE_DEBUG
     } else if (shift != 2) {
-      fatal("Unsupported string shift: %d\n", shift);
+      Pike_fatal("Unsupported string shift: %d\n", shift);
 #endif /* PIKE_DEBUG */
     } else {
       t = (struct pike_string *)alloc_short_pike_string2();
@@ -574,7 +670,7 @@ PMOD_EXPORT struct pike_string *low_end_shared_string(struct pike_string *s)
   s2 = internal_findstring(s->str, len, s->size_shift, h);
 #ifdef PIKE_DEBUG
   if(s2==s) 
-    fatal("end_shared_string called twice! (or something like that)\n");
+    Pike_fatal("end_shared_string called twice! (or something like that)\n");
 #endif
 
   if(s2)
@@ -601,7 +697,7 @@ PMOD_EXPORT struct pike_string *end_shared_string(struct pike_string *s)
   switch(s->size_shift)
   {
     default:
-      fatal("ARGHEL! size_shift:%d\n", s->size_shift);
+      Pike_fatal("ARGHEL! size_shift:%d\n", s->size_shift);
 
     case 2:
       switch(find_magnitude2(STR2(s),s->len))
@@ -643,7 +739,7 @@ PMOD_EXPORT struct pike_string *end_and_resize_shared_string(struct pike_string
   struct pike_string *tmp;
 #ifdef PIKE_DEBUG
   if(len > str->len)
-    fatal("Cannot extend string here!\n");
+    Pike_fatal("Cannot extend string here!\n");
 #endif
   if( str->len <= SHORT_STRING_THRESHOLD ?
       (len <= SHORT_STRING_THRESHOLD) :
@@ -688,7 +784,7 @@ PMOD_EXPORT struct pike_string * debug_make_shared_binary_pcharp(const PCHARP st
     case 2:
       return make_shared_binary_string2((p_wchar2 *)(str.ptr),  len);
     default:
-      fatal("Unknown string width!\n");
+      Pike_fatal("Unknown string width!\n");
   }
   /* NOT REACHED */
   return NULL;	/* Keep the compiler happy */
@@ -795,13 +891,21 @@ PMOD_EXPORT struct pike_string *debug_make_shared_string2(const p_wchar2 *str)
 
 PMOD_EXPORT void unlink_pike_string(struct pike_string *s)
 {
-  size_t h = s->hval % htable_size;
+  size_t h;
+  LOCK_BUCKET(s->hval);
+  h= HMODULO(s->hval);
   propagate_shared_string(s,h);
+#ifdef PIKE_DEBUG
+  if (base_table[h] != s) {
+    Pike_fatal("propagate_shared_string() failed. Probably got bogus pike_string.\n");
+  }
+#endif /* PIKE_DEBUG */
   base_table[h]=s->next;
 #ifdef PIKE_DEBUG
   s->next=(struct pike_string *)(ptrdiff_t)-1;
 #endif
   num_strings--;
+  UNLOCK_BUCKET(s->hval);
 }
 
 PMOD_EXPORT void do_free_string(struct pike_string *s)
@@ -826,18 +930,23 @@ PMOD_EXPORT void really_free_string(struct pike_string *s)
 {
 #ifdef PIKE_DEBUG
   extern int d_flag;
-  if (d_flag) {
-    if (s->refs) {
-      fatal("Freeing string with references!\n");
-    }
-    if(d_flag > 2)
-    {
-      if(s->next == (struct pike_string *)(ptrdiff_t)-1)
-	fatal("Freeing shared string again!\n");
+  if (s->refs) {
+#ifdef DEBUG_MALLOC
+    describe_something(s, T_STRING, 0,2,0, NULL);
+#endif
+    Pike_fatal("Freeing string with %d references.\n", s->refs);
+  }
+  if(d_flag > 2)
+  {
+    if(s->next == (struct pike_string *)(ptrdiff_t)-1)
+      Pike_fatal("Freeing shared string again!\n");
 
-      if(((ptrdiff_t)s->next) & 1)
-	fatal("Freeing shared string again, memory corrupt or other bug!\n");
-    }
+    if(((ptrdiff_t)s->next) & 1)
+      Pike_fatal("Freeing shared string again, memory corrupt or other bug!\n");
+  }
+  if ((s->size_shift < 0) || (s->size_shift > 2)) {
+    Pike_fatal("Freeing string with bad shift (0x%08x); could it be a type?\n",
+	  s->size_shift);
   }
 #endif
   unlink_pike_string(s);
@@ -872,6 +981,7 @@ struct pike_string *add_string_status(int verbose)
     struct pike_string *p;
     for(e=0;e<htable_size;e++)
     {
+      LOCK_BUCKET(e);
       for(p=base_table[e];p;p=p->next)
       {
 	num_distinct_strings++;
@@ -879,7 +989,7 @@ struct pike_string *add_string_status(int verbose)
 	allocd_strings+=p->refs;
 	allocd_bytes+=p->refs*DO_ALIGN(p->len+3,sizeof(void *));
       }
-
+      UNLOCK_BUCKET(e);
     }
     overhead_bytes=(sizeof(struct pike_string)-1)*num_distinct_strings;
     my_strcat("\nShared string hash table:\n");
@@ -909,75 +1019,90 @@ struct pike_string *add_string_status(int verbose)
 /*** PIKE_DEBUG ***/
 #ifdef PIKE_DEBUG
 
+static long last_stralloc_verify=0;
+extern long current_do_debug_cycle;
+
 PMOD_EXPORT void check_string(struct pike_string *s)
 {
-  do_hash(s);
-  if(full_hash_value != s->hval)
+  if(current_do_debug_cycle == last_stralloc_verify)
   {
-    locate_problem(wrong_hash);
-    fatal("Hash value changed?\n");
-  }
+    if(debug_findstring(s) !=s)
+      Pike_fatal("Shared string not shared.\n");
+  }else{
+    if(do_hash(s) != s->hval)
+    {
+      locate_problem(wrong_hash);
+      Pike_fatal("Hash value changed?\n");
+    }
+    
+    if(debug_findstring(s) !=s)
+      Pike_fatal("Shared string not shared.\n");
 
-  if(debug_findstring(s) !=s)
-    fatal("Shared string not shared.\n");
-
-  if(index_shared_string(s,s->len))
-  {
-    locate_problem(improper_zero_termination);
-    fatal("Shared string is not zero terminated properly.\n");
+    if(index_shared_string(s,s->len))
+    {
+      locate_problem(improper_zero_termination);
+      Pike_fatal("Shared string is not zero terminated properly.\n");
+    }
   }
 }
 
 PMOD_EXPORT void verify_shared_strings_tables(void)
 {
   unsigned INT32 e, h, num=0;
-  size_t orig_full_hash = full_hash_value;
   struct pike_string *s;
 
+  last_stralloc_verify=current_do_debug_cycle;
+
   for(e=0;e<htable_size;e++)
   {
     h=0;
+    LOCK_BUCKET(e);
     for(s=base_table[e];s;s=s->next)
     {
       num++;
       h++;
+
+      if (bad_pointer(s)) {
+	Pike_fatal("Odd string pointer in string table!\n");
+      }
+
       if(s->len < 0)
-	fatal("Shared string shorter than zero bytes.\n");
+	Pike_fatal("Shared string shorter than zero bytes.\n");
 
       if(s->refs <= 0)
       {
 	locate_problem(has_zero_refs);
-	fatal("Shared string had too few references.\n");
+	Pike_fatal("Shared string had too few references.\n");
       }
 
       if(index_shared_string(s,s->len))
       {
 	locate_problem(improper_zero_termination);
-	fatal("Shared string didn't end with a zero.\n");
+	Pike_fatal("Shared string didn't end with a zero.\n");
       }
 
-      if(do_hash(s) != e)
+      if(do_hash(s) != s->hval)
+	Pike_fatal("Shared string hashed to other number.\n");
+
+      if(HMODULO(s->hval) != e)
       {
 	locate_problem(wrong_hash);
-	fatal("Shared string hashed to wrong place.\n");
+	Pike_fatal("Shared string hashed to wrong place.\n");
       }
 
-      if(s->hval != full_hash_value)
-	fatal("Shared string hashed to other number.\n");
-	
       if(h>10000)
       {
 	struct pike_string *s2;
 	for(s2=s;s2;s2=s2->next)
 	  if(s2 == s)
-	    fatal("Shared string table is cyclic.\n");
+	    Pike_fatal("Shared string table is cyclic.\n");
 	h=0;
       }
     }
+    UNLOCK_BUCKET(e);
   }
   if(num != num_strings)
-    fatal("Num strings is wrong %d!=%d\n",num,num_strings);
-  full_hash_value = orig_full_hash;
+    Pike_fatal("Num strings is wrong %d!=%d\n",num,num_strings);
 }
 
 PMOD_EXPORT int safe_debug_findstring(struct pike_string *foo)
@@ -987,8 +1112,16 @@ PMOD_EXPORT int safe_debug_findstring(struct pike_string *foo)
   for(e=0;e<htable_size;e++)
   {
     struct pike_string *p;
+    LOCK_BUCKET(e);
     for(p=base_table[e];p;p=p->next)
-      if(p==foo) return 1;
+    {
+      if(p==foo)
+      {
+	UNLOCK_BUCKET(e);
+	return 1;
+      }
+    }
+    UNLOCK_BUCKET(e);
   }
   return 0;
 }
@@ -996,7 +1129,7 @@ PMOD_EXPORT int safe_debug_findstring(struct pike_string *foo)
 PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo)
 {
   struct pike_string *tmp;
-  tmp=propagate_shared_string(foo, foo->hval % htable_size);
+  tmp=propagate_shared_string(foo, HMODULO(foo->hval));
 
 #if 0
   if(!tmp)
@@ -1008,11 +1141,12 @@ PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo)
 	    (long)foo->hval,
 	    (long)foo->len,
 	    foo->str);
-    StrHash(foo->str,foo->len);
+
+    LOCK_BUCKET(foo->hval);
     fprintf(stderr,"------ %p %ld\n",
-	    base_table[foo->hval %htable_size],
-	    (long)full_hash_value);
-    for(tmp2=base_table[foo->hval % htable_size];tmp2;tmp2=tmp2->next)
+	    base_table[HMODULO(foo->hval)],
+	    foo->hval);
+    for(tmp2=base_table[HMODULO(foo->hval)];tmp2;tmp2=tmp2->next)
     {
       if(tmp2 == tmp)
 	fprintf(stderr,"!!%p!!->",tmp2);
@@ -1020,16 +1154,19 @@ PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo)
 	fprintf(stderr,"%p->",tmp2);
     }
     fprintf(stderr,"0\n");
+    UNLOCK_BUCKET(foo->hval);
 
     for(e=0;e<htable_size;e++)
     {
+      LOCK_BUCKET(e);
       for(tmp2=base_table[e];tmp2;tmp2=tmp2->next)
       {
 	if(tmp2 == tmp)
 	  fprintf(stderr,"String found in hashbin %ld (not %ld)\n",
 		  (long)e,
-		  (long)(foo->hval % htable_size));
+		  (long)HMODULO(foo->hval));
       }
+      UNLOCK_BUCKET(e);
     }
   }
 #endif
@@ -1079,8 +1216,12 @@ void dump_stralloc_strings(void)
   unsigned INT32 e;
   struct pike_string *p;
   for(e=0;e<htable_size;e++)
+  {
+    LOCK_BUCKET(e);
     for(p=base_table[e];p;p=p->next)
       debug_dump_pike_string(p, 70);
+    UNLOCK_BUCKET(e);
+  }
 }
 
 #endif
@@ -1146,29 +1287,21 @@ PMOD_EXPORT int c_compare_string(struct pike_string *s, char *foo, int len)
   return s->len == len && s->size_shift == 0 && !MEMCMP(s->str,foo,len);
 }
 
-#ifndef HAVE_STRCOLL
-/* No locale function available */
-static int low_binary_strcmp(char *a, ptrdiff_t alen,
-			     char *b, ptrdiff_t blen)
-{
-  low_quick_binary_strcmp(a,alen,b,blen);
-}
-#else
-
+#if defined(HAVE_STRCOLL) && !defined(DONT_USE_SYSTEM_LOCALE)
 /* takes locale into account */
 static int low_binary_strcmp(char *a, ptrdiff_t alen,
 			     char *b, ptrdiff_t blen)
 {
-  INT32 tmp;
   while(alen>0 && blen>0)
   {
-    tmp=strcoll(a,b);
-    if(tmp) return (int)tmp;
-    tmp=strlen(a)+1;
-    a+=tmp;
-    b+=tmp;
-    alen-=tmp;
-    blen-=tmp;
+    int tmp1 = strcoll(a,b);
+    ptrdiff_t tmp2;
+    if(tmp1) return (int)tmp1;
+    tmp2 = strlen(a)+1;
+    a += tmp2;
+    b += tmp2;
+    alen -= tmp2;
+    blen -= tmp2;
   }
   if(alen==blen) return 0;
   if(alen > blen) return 1;
@@ -1190,6 +1323,9 @@ PMOD_EXPORT ptrdiff_t my_quick_strcmp(struct pike_string *a,
 /* Does take locale into account */
 PMOD_EXPORT ptrdiff_t my_strcmp(struct pike_string *a,struct pike_string *b)
 {
+#if !defined(HAVE_STRCOLL) || defined(DONT_USE_SYSTEM_LOCALE)
+  return my_quick_strcmp(a, b);
+#else /* HAVE_STRCOLL && !DONT_USE_SYSTEM_LOCALE */
   if(a==b) return 0;
 
   switch(TWO_SIZES(a->size_shift,b->size_shift))
@@ -1205,7 +1341,6 @@ PMOD_EXPORT ptrdiff_t my_strcmp(struct pike_string *a,struct pike_string *b)
 	INT32 ac=index_shared_string(a,e);
 	INT32 bc=index_shared_string(b,e);
 
-#ifdef HAVE_STRCOLL
 	if(ac < 256 && bc < 256)
 	{
 	  char atmp[2],btmp[2];
@@ -1217,12 +1352,12 @@ PMOD_EXPORT ptrdiff_t my_strcmp(struct pike_string *a,struct pike_string *b)
 	  if((tmp=strcoll(atmp,btmp)))
 	     return tmp;
 	}else
-#endif
 	  if(ac-bc) return ac-bc;
       }
       return a->len - b->len;
     }
   }
+#endif /* !HAVE_STRCOLL || DONT_USE_SYSTEM_LOCALE */
 }
 
 PMOD_EXPORT struct pike_string *realloc_unlinked_string(struct pike_string *a,
@@ -1302,7 +1437,7 @@ PMOD_EXPORT struct pike_string *modify_shared_string(struct pike_string *a,
   INT32 old_value;
 #ifdef PIKE_DEBUG
   if(index<0 || index>=a->len)
-    fatal("Index out of range in modify_shared_string()\n");
+    Pike_fatal("Index out of range in modify_shared_string()\n");
 #endif
 
 
@@ -1343,7 +1478,7 @@ PMOD_EXPORT struct pike_string *modify_shared_string(struct pike_string *a,
 	return end_shared_string(b);
 
       default:
-	fatal("Odd wide string conversion!\n");
+	Pike_fatal("Odd wide string conversion!\n");
     }
   }
 
@@ -1363,7 +1498,7 @@ PMOD_EXPORT struct pike_string *modify_shared_string(struct pike_string *a,
     switch(a->size_shift)
     {
       case 0:
-	fatal("Unshrinkable!\n");
+	Pike_fatal("Unshrinkable!\n");
 
       case 1:
 	/* Test if we *actually* can shrink it.. */
@@ -1470,13 +1605,14 @@ PMOD_EXPORT ptrdiff_t string_search(struct pike_string *haystack,
 				    ptrdiff_t start)
 {
   SearchMojt mojt;
-  struct generic_mem_searcher s;
   char *r;
 
   if(needle->size_shift > haystack->size_shift ||
      start + needle->len > haystack->len)
     return -1;
 
+  if(!needle->len) return start;
+
   mojt=compile_memsearcher(MKPCHARP_STR(needle),
 			   needle->len,
 			   haystack->len,
@@ -1492,7 +1628,7 @@ PMOD_EXPORT ptrdiff_t string_search(struct pike_string *haystack,
 #ifdef PIKE_DEBUG
   if((r < haystack->str) ||
      (r - haystack->str)>>haystack->size_shift > haystack->len)
-    fatal("string_search did a bobo!\n");
+    Pike_fatal("string_search did a bobo!\n");
 #endif
   return (r-haystack->str)>>haystack->size_shift;
 }
@@ -1504,7 +1640,7 @@ PMOD_EXPORT struct pike_string *string_slice(struct pike_string *s,
 #ifdef PIKE_DEBUG
   if(start < 0 || len<0 || start+len>s->len )
   {
-    fatal("string_slice, start = %ld, len = %ld, s->len = %ld\n",
+    Pike_fatal("string_slice, start = %ld, len = %ld, s->len = %ld\n",
 	  DO_NOT_WARN((long)start),
 	  DO_NOT_WARN((long)len),
 	  DO_NOT_WARN((long)s->len));
@@ -1528,7 +1664,7 @@ PMOD_EXPORT struct pike_string *string_slice(struct pike_string *s,
     case 2:
       return make_shared_binary_string2(STR2(s)+start,len);
   }
-  fatal("Illegal shift size!\n");
+  Pike_fatal("Illegal shift size!\n");
   return 0;
 }
 
@@ -1583,7 +1719,7 @@ PMOD_EXPORT struct pike_string *string_replace(struct pike_string *str,
       case 1: f=(replace_searchfunc)mojt.vtab->func1; break;
       case 2: f=(replace_searchfunc)mojt.vtab->func2; break;
 #ifdef PIKE_DEBUG
-      default: fatal("Illegal shift.\n");
+      default: Pike_fatal("Illegal shift.\n");
 #endif
     }
 
@@ -1600,7 +1736,7 @@ PMOD_EXPORT struct pike_string *string_replace(struct pike_string *str,
       case 1: f=(replace_searchfunc)mojt.vtab->func1; break;
       case 2: f=(replace_searchfunc)mojt.vtab->func2; break;
 #ifdef PIKE_DEBUG
-      default: fatal("Illegal shift.\n");
+      default: Pike_fatal("Illegal shift.\n");
 #endif
     }
 
@@ -1626,7 +1762,7 @@ PMOD_EXPORT struct pike_string *string_replace(struct pike_string *str,
   {
 #ifdef PIKE_DEBUG
     if(tmp + (del->len << str->size_shift) > end)
-      fatal("generic_memory_search found a match beyond end of string!!!\n");
+      Pike_fatal("generic_memory_search found a match beyond end of string!!!\n");
 #endif
     generic_memcpy(r,MKPCHARP(s,str->size_shift),(tmp-s)>>str->size_shift);
     INC_PCHARP(r,(tmp-s)>>str->size_shift);
@@ -1643,10 +1779,23 @@ PMOD_EXPORT struct pike_string *string_replace(struct pike_string *str,
 /*** init/exit memory ***/
 void init_shared_string_table(void)
 {
+  init_short_pike_string0_blocks();
+  init_short_pike_string1_blocks();
+  init_short_pike_string2_blocks();
   for(hashprimes_entry=0;hashprimes[hashprimes_entry]<BEGIN_HASH_SIZE;hashprimes_entry++);
-  htable_size=hashprimes[hashprimes_entry];
+  SET_HSIZE(hashprimes_entry);
   base_table=(struct pike_string **)xalloc(sizeof(struct pike_string *)*htable_size);
   MEMSET((char *)base_table,0,sizeof(struct pike_string *)*htable_size);
+#ifdef PIKE_RUN_UNLOCKED
+  {
+    int h;
+    bucket_locks=(PIKE_MUTEX_T *)xalloc(sizeof(PIKE_MUTEX_T)*BUCKET_LOCKS);
+    for(h=0;h<BUCKET_LOCKS;h++) mt_init(bucket_locks + h);
+  }
+#endif
+  empty_pike_string = make_shared_string("");
+
+  initialize_global_buf();
 }
 
 #ifdef DEBUG_MALLOC
@@ -1659,6 +1808,11 @@ void cleanup_shared_string_table(void)
   unsigned INT32 e;
   struct pike_string *s,*next;
 
+  if (empty_pike_string) {
+    free_string(empty_pike_string);
+    empty_pike_string = 0;
+  }
+
 #if defined(PIKE_DEBUG) && defined(DEBUG_MALLOC)
   while(all_shared_string_locations)
   {
@@ -1682,6 +1836,7 @@ void cleanup_shared_string_table(void)
 
   for(e=0;e<htable_size;e++)
   {
+    LOCK_BUCKET(e);
     for(s=base_table[e];s;s=next)
     {
       next=s->next;
@@ -1692,16 +1847,17 @@ void cleanup_shared_string_table(void)
 #endif
     }
     base_table[e]=0;
+    UNLOCK_BUCKET(e);
   }
   free((char *)base_table);
   base_table=0;
   num_strings=0;
 
-#ifdef DEBUG_MALLOC
+#ifdef DO_PIKE_CLEANUP
   free_all_short_pike_string0_blocks();
   free_all_short_pike_string1_blocks();
   free_all_short_pike_string2_blocks();
-#endif /* DEBUG_MALLOC */
+#endif /* DO_PIKE_CLEANUP */
 }
 
 void count_memory_in_strings(INT32 *num, INT32 *size)
@@ -1716,15 +1872,17 @@ void count_memory_in_strings(INT32 *num, INT32 *size)
   for(e=0;e<htable_size;e++)
   {
     struct pike_string *p;
+    LOCK_BUCKET(e);
     for(p=base_table[e];p;p=p->next)
     {
       num_++;
       size_+=sizeof(struct pike_string)+(p->len<<p->size_shift);
     }
+    UNLOCK_BUCKET(e);
   }
 #ifdef PIKE_DEBUG
   if(num_strings != num_)
-    fatal("Num strings is wrong! %d!=%d.\n",num_strings, num_);
+    Pike_fatal("Num strings is wrong! %d!=%d.\n",num_strings, num_);
 #endif
   num[0]=num_;
   size[0]=size_;
@@ -1743,7 +1901,6 @@ unsigned gc_touch_all_strings(void)
   }
   return n;
 }
-#endif
 
 void gc_mark_all_strings(void)
 {
@@ -1755,6 +1912,23 @@ void gc_mark_all_strings(void)
     for(p=base_table[e];p;p=p->next) gc_is_referenced(p);
   }
 }
+#endif
+
+struct pike_string *next_pike_string (struct pike_string *s)
+{
+  struct pike_string *next = s->next;
+  if (!next) {
+    size_t h = s->hval;
+    do {
+      h++;
+      LOCK_BUCKET(h);
+      h = HMODULO(h);
+      next = base_table[h];
+      UNLOCK_BUCKET(h);
+    } while (!next);
+  }
+  return next;
+}
 
 PMOD_EXPORT void init_string_builder(struct string_builder *s, int mag)
 {
@@ -1840,7 +2014,7 @@ PMOD_EXPORT void string_builder_binary_strcat(struct string_builder *s, char *st
     case 1: convert_0_to_1(STR1(s->s)+s->s->len,(p_wchar0 *)str,len); break;
     case 2: convert_0_to_2(STR2(s->s)+s->s->len,(p_wchar0 *)str,len); break;
     default:
-      fatal("Illegal magnitude!\n");
+      Pike_fatal("Illegal magnitude!\n");
   }
   s->s->len+=len;
   /* Ensure NUL-termination */
@@ -1878,7 +2052,7 @@ PMOD_EXPORT void string_builder_fill(struct string_builder *s,
 
 #ifdef PIKE_DEBUG
   if(len<=0)
-    fatal("Cannot fill with zero length strings!\n");
+    Pike_fatal("Cannot fill with zero length strings!\n");
 #endif
   if(howmany<=0) return;
 
@@ -1914,7 +2088,6 @@ PMOD_EXPORT void string_builder_fill(struct string_builder *s,
   howmany-=tmp;
   if(howmany > 0)
   {
-    void *new_from;
     PCHARP to;
     tmp=MINIMUM(howmany, len);
     to=MKPCHARP_STR_OFF(s->s,s->s->len);
@@ -1990,7 +2163,7 @@ PMOD_EXPORT PCHARP MEMCHR_PCHARP(PCHARP ptr, int chr, ptrdiff_t len)
     case 1: return MKPCHARP(MEMCHR1((p_wchar1 *)ptr.ptr,chr,len),1);
     case 2: return MKPCHARP(MEMCHR2((p_wchar2 *)ptr.ptr,chr,len),2);
   }
-  fatal("Illegal shift in MEMCHR_PCHARP.\n");
+  Pike_fatal("Illegal shift in MEMCHR_PCHARP.\n");
   return MKPCHARP(0,0); /* make wcc happy */
 }
 
@@ -2000,15 +2173,17 @@ PMOD_EXPORT PCHARP MEMCHR_PCHARP(PCHARP ptr, int chr, ptrdiff_t len)
 
 PMOD_EXPORT long STRTOL_PCHARP(PCHARP str, PCHARP *ptr, int base)
 {
-  register long val;
-  register int c;
-  int xx, neg = 0;
+  /* Note: Code duplication in STRTOL and pcharp_to_svalue_inumber. */
+
+  unsigned long val, mul_limit;
+  int c;
+  int xx, neg = 0, add_limit, overflow = 0;
 
   if (ptr)  *ptr = str;
   if (base < 0 || base > MBASE)  return 0;
   if (!WIDE_ISALNUM(c = EXTRACT_PCHARP(str)))
   {
-    while (ISSPACE(c))
+    while (WIDE_ISSPACE(c))
     {
       INC_PCHARP(str,1);
       c=EXTRACT_PCHARP(str);
@@ -2041,16 +2216,41 @@ PMOD_EXPORT long STRTOL_PCHARP(PCHARP str, PCHARP *ptr, int base)
     INC_PCHARP(str,2);
     c = EXTRACT_PCHARP(str);		/* skip over leading "0x" or "0X" */
   }
-  val=-DIGIT(c);
+
+  if (neg) {
+    mul_limit = (unsigned long) LONG_MIN / base;
+    add_limit = (int) ((unsigned long) LONG_MIN % base);
+  }
+  else {
+    mul_limit = LONG_MAX / base;
+    add_limit = (int) (LONG_MAX % base);
+  }
+
+  val=DIGIT(c);
   while(1)
   {
     INC_PCHARP(str,1);
     c=EXTRACT_PCHARP(str);
     if(!(WIDE_ISALNUM(c)  && (xx=DIGIT(c)) < base)) break;
-    val = base * val - xx;
+    if (val > mul_limit || (val == mul_limit && xx > add_limit))
+      overflow = 1;
+    else
+      val = base * val + xx;
   }
+
   if (ptr) *ptr = str;
-  return (neg ? val : -val);
+  if (overflow) {
+    errno = ERANGE;
+    return neg ? LONG_MIN : LONG_MAX;
+  }
+  else {
+    if (neg)
+      return val > (unsigned long) LONG_MAX ?
+	-(long) (val - (unsigned long) LONG_MAX) - LONG_MAX :
+	-(long) val;
+    else
+      return (long) val;
+  }
 }
 
 PMOD_EXPORT int string_to_svalue_inumber(struct svalue *r,
@@ -2092,11 +2292,13 @@ PMOD_EXPORT int pcharp_to_svalue_inumber(struct svalue *r,
 					 int base,
 					 ptrdiff_t maxlength)
 {
+  /* Note: Code duplication in STRTOL and STRTOL_PCHARP. */
+
   PCHARP str_start;
   
-  INT_TYPE xx, neg = 0, is_bignum = 0, implicit_base = 0;
-  INT_TYPE val;
-  INT_TYPE c;
+  unsigned INT_TYPE val, mul_limit;
+  int c;
+  int xx, neg = 0, add_limit, overflow = 0;
 
   maxlength--;   /* max_length <= 0 means no max length. */
   str_start = str;
@@ -2132,8 +2334,6 @@ PMOD_EXPORT int pcharp_to_svalue_inumber(struct svalue *r,
   
   if(base == 0)
   {
-    implicit_base = 1;
-    
     if(c != '0')
       base = 10;
     else if(INDEX_PCHARP(str,1) == 'x' || INDEX_PCHARP(str,1) == 'X')
@@ -2151,7 +2351,7 @@ PMOD_EXPORT int pcharp_to_svalue_inumber(struct svalue *r,
   if(!WIDE_ISALNUM(c) || (xx = DIGIT(c)) >= base)
     return 0;   /* No number formed. */
   
-  if(implicit_base && c == '0' &&
+  if((base==16 || base == 2) && c == '0' &&
      INDEX_PCHARP(str,2) < 256 && /* Don't trust isxdigit... */
      isxdigit(INDEX_PCHARP(str,2)) &&
      ((base==16 && (INDEX_PCHARP(str,1)=='x' || INDEX_PCHARP(str,1)=='X')) ||
@@ -2161,63 +2361,61 @@ PMOD_EXPORT int pcharp_to_svalue_inumber(struct svalue *r,
     INC_PCHARP(str,2);
     c=EXTRACT_PCHARP(str);
   }
+  str_start=str;
+
+  if (neg) {
+    mul_limit = (unsigned INT_TYPE) MIN_INT_TYPE / base;
+    add_limit = (int) ((unsigned INT_TYPE) MIN_INT_TYPE % base);
+  }
+  else {
+    mul_limit = MAX_INT_TYPE / base;
+    add_limit = (int) (MAX_INT_TYPE % base);
+  }
   
-  for(val = -DIGIT(c);
+  for(val = DIGIT(c);
       (INC_PCHARP(str,1), WIDE_ISALNUM(c = EXTRACT_PCHARP(str) )) &&
 	(xx = DIGIT(c)) < base &&
 	0 != maxlength--; )
   {
-#ifdef AUTO_BIGNUM
-    if(INT_TYPE_MUL_OVERFLOW(val, base))
-      is_bignum = 1;
-#endif /* AUTO_BIGNUM */
-    val = base * val;
-    /* Accumulating a negative value avoids surprises near MIN_TYPE_INT. */
-#ifdef AUTO_BIGNUM
-    if(INT_TYPE_SUB_OVERFLOW(val, xx))
-      is_bignum = 1;
-#endif /* AUTO_BIGNUM */
-    val -= xx;
+    if (val > mul_limit || (val == mul_limit && xx > add_limit))
+      overflow = 1;
+    else
+      val = base * val + xx;
   }
   
   if(ptr != 0)
     *ptr = str;
 
-  if(neg)
-    r->u.integer = val;
-  else
-  {
+  if (overflow) {
 #ifdef AUTO_BIGNUM
-    if(INT_TYPE_NEG_OVERFLOW(val))
-      is_bignum = 1;
-#endif /* AUTO_BIGNUM */
-    r->u.integer = -val;
-  }
-
-#ifdef AUTO_BIGNUM
-  if(is_bignum)
-  {
     push_string(make_shared_binary_pcharp(str_start,
 					  SUBTRACT_PCHARP(str,str_start)));
-    /* Note that this can concievably throw errors()
+    /* Note that this can conceivably throw errors()
      * in some situations that might not be desirable...
      * take care.
      * /Hubbe
+     *
+     * It could probably also be faster...
      */
-    if(implicit_base)
-    {
-      convert_stack_top_to_bignum();
-    }
-    else
-    {
-      push_int(base);
-      convert_stack_top_with_base_to_bignum();
-    }
+    push_int(base);
+    convert_stack_top_with_base_to_bignum();
+    if(neg) o_negate();
     
-    *r = *--sp;
+    *r = *--Pike_sp;
+    dmalloc_touch_svalue (r);
+#else  /* !AUTO_BIGNUM */
+    r->u.integer = neg ? MIN_INT_TYPE : MAX_INT_TYPE;
+#endif
   }
-#endif /* AUTO_BIGNUM */
-  
+  else {
+    if (neg)
+      r->u.integer = val > (unsigned INT_TYPE) MAX_INT_TYPE ?
+	-(INT_TYPE) (val - (unsigned INT_TYPE) MAX_INT_TYPE) - MAX_INT_TYPE :
+	-(INT_TYPE) val;
+    else
+      r->u.integer = (INT_TYPE) val;
+  }
+
   return 1;
 }
 
@@ -2226,13 +2424,13 @@ PMOD_EXPORT int convert_stack_top_string_to_inumber(int base)
   struct svalue r;
   int i;
 
-  if(sp[-1].type != T_STRING)
+  if(Pike_sp[-1].type != T_STRING)
     Pike_error("Cannot convert stack top to integer number.\n");
   
-  i=pcharp_to_svalue_inumber(&r, MKPCHARP_STR(sp[-1].u.string), 0, base, 0);
+  i=pcharp_to_svalue_inumber(&r, MKPCHARP_STR(Pike_sp[-1].u.string), 0, base, 0);
   
-  free_string(sp[-1].u.string);
-  sp[-1] = r;
+  free_string(Pike_sp[-1].u.string);
+  Pike_sp[-1] = r;
 
   return i;
 }
@@ -2241,6 +2439,8 @@ PMOD_EXPORT int convert_stack_top_string_to_inumber(int base)
    character after the last one used in the number is put in *ENDPTR.  */
 PMOD_EXPORT double STRTOD_PCHARP(PCHARP nptr, PCHARP *endptr)
 {
+  /* Note: Code duplication in STRTOD. */
+
   register PCHARP s;
   short int sign;
 
@@ -2370,12 +2570,14 @@ PMOD_EXPORT double STRTOD_PCHARP(PCHARP nptr, PCHARP *endptr)
  overflow:
   /* Return an overflow error.  */
   errno = ERANGE;
-  return HUGE * sign;
+  return HUGE_VAL * sign;
 
  underflow:
   /* Return an underflow error.  */
+#if 0
   if (endptr != NULL)
     *endptr = nptr;
+#endif
   errno = ERANGE;
   return 0.0;
   
@@ -2400,7 +2602,7 @@ PMOD_EXPORT p_wchar0 *require_wstring0(struct pike_string *s,
       return 0;
 
     default:
-      fatal("Illegal shift size in string.\n");
+      Pike_fatal("Illegal shift size in string.\n");
   }
   return 0;
 }
@@ -2423,7 +2625,7 @@ PMOD_EXPORT p_wchar1 *require_wstring1(struct pike_string *s,
       return 0;
 
     default:
-      fatal("Illegal shift size in string.\n");
+      Pike_fatal("Illegal shift size in string.\n");
   }
   return 0;
 }
@@ -2449,8 +2651,7 @@ PMOD_EXPORT p_wchar2 *require_wstring2(struct pike_string *s,
       return STR2(s);
 
     default:
-      fatal("Illegal shift size in string.\n");
+      Pike_fatal("Illegal shift size in string.\n");
   }
   return 0;
 }
-
