@@ -1,12 +1,10 @@
-/*\
-||| This file a part of Pike, and is copyright by Fredrik Hubinette
-||| Pike is distributed as GPL (General Public License)
-||| See the files COPYING and DISCLAIMER for more information.
-\*/
-
 /*
- * $Id: svalue.h,v 1.77 2001/09/28 23:18:56 hubbe Exp $
- */
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: svalue.h,v 1.116 2003/09/09 14:58:54 mast Exp $
+*/
+
 #ifndef SVALUE_H
 #define SVALUE_H
 
@@ -52,9 +50,12 @@ struct processing
   struct processing *next;
   void *pointer_a, *pointer_b;
 };
+   
+struct ref_dummy;
 
 union anything
 {
+  INT_TYPE integer;		/* Union initializations assume this first. */
   struct callable *efun;
   struct array *array;
   struct mapping *mapping;
@@ -62,16 +63,22 @@ union anything
   struct object *object;
   struct program *program;
   struct pike_string *string;
+  struct pike_type *type;
   INT32 *refs;
-  INT_TYPE integer;
+  struct ref_dummy *dummy;
   FLOAT_TYPE float_number;
   struct svalue *lval;   /* only used on stack */
   union anything *short_lval;   /* only used on stack */
+  void *ptr;
 };
 
 #ifndef STRUCT_SVALUE_DECLARED
 #define STRUCT_SVALUE_DECLARED
 #endif
+
+/* Note: At least multisets overlays the type field and uses the top 4
+ * bits in it internally. */
+
 struct svalue
 {
   unsigned INT16 type;
@@ -86,7 +93,7 @@ struct svalue
 #define PIKE_T_FUNCTION 4
 #define PIKE_T_PROGRAM 5
 #define PIKE_T_STRING 6
-#define PIKE_T_TYPE 7		/* Not supported yet */
+#define PIKE_T_TYPE 7
 #define PIKE_T_INT 8
 #define PIKE_T_FLOAT 9
 
@@ -95,7 +102,8 @@ struct svalue
 #define T_VOID       16	/* Can't return any value */
 #define T_MANY       17
 
-#define T_MAPPING_DATA 242
+#define PIKE_T_RING 240
+#define PIKE_T_NAME 241		/* Named type. */
 #define PIKE_T_SCOPE 243	/* Not supported yet */
 #define PIKE_T_TUPLE 244	/* Not supported yet */
 #define T_ASSIGN 245
@@ -109,6 +117,11 @@ struct svalue
 #define T_AND 254
 #define T_OR 255
 
+/* These are only used together with describe() and friends. */
+#define T_STORAGE 10000
+#define T_MAPPING_DATA 10001
+#define T_PIKE_FRAME 10002
+#define T_MULTISET_DATA 10003
 
 #define tArr(VAL) "\000" VAL
 #define tArray tArr(tMix)
@@ -132,11 +145,11 @@ struct svalue
 #define MagictFunc(RET,ARGS) tFuncV(ARGS "", tVoid, RET)
 #define tFunction tFuncV("" ,tOr(tZero,tVoid),tOr(tMix,tVoid))
 #define tNone ""
-#define tPrg "\005"
-#define tProgram "\005"
+#define tPrg(X) "\005" X
+#define tProgram(X) "\005" X
 #define tStr "\006"
 #define tString "\006"
-#define tType "\007"
+#define tType(T) "\007" T
 #define tInt "\010\200\000\000\000\177\377\377\377"
 #define tInt0 "\010\000\000\000\000\000\000\000\000"
 #define tInt1 "\010\000\000\000\001\000\000\000\001"
@@ -171,11 +184,21 @@ struct svalue
 #define tOr9(A,B,C,D,E,F,G,H,I) tOr(A,tOr8(B,C,D,E,F,G,H,I))
 #define tMix "\373"
 #define tMixed "\373"
-#define tComplex tOr6(tArray,tMapping,tMultiset,tObj,tFunction,tProgram)
-#define tStringIndicable tOr5(tMapping,tObj,tFunction,tProgram,tMultiset)
+#define tComplex tOr6(tArray,tMapping,tMultiset,tObj,tFunction,tPrg(tObj))
+#define tStringIndicable tOr5(tMapping,tObj,tFunction,tPrg(tObj),tMultiset)
 #define tRef tOr(tString,tComplex)
 #define tIfnot(X,Y) tAnd(tNot(X),Y)
 #define tAny tOr(tVoid,tMix)
+#define tName(X,Y) "\361\0"X"\0"Y
+#if PIKE_BYTEORDER == 1234
+/* Little endian */
+#define tName1(X,Y) "\361\5"X"\0\0"Y
+#define tName2(X,Y) "\361\6"X"\0\0\0\0"Y
+#else /* PIKE_BYTEORDER != 1234 */
+/* Big endian */
+#define tName1(X,Y) "\361\1"X"\0\0"Y
+#define tName2(X,Y) "\361\2"X"\0\0\0\0"Y
+#endif /* PIKE_BYTEORDER == 1234 */
 
 #define tSimpleCallable tOr3(tArray,tFunction,tObj)
 #define tCallable tOr3(tArr(tSimpleCallable),tFunction,tObj)
@@ -220,14 +243,20 @@ struct svalue
 /* Max type handled by svalue primitives */
 #define MAX_TYPE PIKE_T_FLOAT
 
-#define NUMBER_NUMBER 0
-#define NUMBER_UNDEFINED 1
+#define NUMBER_NUMBER     0
+#define NUMBER_UNDEFINED  1
 #define NUMBER_DESTRUCTED 2
+#define NUMBER_LOCKED     USHRT_MAX	/* NOTE: Kludge, only in Pike 7.4. */
 
 #define FUNCTION_BUILTIN USHRT_MAX
 
 #define is_gt(a,b) is_lt(b,a)
-#define IS_ZERO(X) ((X)->type==PIKE_T_INT?(X)->u.integer==0:(1<<(X)->type)&(BIT_OBJECT|BIT_FUNCTION)?!svalue_is_true(X):0)
+
+/* SAFE_IS_ZERO is compatible with the old IS_ZERO, but you should
+ * consider using UNSAFE_IS_ZERO instead, since exceptions thrown from
+ * `! functions will be propagated correctly then. */
+#define UNSAFE_IS_ZERO(X) ((X)->type==PIKE_T_INT?(X)->u.integer==0:(1<<(X)->type)&(BIT_OBJECT|BIT_FUNCTION)?!svalue_is_true(X):0)
+#define SAFE_IS_ZERO(X) ((X)->type==PIKE_T_INT?(X)->u.integer==0:(1<<(X)->type)&(BIT_OBJECT|BIT_FUNCTION)?!safe_svalue_is_true(X):0)
 
 #define IS_UNDEFINED(X) ((X)->type==PIKE_T_INT&&!(X)->u.integer&&(X)->subtype==1)
 
@@ -261,30 +290,58 @@ do{ \
   } \
 }while(0)
 
+
+#ifdef PIKE_RUN_UNLOCKED
+#define add_ref(X) pike_atomic_inc32(&(X)->refs)
+#define sub_ref(X) pike_atomic_dec_and_test32(&(X)->refs)
+
+#if 0
+#define IF_LOCAL_MUTEX(X) X
+#define USE_LOCAL_MUTEX
+#define pike_lock_data(X) mt_lock(&(X)->mutex)
+#define pike_unlock_data(X) mt_unlock(&(X)->mutex)
+#else
+#define IF_LOCAL_MUTEX(X)
+#define pike_lock_data(X) pike_lockmem((X))
+#define pike_unlock_data(X) pike_unlockmem((X))
+#endif
+
+#else
+#define IF_LOCAL_MUTEX(X)
+#define add_ref(X) (void)((X)->refs++)
+#define sub_ref(X) (--(X)->refs > 0)
+#define pike_lock_data(X) (void)(X)
+#define pike_unlock_data(X) (void)(X)
+#endif
+
+
 #ifdef PIKE_DEBUG
 PMOD_EXPORT extern void describe(void *); /* defined in gc.c */
-#define check_type(T) if(T > MAX_TYPE && T!=T_LVALUE && T!=T_SHORT_LVALUE && T!=T_VOID && T!=T_DELETED && T!=T_ARRAY_LVALUE) fatal("Type error: %d\n",T)
+PMOD_EXPORT extern const char msg_type_error[];
+#define check_type(T) if(T > MAX_TYPE && T!=T_LVALUE && T!=T_SHORT_LVALUE && T!=T_VOID && T!=T_DELETED && T!=T_ARRAY_LVALUE) Pike_fatal(msg_type_error,T)
 
 #define check_svalue(S) debug_check_svalue(dmalloc_check_svalue(S,DMALLOC_LOCATION()))
 
+PMOD_EXPORT extern const char msg_sval_obj_wo_refs[];
 #define check_refs(S) do {\
  if((S)->type <= MAX_REF_TYPE && (!(S)->u.refs || (S)->u.refs[0] < 0)) { \
  describe((S)->u.refs); \
- fatal("Svalue to object without references.\n"); \
+ Pike_fatal(msg_sval_obj_wo_refs); \
 } }while(0)
 
+PMOD_EXPORT extern const char msg_ssval_obj_wo_refs[];
 #define check_refs2(S,T) do { \
 if((T) <= MAX_REF_TYPE && (S)->refs && (S)->refs[0] <= 0) {\
  describe((S)->refs); \
- fatal("(short) Svalue to object without references.\n"); \
+ Pike_fatal(msg_ssval_obj_wo_refs); \
 } }while(0)
 
 #ifdef DEBUG_MALLOC
 static inline struct svalue *dmalloc_check_svalue(struct svalue *s, char *l)
 {
   debug_malloc_update_location(s,l);
-#if 0
-  if(s->type <= MAX_REF_TYPE)
+#if 1
+  if(s && s->type <= MAX_REF_TYPE)
     debug_malloc_update_location(s->u.refs,l);
 #endif
   return s;
@@ -293,20 +350,30 @@ static inline struct svalue *dmalloc_check_svalue(struct svalue *s, char *l)
 static inline union anything *dmalloc_check_union(union anything *u,int type, char * l)
 {
   debug_malloc_update_location(u,l);
-#if 0
-  if(type <= MAX_REF_TYPE)
+#if 1
+  if(u && type <= MAX_REF_TYPE)
     debug_malloc_update_location(u->refs,l);
 #endif
   return u;
 }
 
-#define add_ref(X) ((INT32 *)debug_malloc_pass( &((X)->refs)))[0]++
+#undef add_ref
+#undef sub_ref
+
+#ifdef PIKE_RUN_UNLOCKED
+#define add_ref(X) pike_atomic_inc32((INT32 *)debug_malloc_pass( &((X)->refs)))
+#define sub_ref(X) pike_atomic_dec_and_test32((INT32 *)debug_malloc_pass( &((X)->refs)))
+#else
+#define add_ref(X) (((INT32 *)debug_malloc_pass( &((X)->refs)))[0]++)
+#define sub_ref(X) (--((INT32 *)debug_malloc_pass( &((X)->refs)))[0] > 0)
+#endif
+
+
 #else
 #define dmalloc_check_svalue(S,L) (S)
 #define dmalloc_check_union(U,T,L) (U)
-#define add_ref(X) (X)->refs++
-#endif
 
+#endif
 
 #else
 
@@ -316,80 +383,178 @@ static inline union anything *dmalloc_check_union(union anything *u,int type, ch
 #define check_refs2(S,T)
 #define dmalloc_check_svalue(S,L) (S)
 #define dmalloc_check_union(U,T,L) (U)
-#define add_ref(X) (X)->refs++
 
 #endif
 
-#define free_svalue(X) do { struct svalue *_s=(X); check_type(_s->type); check_refs(_s); if(_s->type<=MAX_REF_TYPE) { debug_malloc_touch(_s->u.refs); if(--*(_s->u.refs) <=0) { really_free_svalue(_s); } DO_IF_DMALLOC(_s->u.refs=(void *)-1;)  }}while(0)
-#define free_short_svalue(X,T) do { union anything *_s=(X); TYPE_T _t=(T); check_type(_t); check_refs2(_s,_t); if(_t<=MAX_REF_TYPE && _s->refs) if(--*(_s->refs) <= 0) { really_free_short_svalue(_s,_t); } DO_IF_DMALLOC(_s->refs=(void *)-1;) }while(0)
-#define add_ref_svalue(X) do { struct svalue *_tmp=(X); check_type(_tmp->type); check_refs(_tmp); if(_tmp->type <= MAX_REF_TYPE) { debug_malloc_touch(_tmp->u.refs); _tmp->u.refs[0]++; } }while(0)
-#define assign_svalue_no_free(X,Y) do { struct svalue _tmp, *_to=(X), *_from=(Y); check_type(_from->type); check_refs(_from);  *_to=_tmp=*_from; if(_tmp.type <= MAX_REF_TYPE) { debug_malloc_touch(_tmp.u.refs); _tmp.u.refs[0]++; } }while(0)
-#define assign_svalue(X,Y) do { struct svalue *_to2=(X), *_from2=(Y); if (_to2 != _from2) { free_svalue(_to2); assign_svalue_no_free(_to2, _from2); }  }while(0)
+
+/* This define
+ * should check that the svalue address (X) is on the local stack,
+ * the processor stack or in a locked memory object
+ *
+ * Or, it could just try to make sure it's not in an unlocked memory
+ * object...
+ */
+#define assert_svalue_locked(X)
+
+
+#define swap_svalues_unlocked(X,Y)  do {		\
+  struct svalue *_a=(X);				\
+  struct svalue *_b=(Y);				\
+  struct svalue _tmp;					\
+  assert_svalue_locked(_a); assert_svalue_locked(_b);	\
+  _tmp=*_a; *_a=*_b; *_b=_tmp;				\
+}while(0)
+
+#define free_svalue_unlocked(X) do {				\
+  struct svalue *_s=(X);					\
+  assert_svalue_locked(_s);					\
+  check_type(_s->type); check_refs(_s);				\
+  if(_s->type<=MAX_REF_TYPE) {					\
+    debug_malloc_touch(_s->u.refs);				\
+    if(sub_ref(_s->u.dummy) <=0) { really_free_svalue(_s); }	\
+  }								\
+  DO_IF_DMALLOC(_s->type=PIKE_T_UNKNOWN;_s->u.refs=(void *)-1); \
+  PIKE_MEM_WO(*_s);						\
+}while(0)
+
+#define free_short_svalue_unlocked(X,T) do {				\
+  union anything *_s=(X); TYPE_T _t=(T);				\
+  check_type(_t); check_refs2(_s,_t);					\
+  assert_svalue_locked(_s);						\
+  if(_t<=MAX_REF_TYPE && _s->refs) {					\
+    if(sub_ref(_s->dummy) <= 0) really_free_short_svalue(_s,_t);	\
+  }									\
+  DO_IF_DMALLOC(_s->refs=(void *)-1);					\
+  PIKE_MEM_WO(_s->refs);						\
+}while(0)
+
+#define add_ref_svalue_unlocked(X) do {				\
+  struct svalue *_tmp=(X);					\
+  check_type(_tmp->type); check_refs(_tmp);			\
+  if(_tmp->type <= MAX_REF_TYPE) add_ref(_tmp->u.dummy);	\
+}while(0)
+
+#define assign_svalue_no_free_unlocked(X,Y) do {	\
+  struct svalue _tmp;					\
+  struct svalue *_to=(X);				\
+  const struct svalue *_from=(Y);			\
+  check_type(_from->type); check_refs(_from);		\
+  *_to=_tmp=*_from;					\
+  if(_tmp.type <= MAX_REF_TYPE) add_ref(_tmp.u.dummy);	\
+}while(0)
+
+#define assign_svalue_unlocked(X,Y) do {	\
+  struct svalue *_to2=(X);			\
+  const struct svalue *_from2=(Y);		\
+  if (_to2 != _from2) {				\
+    free_svalue(_to2);				\
+     assign_svalue_no_free(_to2, _from2);	\
+  }						\
+}while(0)
+
+#define move_svalue(TO, FROM) do {					\
+    struct svalue *_to = (TO);						\
+    struct svalue *_from = (FROM);					\
+    dmalloc_touch_svalue(_from);					\
+    *_to = *_from;							\
+    DO_IF_DMALLOC(_from->type = PIKE_T_UNKNOWN; _from->u.refs = (void *) -1); \
+    PIKE_MEM_WO(*_from);						\
+  } while (0)
 
 extern struct svalue dest_ob_zero;
 
+#define free_mixed_svalues(X,Y) do {		\
+  struct svalue *s_=(X);			\
+  ptrdiff_t num_=(Y);				\
+  while(num_--)					\
+  {						\
+    dmalloc_touch_svalue(s_);			\
+    free_svalue(s_++);				\
+  }						\
+}while(0);
+
 #ifdef DEBUG_MALLOC
-#define free_svalues(X,Y,Z) debug_free_svalues((X),(Y),(Z), DMALLOC_LOCATION())
+#define free_svalues(X,Y,Z) debug_free_svalues((X),(Y),(Z), DMALLOC_LOCATION());
 #else
-#define free_svalues(X,Y,Z) debug_free_svalues((X),(Y),(Z))
+#define free_svalues(X,Y,Z) debug_free_svalues((X),(Y),(Z));
 #endif
 
+#define low_clear_svalues(X,Y,N) do {		\
+  struct svalue *s_=(X);			\
+  ptrdiff_t num_=(Y);				\
+  for(;num_-- > 0;s_++)				\
+  {						\
+      s_->type=PIKE_T_INT;			\
+      s_->subtype=(N);				\
+      s_->u.integer=0;				\
+  }						\
+}while(0)
+
+#define clear_svalues(X,Y) low_clear_svalues((X),(Y),NUMBER_NUMBER)
+#define clear_svalues_undefined(X,Y) low_clear_svalues((X),(Y),NUMBER_UNDEFINED)
+
+#define really_free_short_svalue(U, TYPE) do {				\
+    union anything *any_ = (U);						\
+    really_free_short_svalue_ptr (&any_->ptr, (TYPE));			\
+  } while (0)
+
 /* Prototypes begin here */
-PMOD_EXPORT void really_free_short_svalue(union anything *s, TYPE_T type);
+PMOD_EXPORT void really_free_short_svalue_ptr(void **s, TYPE_T type);
 PMOD_EXPORT void really_free_svalue(struct svalue *s);
 PMOD_EXPORT void do_free_svalue(struct svalue *s);
 PMOD_EXPORT void debug_free_svalues(struct svalue *s, size_t num, INT32 type_hint DMALLOC_LINE_ARGS);
+PMOD_EXPORT void debug_free_mixed_svalues(struct svalue *s, size_t num, INT32 type_hint DMALLOC_LINE_ARGS);
 PMOD_EXPORT void assign_svalues_no_free(struct svalue *to,
-			    struct svalue *from,
+			    const struct svalue *from,
 			    size_t num,
-			    INT32 type_hint);
+			    TYPE_FIELD type_hint);
 PMOD_EXPORT void assign_svalues(struct svalue *to,
-		    struct svalue *from,
+		    const struct svalue *from,
 		    size_t num,
 		    TYPE_FIELD types);
 PMOD_EXPORT void assign_to_short_svalue(union anything *u,
 			    TYPE_T type,
-			    struct svalue *s);
+			    const struct svalue *s);
 PMOD_EXPORT void assign_to_short_svalue_no_free(union anything *u,
 				    TYPE_T type,
-				    struct svalue *s);
+				    const struct svalue *s);
 PMOD_EXPORT void assign_from_short_svalue_no_free(struct svalue *s,
-				      union anything *u,
+				      const union anything *u,
 				      TYPE_T type);
 PMOD_EXPORT void assign_short_svalue_no_free(union anything *to,
-				 union anything *from,
+				 const union anything *from,
 				 TYPE_T type);
 PMOD_EXPORT void assign_short_svalue(union anything *to,
-			 union anything *from,
+			 const union anything *from,
 			 TYPE_T type);
-PMOD_EXPORT unsigned INT32 hash_svalue(struct svalue *s);
-PMOD_EXPORT int svalue_is_true(struct svalue *s);
-PMOD_EXPORT int is_identical(struct svalue *a, struct svalue *b);
-PMOD_EXPORT int is_eq(struct svalue *a, struct svalue *b);
-PMOD_EXPORT int low_is_equal(struct svalue *a,
-		 struct svalue *b,
+PMOD_EXPORT unsigned INT32 hash_svalue(const struct svalue *s);
+PMOD_EXPORT int svalue_is_true(const struct svalue *s);
+PMOD_EXPORT int safe_svalue_is_true(const struct svalue *s);
+PMOD_EXPORT int is_identical(const struct svalue *a, const struct svalue *b);
+PMOD_EXPORT int is_eq(const struct svalue *a, const struct svalue *b);
+PMOD_EXPORT int low_is_equal(const struct svalue *a,
+		 const struct svalue *b,
 		 struct processing *p);
 PMOD_EXPORT int low_short_is_equal(const union anything *a,
 		       const union anything *b,
 		       TYPE_T type,
 		       struct processing *p);
-PMOD_EXPORT int is_equal(struct svalue *a,struct svalue *b);
-PMOD_EXPORT int is_lt(struct svalue *a,struct svalue *b);
-PMOD_EXPORT void describe_svalue(struct svalue *s,int indent,struct processing *p);
-PMOD_EXPORT void print_svalue (FILE *out, struct svalue *s);
-PMOD_EXPORT void clear_svalues(struct svalue *s, ptrdiff_t num);
-PMOD_EXPORT void clear_svalues_undefined(struct svalue *s, ptrdiff_t num);
+PMOD_EXPORT int is_equal(const struct svalue *a, const struct svalue *b);
+PMOD_EXPORT int is_lt(const struct svalue *a, const struct svalue *b);
+PMOD_EXPORT void describe_svalue(const struct svalue *s,int indent,struct processing *p);
+PMOD_EXPORT void print_svalue (FILE *out, const struct svalue *s);
+PMOD_EXPORT void print_svalue_compact (FILE *out, const struct svalue *s);
 PMOD_EXPORT void copy_svalues_recursively_no_free(struct svalue *to,
-				      struct svalue *from,
+				      const struct svalue *from,
 				      size_t num,
 				      struct processing *p);
-void check_short_svalue(union anything *u, TYPE_T type);
-void debug_check_svalue(struct svalue *s);
-PMOD_EXPORT void real_gc_xmark_svalues(struct svalue *s, ptrdiff_t num);
-PMOD_EXPORT void real_gc_check_svalues(struct svalue *s, size_t num);
-void gc_check_weak_svalues(struct svalue *s, size_t num);
-PMOD_EXPORT void real_gc_check_short_svalue(union anything *u, TYPE_T type);
-void gc_check_weak_short_svalue(union anything *u, TYPE_T type);
+void check_short_svalue(const union anything *u, TYPE_T type);
+void debug_check_svalue(const struct svalue *s);
+PMOD_EXPORT void real_gc_xmark_svalues(const struct svalue *s, ptrdiff_t num);
+PMOD_EXPORT void real_gc_check_svalues(const struct svalue *s, size_t num);
+void gc_check_weak_svalues(const struct svalue *s, size_t num);
+PMOD_EXPORT void real_gc_check_short_svalue(const union anything *u, TYPE_T type);
+void gc_check_weak_short_svalue(const union anything *u, TYPE_T type);
 PMOD_EXPORT TYPE_FIELD real_gc_mark_svalues(struct svalue *s, size_t num);
 TYPE_FIELD gc_mark_weak_svalues(struct svalue *s, size_t num);
 int real_gc_mark_short_svalue(union anything *u, TYPE_T type);
@@ -402,7 +567,7 @@ PMOD_EXPORT int real_gc_cycle_check_short_svalue(union anything *u, TYPE_T type)
 int gc_cycle_check_weak_short_svalue(union anything *u, TYPE_T type);
 void real_gc_free_svalue(struct svalue *s);
 void real_gc_free_short_svalue(union anything *u, TYPE_T type);
-PMOD_EXPORT INT32 pike_sizeof(struct svalue *s);
+PMOD_EXPORT INT32 pike_sizeof(const struct svalue *s);
 int svalues_are_constant(struct svalue *s,
 			 INT32 num,
 			 TYPE_FIELD hint,
@@ -443,5 +608,160 @@ int svalues_are_constant(struct svalue *s,
 
 #endif /* !NO_PIKE_SHORTHAND */
 
+#if 0 /* PIKE_RUN_UNLOCKED */
+
+#include "pike_error.h"
+
+/*#define swap_svalues swap_svalues*/
+/*#define free_svalue free_svalue_unlocked*/
+/*#define free_short_svalue free_short_svalue_unlocked */
+/*#define add_ref_svalue add_ref_svalue_unlocked*/
+
+/*
+ * These don't work right now - Hubbe
+ */
+#define assign_svalue_no_free assign_svalue_no_free_unlocked
+#define assign_svalue assign_svalue_unlocked
+
+/* FIXME:
+ * These routines assumes that pointers are 32 bit
+ * and svalues 64 bit!!!!! - Hubbe
+ */
+
+#ifndef swap_svalues 
+#define swap_svalues swap_svalues_unlocked
+#endif
+
+#ifndef free_svalue
+static inline void free_svalue(struct svalue *s)
+{
+  INT64 tmp;
+  struct svalue zero;
+  zero.type=PIKE_T_INT;
+  tmp=pike_atomic_swap64((INT64 *)s, *(INT64 *)&zero);
+  free_svalue_unlocked((struct svalue *)&tmp);
+}
+#endif
+
+#ifndef free_short_svalue
+static inline void free_short_svalue(union anything *s, INT16 t)
+{
+  if(t <= MAX_REF_TYPE)
+  {
+    INT32 tmp;
+    tmp=pike_atomic_swap32((INT32 *)s, 0);
+    free_short_svalue_unlocked((union anything *)&tmp, t);
+  }
+}
+#endif
+
+#ifndef add_ref_svalue
+static inline void add_ref_svalue(struct svalue *s)
+{
+  INT64 sv;
+  sv=pike_atomic_get64((INT64 *)s);
+  add_ref_svalue_unlocked((struct svalue *)&sv);
+}
+#endif
+
+#ifndef assign_svalue_no_free
+void assign_svalue_no_free(struct svalue *to, const struct svalue *from)
+{
+  INT64 tmp, sv;
+  sv=pike_atomic_get64((INT64 *)from);
+#ifdef PIKE_DEBUG
+  if(sv != *(INT64*)from)
+  {
+    fprintf(stderr,"pike_atomic_get64() is broken %llx != %llx (%08x%08x)!\n",
+	    sv,
+	    *(INT64*)from,
+	    ((INT32*)from)[1], ((INT32*)from)[0]);
+    abort();
+  }
+#endif
+  add_ref_svalue_unlocked((struct svalue *)&sv);
+  pike_atomic_set64((INT64 *)to, sv);
+#ifdef PIKE_DEBUG
+  if(*(INT64*)to != *(INT64*)from)
+  {
+    fprintf(stderr,"pike_atomic_set64() is broken!\n");
+    abort();
+  }
+#endif
+}
+#endif
+
+#ifndef assign_svalue
+static inline void assign_svalue(struct svalue *to, const struct svalue *from)
+{
+  INT64 tmp, sv;
+  if(to != from)
+  {
+    sv=pike_atomic_get64((INT64 *)from);
+    add_ref_svalue_unlocked((struct svalue *)&sv);
+    tmp=pike_atomic_swap64((INT64 *)to, sv);
+    free_svalue_unlocked((struct svalue *)&tmp);
+  }
+}
+#endif
+
+#else /* FOO_PIKE_RUN_UNLOCKED */
+#define swap_svalues swap_svalues
+#define free_svalue free_svalue_unlocked
+#define free_short_svalue free_short_svalue_unlocked 
+#define add_ref_svalue add_ref_svalue_unlocked
+#define assign_svalue_no_free assign_svalue_no_free_unlocked
+#define assign_svalue assign_svalue_unlocked
+#endif /* FOO_PIKE_RUN_UNLOCKED */
+
+#ifdef PIKE_RUN_UNLOCKED
+#include "pike_threadlib.h"
+#endif
+
+/* 
+ * Note to self:
+ * It might be better to use a static array of mutexes instead
+ * and just lock mutex ptr % array_size instead.
+ * That way I wouldn't need a mutex in each memory object,
+ * but it would cost a couple of cycles in every lock/unlock
+ * operation instead.
+ */
+#define PIKE_MEMORY_OBJECT_MEMBERS \
+  INT32 refs \
+  DO_IF_SECURITY(; struct object *prot) \
+  IF_LOCAL_MUTEX(; PIKE_MUTEX_T mutex)
+
+#ifdef PIKE_SECURITY
+#ifdef USE_LOCAL_MUTEX
+#define PIKE_CONSTANT_MEMOBJ_INIT(refs) refs, 0, PTHREAD_MUTEX_INITIALIZER
+#else
+#define PIKE_CONSTANT_MEMOBJ_INIT(refs) refs, 0
+#endif
+#else
+#ifdef USE_LOCAL_MUTEX
+#define PIKE_CONSTANT_MEMOBJ_INIT(refs) refs, PTHREAD_MUTEX_INITIALIZER
+#else
+#define PIKE_CONSTANT_MEMOBJ_INIT(refs) refs
+#endif
+#endif
+
+#define INIT_PIKE_MEMOBJ(X) do {			\
+  struct ref_dummy *v_=(struct ref_dummy *)(X);		\
+  v_->refs=1;						\
+  DO_IF_SECURITY( INITIALIZE_PROT(v_) );		\
+  IF_LOCAL_MUTEX(mt_init_recursive(&(v_->mutex)));	\
+}while(0)
+
+#define EXIT_PIKE_MEMOBJ(X) do {		\
+  struct ref_dummy *v_=(struct ref_dummy *)(X);		\
+  DO_IF_SECURITY( FREE_PROT(v_) );		\
+  IF_LOCAL_MUTEX(mt_destroy(&(v_->mutex)));	\
+}while(0)
+
+
+struct ref_dummy
+{
+  PIKE_MEMORY_OBJECT_MEMBERS;
+};
 
 #endif /* !SVALUE_H */
